/*
 ******************************************************************************
 * ESO/DMD
 *
 * "@(#) $Id: ngamsCClientLib.c,v 1.145 2006/02/01 16:59:38 jknudstr Exp $" 
 *
 * who       when      what
 * --------  --------  --------------------------------------------------
 * jknudstr  30/08/01  created
 */

/**
NAME
ngamsCClientLib - NG/AMS C-API library

SYNOPSIS
#include "ngams.h"

If HTTP authorization is enabled on the server side, it is necessary to
call the function ngamsSetAuthorization() with an access code, which is
provided by the responsibles of the server side before starting to 
communicate with the remote server.

 
In general for the NG/AMS interface functions listed below,
the "host" parameter is the name of the host where the NG/AMS Server
is running. E.g.: "arcdev1.hq.eso.org". The "port" parameter is the
socket port, which the NG/AMS Server is waiting on.

If the parameter "wait" is set to 0, an immediate reply to the
request will be generated, i.e. before the request has been handled.

The parameter "status" is a structure containing the following members:

Data Type       Member     Description
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
ngamsSMALL_BUF  date       Date for handling query.
ngamsSTAT       errorCode  Error code giving status for the query.
                           See #1.
ngamsSMALL_BUF  hostId     Host ID for host where the NG/AMS Server 
                           is running.
ngamsHUGE_BUF   message    Message from the NG/AMS Server.
ngamsSMALL_BUF  status     Status of query ("OK" | "FAILURE").
ngamsSMALL_BUF  state      State of the NG/AMS Server.
ngamsSMALL_BUF  subState   Sub-State of the NG/AMS Server.
ngamsSMALL_BUF  version    Version of the NG/AMS Server.
char            replyData  Pointer array of pointers pointing to 
                           allocated buffers containing the reply data
                           from the request.


#1: The following error codes (internal to the NG/AMS C API) 
    are defined (data type: ngamsSTAT):

Error Macro             Description
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
ngamsSTAT_SUCCESS       Query successfully executed.

ngamsERR_HOST           No such host.
ngamsERR_SOCK           Cannot create socket.
ngamsERR_CON            Cannot connect to host/server.

ngamsERR_COM            Problem occurred during socket connection.
ngamsERR_TIMEOUT        Timeout encountered while communication with
                        server.

ngamsERR_WR_HD          Write error on socket while writing header.
ngamsERR_WR_DATA        Write error on socket while writing data.
ngamsERR_RD_DATA        Read error while reading data.
ngamsERR_INV_REPLY      Invalid reply from data server.

ngamsERR_FILE           Invalid filename specified.
ngamsERR_ALLOC_MEM      Cannot allocate memory.

ngamsERR_UNKNOWN_STAT   Unknown status code.
ngamsERR_UNKNOWN_CMD    Unknown command issued.
ngamsERR_INV_TARG_FILE  Invalid target filename specified.
ngamsERR_INV_PARS       Invalid parameters given.

ngamsSRV_OK             Request successfully handled by server.
ngamsSRV_REDIRECT       The reply is an HTTP redirection response.

ngamsSRV_INV_QUERY      Invalid query.   

 
Apart from that, the errors defined by NG/AMS can be returned.


All functions return ngamsSTAT_SUCCESS in case of success. In case of
an error a termination status within the set of status codes 
given above.


The following macros are defined for referring to NG/AMS commands:
  
Command Macros (#2)       Description
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
ngamsCMD_ARCHIVE          Archive file.
ngamsCMD_CLONE            Clone files.
ngamsCMD_EXIT             Make NG/AMS Server exit.
ngamsCMD_INIT             Re-initialize the server (same as Offline->Online).
ngamsCMD_LABEL            Make NG/AMS print out a disk label.
ngamsCMD_ONLINE           Bring NG/AMS Server Online.
ngamsCMD_OFFLINE          Bring NG/AMS Server Offline.
ngamsCMD_REGISTER         Register files on a disk.
ngamsCMD_REMDISK          Remove a disk from NGAS.
ngamsCMD_REMFILE          Remove a file from NGAS.
ngamsCMD_RETRIEVE         Retrieve a file.
ngamsCMD_STATUS           Query status information from NG/AMS.
ngamsCMD_SUBSCRIBE        Subscribe to a NG/AMS Server.
ngamsCMD_UNSUBSCRIBE      Un-subscribe/cancel a previous subscription.

#2: All command macros exist also as a string, which carries the name
    of the enumerated macro name with a "_STR" appended.
  

In the following, the functions provided for interacting with NG/AMS
are listed. The specific parameters for each function are listed in
connection with the function. The parameters used by several 
functions are as follows:

host:         Host name of NG/AMS Server to contact.
 
port:         Port number used by NG/AMS Server to contact.

timeoutSecs:  Timeout in seconds to apply while executing the request.

wait:         Wait for the NG/AMS Server to finish the request (=1)
              completely, or return an immediate response (=0).

status:       Pointer to ngamsSTATUS structure containing the
              status of the query.

parArray:     Instance of the ngamsPAR_ARRAY structure containing the
              parameters of the query. This should be filled with the 
              function: ngamsAddParAndVal(). Before usage, the structure
              should be initialized invoking ngamsResetParArray() on it.

EXAMPLES
To archive a file using the API the following must be called from
the application:

#include "ngams.h"

ngamsSTATUS      status;
if (ngamsArchive("wfinau", "7171", 10, "/home/data/MyFile.fits", "",
                 1, 0, &status) != ngamsSTAT_SUCCESS)
    {
    ngamsDumpStatStdout(&status);
    ... error handling ...
    }


To retrieve a file into the directory "/home/data/target_dir". The
name will be the same as the File ID:

#include "ngams.h"

ngamsSTATUS      status;
if (ngamsRetrieve2File("wfinau", "7171", 30, 
                       "WFI.2001-10-21T23:24:03.925",
                       -1, "", "", "/home/data/target_dir", 
                       &status) != ngamsSTAT_SUCCESS)
    {
    ngamsDumpStatStdout(&status);
    ... error handling ...
    }


CAUTIONS
Remember to perform a call to ngamsInitStatus() right after declaring 
an instance of the ngamsSTATUS class, and a call to ngamsFreeStatus()
after each call to one of the commands used to issue commands to NG/AMS.
Memory may be allocated dynamically and needs to be freed.

If data is returned by the NG/AMS Server the member "replyData" will
have a pointer pointing to each block of data. It is the responsibility
of the calling application to free these data chunks. The function
"ngamsFreeStatus()" can be used for this.

-----------------------------------------------------------------------------
*/

static char *rcsId="@(#) $Id: ngamsCClientLib.c,v 1.145 2006/02/01 16:59:38 jknudstr Exp $"; 
static void *use_rcsId = ((void)&use_rcsId,(void *) &rcsId);

#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <poll.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <limits.h> 
#include <sys/time.h>
#include <stdarg.h>
#include <sys/types.h>
#include <dirent.h>
#include <pthread.h>
#include <time.h>

#ifndef S_IFMT
    #define S_IFMT   __S_IFMT
    #define S_IFDIR  __S_IFDIR
#endif

#include "ngams.h"
#include "ngamsVERSION.h"


char *_ngamsLicense(void);
char *_ngamsManPage(void);

/* IMPL: Add entry/exit debug logs in all functions where relevant. */

/**
*******************************************************************************
Some global variables - should get rid of them.
*******************************************************************************
*/

/* Server contact information. */
ngamsSRV_INFO   _srvInfo;
ngamsSRV_INFO   *_ngamsSrvInfo() {return &_srvInfo;}
void             _ngamsShuffleSrvs()
{
    int             idx, idx1[ngamsMAX_SRVS],  idx2[ngamsMAX_SRVS], count, n;
    ngamsSRV_INFO   srvInfo;

    ngamsLogDebug("Entering _ngamsShuffleSrvs() ...");

    /* IMPL: This function is not thread safe. */
    for (n = 0; n < _srvInfo.numberOfSrvs; n++) idx2[n] = -1;
    count = 0;
    srand(time(NULL));
    while (1)
	{
	idx = ((double)_srvInfo.numberOfSrvs * rand() / (double)RAND_MAX);
	if (idx2[idx] == -1)
	    {
	    idx1[count] = idx;
	    idx2[idx] = 0;
	    count++;
	    }
	if (count == _srvInfo.numberOfSrvs) break;
	}
    memcpy(&srvInfo, &_srvInfo, sizeof(ngamsSRV_INFO));
    for (n = 0; n < _srvInfo.numberOfSrvs; n++) 
	{
	strcpy(srvInfo.hosts[idx1[n]], _srvInfo.hosts[n]);
	srvInfo.ports[idx1[n]] = _srvInfo.ports[n];
	}
    memcpy(&_srvInfo, &srvInfo, sizeof(ngamsSRV_INFO));

    ngamsLogDebug("Leaving _ngamsShuffleSrvs()");
}
ngamsSTAT       _ngamsGetNextSrv(char  **host,
				 int   *port)
{
    ngamsLogDebug("Entering _ngamsGetNextSrv() ...");

    /* IMPL: Should maybe make this thread-safe using a semaphore */
    _ngamsSrvInfo()->srvIdx = ((_ngamsSrvInfo()->srvIdx + 1) % 
			       _ngamsSrvInfo()->numberOfSrvs);
    *host = _ngamsSrvInfo()->hosts[_ngamsSrvInfo()->srvIdx];
    *port = _ngamsSrvInfo()->ports[_ngamsSrvInfo()->srvIdx]; 
    ngamsLogDebug("Leaving _ngamsGetNextSrv()");
    return ngamsSTAT_SUCCESS;
}


/* Authorization user/password */
char           *_authorization = NULL;


/* Time-out for poll() */
float          _replyTimeout;

/* Log Conditions */
ngamsMED_BUF   _logFile;
int            _logLevel = 0;
int            _logRotate = -1;
int            _logHistory = -1;
int            _verboseLevel = 0;

/**
*******************************************************************************
*/


/**
FUNCTIONS
*/

/**
*******************************************************************************
Miscellaneous, internal utility functions.
*******************************************************************************
*/

void _setReplyTimeoutFromFloat(const float timeoutSecs)
{
    if (timeoutSecs <= 0) 
	_replyTimeout = 100000000;
    else
	_replyTimeout = timeoutSecs;
}

float _getTimeOut()
{
    float timeOut;
    timeOut = _replyTimeout;
    if (timeOut == 0) timeOut = 100000;
    return timeOut;
}

/*
int _pollFd(const int  fd,
            const float timeOut)

Poll file descriptor with a timeout in seconds. If timeout is 0 (or less), 
the internal timeout in the C-API is taken.

Returns:   ngamsSTAT_SUCCESS = OK
           ngamsERR_COM      = Problem with socket connection (broken).
           ngamsERR_TIMEOUT  = Timeout.
*/
int _pollFd(const int   fd,
	    const float timeOut)
{
    int readyFds = 0, locTimeOut, stat;
    struct pollfd *fds;
    struct pollfd fdStr;

    ngamsLogDebug("Entering _pollFd() ...");
    memset(&fdStr, 0, sizeof(struct pollfd));
    fdStr.fd = fd;
    fdStr.events = POLLIN;
    fds = &fdStr;
    ngamsLogDebug("Requested timeout: %.6fs", timeOut);
    if (timeOut > 0)
	locTimeOut = (1000 * timeOut);
    else
	locTimeOut = (1000 * _getTimeOut());
    ngamsLogDebug("Polling for ready fd's. Timeout: %dms", locTimeOut);
    readyFds = poll(fds, 1, locTimeOut);
    if (readyFds <= -1)
	{
        stat = ngamsERR_COM;
	goto errExit;
	}
    else if (readyFds == 0)
	{
        stat = ngamsERR_TIMEOUT;
	goto errExit;
	}
    ngamsLogDebug("Leaving _pollFd()");
    return ngamsSTAT_SUCCESS;

 errExit:
    ngamsLogDebug("Leaving _pollFd()/FAILURE. Status: %d", stat);
    return stat;
}

/**
*******************************************************************************
*/

/**
*******************************************************************************
Functions to Handle NG/AMS Commands.
*******************************************************************************
*/

/**
ngamsSTAT ngamsUnpackStatus(const char    *xmlDoc,
			    ngamsSTATUS   *status)
Unpack an XML Status Document received from the NG/AMS Server.

xmlDoc:     XML Status Document (string).

status:     Instance of the ngamsSTATUS structure, which will contain
            the deoded status.

Returns:    One of the following status codes:

              ngamsSTAT_SUCCESS or ngamsERR_INV_REPLY
*/
ngamsSTAT ngamsUnpackStatus(const char    *xmlDoc,
			    ngamsSTATUS   *status)
{ 
    char  errCode[64], *tmpBuf = NULL;
    int   OK = ngamsSTAT_SUCCESS, stat;

    ngamsLogDebug("Entering ngamsUnpackStatus() ...");
    if ((ngamsGetXmlAttr(xmlDoc, "Status", "Date", 
			 sizeof(ngamsSMALL_BUF), status->date) != OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "HostId", 
			 sizeof(ngamsSMALL_BUF), status->hostId) != OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "Message", 
			 sizeof(ngamsHUGE_BUF), status->message) != OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "State", 
			 sizeof(ngamsSMALL_BUF), status->state) != OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "SubState", 
			 sizeof(ngamsSMALL_BUF), status->subState)!= OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "Status", 
			 sizeof(ngamsSMALL_BUF), status->status) != OK) ||
	(ngamsGetXmlAttr(xmlDoc, "Status", "Version", 
			 sizeof(ngamsSMALL_BUF), status->version) != OK))
	{
	ngamsStat2Str(ngamsERR_INV_REPLY, status->message);
	status->errorCode = ngamsERR_INV_REPLY;
	stat = ngamsERR_INV_REPLY;
	goto errExit;
	}
    if ((strcmp(status->status, NGAMS_FAILURE) == 0) &&
	((strstr(status->message, ":ERROR:") != NULL) ||
	 (strstr(status->message, ":WARNING:") != NULL) ||
	 (strstr(status->message, ":ALERT:") != NULL)))
	{
	tmpBuf = malloc(strlen(status->message) + 1);
	strcpy(tmpBuf, status->message);
	strtok(tmpBuf, ":");
	memset(errCode, 0, (64 * sizeof(char)));
	strncpy(errCode, strtok(NULL, ":"), 64);
	if (errCode != NULL) status->errorCode = atoi(errCode);
	free(tmpBuf);
	}
    else
	status->errorCode = ngamsSTAT_SUCCESS;

    if ((strcmp(status->status, NGAMS_FAILURE) != 0) &&
	(strcmp(status->status, NGAMS_SUCCESS) != 0))
	{
	ngamsStat2Str(ngamsERR_INV_REPLY, status->message);
	status->errorCode = ngamsERR_INV_REPLY;
	stat = ngamsERR_INV_REPLY;
	goto errExit;
	}

    /* Make the data reply array refer to the data received */
    status->replyData[0] = (char *)xmlDoc;

    ngamsLogDebug("Leaving ngamsUnpackStatus().");
    return ngamsSTAT_SUCCESS;

 errExit:
    ngamsLogDebug("Leaving ngamsUnpackStatus()/FAILURE. Status: %d", stat);
    return stat;
}


/**
ngamsSTAT ngamsHandleStatus(int             retCode,
			    ngamsHTTP_DATA  *repDataRef,
			    ngamsSTATUS     *status)
Handle the status of an HTTP response. The retCode is return value from
the ngamsHttpGet() or ngamsHttpPost() functions.

retCode:       Return code from ngamsHttpGet() or ngamsHttpPost().

repDataRef:    Instance of the ngamsHTTP_DATA structure containing the
               the reference to the data in the HTTP response.

status:        Instance of the ngamsSTATUS structure, which will contain
               the deoded status.

Returns:       ngamsSTAT_SUCCESS.
*/
ngamsSTAT ngamsHandleStatus(int             retCode,
			    ngamsHTTP_DATA  *repDataRef,
			    ngamsSTATUS     *status)
{
    ngamsSMALL_BUF  errBuf;

    ngamsLogDebug("Entering ngamsHandleStatus() ...");
    ngamsInitStatus(status);
    if (repDataRef->pdata != NULL) 
	ngamsUnpackStatus(repDataRef->pdata, status);
    else
	{
	ngamsStat2Str(retCode, status->message);
	if ((retCode != ngamsSRV_OK) && ((retCode != ngamsSTAT_SUCCESS)))
	    strcpy(status->status, NGAMS_FAILURE);
	else
	    strcpy(status->status, NGAMS_SUCCESS);
	status->errorCode = retCode;
	}

    /* If the error reported is ngamsERR_TIMEOUT we add the time-out time
     * to the message for completenes.
     */
    if (status->errorCode == ngamsERR_TIMEOUT)
        {
	sprintf(errBuf, ". Timeout: %.3fs", (float)(_replyTimeout));
	strcat(status->message, errBuf);
	}

    ngamsLogDebug("Leaving ngamsHandleStatus()");    
    return ngamsSTAT_SUCCESS;
}


/**
ngamsSTAT ngamsGenSendData(const char            *host,
                           const int             port,
                           const ngamsCMD        cmdCode,
                           const float           timeoutSecs,
		           const char            *fileUri,
			   const char            *mimeType,
			   const ngamsPAR_ARRAY  *parArray,
			   ngamsSTATUS           *status)

If the URI referring to the file is a filename, the function will carry
out an HTTP POST request. Otherwise if the URI is a URL, the function
will make an HTTP GET request.

fileUri:        URI referring to file.

mimeType:       Mime-type of data. If set to '\0' a generic mime-type will
                be contained in the HTTP header 'Content-type' and the
                NG/AMS Server will determine the mime-type itself.

parArray:       Instance of the ngamsPAR_ARRAY structure containing the
                parameters of the query.

Returns:        Returns ngamsSTAT_SUCCESS or a status code returned by
                one of the functions:

                  ngamsHttpGet(), ngamsHttpPost()
*/
ngamsSTAT ngamsGenSendData(const char            *host,
			   const int             port,
			   const ngamsCMD        cmdCode,
			   const float           timeoutSecs,
			   const char            *fileUri,
			   const char            *mimeType,
			   const ngamsPAR_ARRAY  *parArray,
			   ngamsSTATUS           *status)
{
    int              retCode, repDataLen, i; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     mtBuf, url, tmpBuf, tmpFileUri, cmd;
    ngamsHUGE_BUF    contDisp;

    ngamsLogDebug("Entering ngamsGenSendData() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    memset(url, 0, sizeof(ngamsMED_BUF));

    ngamsCmd2Str(cmdCode, cmd);

    /* Encode the file URI */
    strcpy(tmpFileUri, ngamsEncodeUrlVal(fileUri, 1));

    if ((strstr(fileUri, "file:") != NULL) ||	
	(strstr(fileUri, "http:") != NULL) ||
	(strstr(fileUri, "ftp:") != NULL))
	{
	/* Data is pulled */
	sprintf(url, "%s?filename=\"%s\"&mime_type=\"%s\"",
		cmd, tmpFileUri, mimeType);
	for (i = 0; i < parArray->idx; i++)
	    {
	    sprintf(tmpBuf, "&%s=\"%s\"", parArray->parArray[i], 
		    ngamsEncodeUrlVal(parArray->valArray[i], 1));
	    strcat(url, tmpBuf);
	    }
	if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, url, 1,
				    &repDataRef, &repDataLen, &httpResp,
				    httpHdr)) != ngamsSTAT_SUCCESS)
	    goto errExit;
	}
    else
	{
	/* Data is pushed */
	sprintf(contDisp, "attachment; filename=\"%s\"", tmpFileUri);
	for (i = 0; i < parArray->idx; i++)
	    {
	    sprintf(tmpBuf, "; %s=\"%s\"", parArray->parArray[i], 
		    ngamsEncodeUrlVal(parArray->valArray[i], 1));
	    strcat(contDisp, tmpBuf);
	    }
	if (*mimeType == '\0') 
	    strcpy(mtBuf, ngamsARHIVE_REQ_MT);
	else
	    strcpy(mtBuf, mimeType);
	if ((retCode = ngamsHttpPost(host, port, ngamsUSER_AGENT, cmd, mtBuf,
				     contDisp, fileUri, "", 0, &repDataRef, 
				     &repDataLen, &httpResp, httpHdr))
	    != ngamsSTAT_SUCCESS)
	    goto errExit;
	}
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsGenSendData()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsGenSendData()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsArchive(const char          *host,
                       const int           port,
                       const float         timeoutSecs,
                       const char          *fileUri,
                       const char          *mimeType,
                       const int           noVersioning,
   	               const int           wait,
  	               ngamsSTATUS         *status)
Archive a file into the NGAS Archive.

fileUri:      Reference name for the file to archive.

mimeType:     The mime-type of the file to archive. In some cases
              it is not possible for NG/AMS to determine the mime-type
              of a data file to be archived, e.g. when the file being is
              archived is RETRIEVEd from another NGAS Host. For efficiency
              it is thus better to indicate the mime-type to enable
              NG/AMS to store the file directly on the target disk.
              If not use this can be put to "".

noVersioning: If set to 1 no new File Version will be generated for
              the file being archived even though a file with that
              File ID was already existing.

Returns:      Execution status (ngamsSTAT_SUCCESS|ngamsSTAT_FAILURE).
*/
ngamsSTAT ngamsArchive(const char          *host,
		       const int           port,
		       const float         timeoutSecs,
		       const char          *fileUri,
		       const char          *mimeType, 
		       const int           noVersioning,
		       const int           wait,
		       ngamsSTATUS         *status)
{
    int             stat;
    ngamsMED_BUF    tmpBuf;
    ngamsPAR_ARRAY  parArray;

    ngamsLogDebug("Entering ngamsArchive() ...");
    ngamsResetParArray(&parArray);
    sprintf(tmpBuf, "%d", noVersioning);
    ngamsAddParAndVal(&parArray, "no_versioning", tmpBuf);
    sprintf(tmpBuf, "%d", wait);
    ngamsAddParAndVal(&parArray, "wait", tmpBuf);
    if (timeoutSecs != -1)
	sprintf(tmpBuf, "%d", (int)(timeoutSecs + 0.5));
    else
	sprintf(tmpBuf, "-1");
    ngamsAddParAndVal(&parArray, "time_out", tmpBuf);
    stat = ngamsGenSendData(host, port, ngamsCMD_ARCHIVE, timeoutSecs, fileUri,
			    mimeType, &parArray, status);
    ngamsLogDebug("Leaving ngamsArchive()");
    return stat;
}

/**
ngamsSTAT ngamsArchiveFromMem(const char          *host,
                              const int           port,
                              const float         timeoutSecs,
                              const char          *fileUri,
                              const char          *bufPtr,
                              const int           size,
                              const char          *mimeType,
                              const int           noVersioning,
                              const int           wait,
                              ngamsSTATUS         *status)
Archive a file which contents have been loaded into a buffer in memory.

fileUri,  
mimeType,
noVersioning: See ngamsArchive().

bufPtr:       Pointer to buffer containing the contents of the file.

size:         Size in bytes of the data loaded into memory.

Returns:      See ngamsArchive().
*/
ngamsSTAT ngamsArchiveFromMem(const char          *host,
			      const int           port,
			      const float         timeoutSecs,
			      const char          *fileUri,
			      const char          *bufPtr,
			      const int           size,
			      const char          *mimeType,
			      const int           noVersioning,
			      const int           wait,
			      ngamsSTATUS         *status)
{
    int              retCode, repDataLen, locTimeout;
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl, contDisp, mtBuf, tmpFileUri;

    ngamsLogDebug("Entering ngamsArchiveFromMem() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    if (*mimeType != '\0') sprintf(mtBuf, "&mime_type=\"%s\"", mimeType);

    /* Encode the file URI */
    strcpy(tmpFileUri, ngamsEncodeUrlVal(fileUri, 1));

    /* Perform an Archive Push */
    if (timeoutSecs == -1)
	locTimeout = -1;
    else
	locTimeout = (int)(timeoutSecs + 0.5);
    sprintf(contDisp, "attachment; filename=\"%s\"; wait=\"%d\"; "
	    "no_versioning=\"%d\"; time_out\"%d\"", tmpFileUri, wait,
	    noVersioning, locTimeout);
    if (*mimeType != '\0') strcat(tmpUrl, mtBuf);
    if ((retCode = ngamsHttpPost(host, port, ngamsUSER_AGENT, "ARCHIVE",
				 ngamsARHIVE_REQ_MT, contDisp, "",
				 bufPtr, size, &repDataRef, &repDataLen, 
				 &httpResp, httpHdr)) != ngamsSTAT_SUCCESS)
        goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsArchiveFromMem()");
    return ngamsSTAT_SUCCESS;

 errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsArchiveFromMem()/FAILURE. Status: %d",retCode);
    return retCode;
}

/**
ngamsSTAT ngamsClone(const char          *host,
                     const int           port,
                     const float         timeoutSecs,
                     const char          *fileId, 
                     const int           fileVersion,
                     const char          *diskId,
                     const char          *targetDiskId,
                     const int           wait,
                     ngamsSTATUS         *status)
Execute a CLONE command. For the exact interpretation of various
combinations of fileId, fileVersion and diskId, consult the man-page
for the NG/AMS Python module "ngamsCloneCmd", function: "clone()".

fileId:          ID of file to clone.

fileVersion:     Version of files to be taken into account for the 
                 cloning.

diskId:          Disk ID for the files to be taken into account.

targetDiskId:    ID of disk where to stored the cloned files.

Returns:         ngamsSTAT_SUCCESS or one of the error codes returned by
                 the function ngamsHttpGet()
*/
ngamsSTAT ngamsClone(const char          *host,
                     const int           port,
		     const float         timeoutSecs,
		     const char          *fileId, 
		     const int           fileVersion,
		     const char          *diskId,
		     const char          *targetDiskId,
		     const int           wait,
		     ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef; 
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl, tmpBuf;

    ngamsLogDebug("Entering ngamsClone() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?wait=\"%d\"", ngamsCMD_CLONE_STR, wait);
    if (*fileId != '\0')
        {
	sprintf(tmpBuf, "&file_id=\"%s\"", fileId);
	strcat(tmpUrl, tmpBuf);
	}
    if (fileVersion != -1)
        {
	sprintf(tmpBuf, "&file_version=%d", fileVersion);
	strcat(tmpUrl, tmpBuf);
	}
    if (*diskId != '\0')
        {
	sprintf(tmpBuf, "&disk_id=\"%s\"", diskId);
	strcat(tmpUrl, tmpBuf);
	}
    if (*targetDiskId != '\0')
        {
	sprintf(tmpBuf, "&target_disk_id=\"%s\"", targetDiskId);
	strcat(tmpUrl, tmpBuf);
	}
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsClone()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsClone()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsExit(const char          *host,
                    const int           port,
                    const float         timeoutSecs,
                    const int           wait,
                    ngamsSTATUS         *status)
Send an EXIT command to the NG/AMS Server to make it clean up and
terminate execution.

Returns:         ngamsSTAT_SUCCESS or one of the error codes returned
                 by the function ngamsHttpGet().
*/
ngamsSTAT ngamsExit(const char          *host,
		    const int           port,
		    const float         timeoutSecs,
		    const int           wait,
		    ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsExit() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?wait=\"%d\"", ngamsCMD_EXIT_STR, wait);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);
    
    ngamsLogDebug("Leaving ngamsExit()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsExit()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsLabel(const char          *host,
                     const int           port,
                     const float         timeoutSecs,
                     const char          *slotId,
                     ngamsSTATUS         *status)
Send a LABEL command to the NG/AMS Server.

slotId:    ID of slot hosting the disk for which to generate the label.

Returns:   ngamsSTAT_SUCCESS or one of the error codes returned
           by the function ngamsHttpGet().
*/
ngamsSTAT ngamsLabel(const char        *host,
		     const int         port,
		     const float       timeoutSecs,
		     const char        *slotId,
		     ngamsSTATUS       *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsLabel() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?slot_id=\"%s\"", ngamsCMD_LABEL_STR, 
	    ngamsEncodeUrlVal(slotId, 1));
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsLabel()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsLabel()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsOnline(const char          *host,
                      const int           port,
                      const float         timeoutSecs,
                      const int           wait,
                      ngamsSTATUS         *status)
Send an ONLINE command to the NG/AMS Server to bring it to Online State.

Returns:         ngamsSTAT_SUCCESS or one of the error codes returned
                 by the function ngamsHttpGet().
*/
ngamsSTAT ngamsOnline(const char          *host,
		      const int           port,
		      const float         timeoutSecs,
		      const int           wait,
		      ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsOnline() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?wait=\"%d\"", ngamsCMD_ONLINE_STR, wait);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsOnline()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsOnline()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsOffline(const char          *host,
                       const int           port,
                       const float         timeoutSecs,
                       const int           force,
                       const int           wait,
                       ngamsSTATUS         *status)

Send an OFFLINE command to the NG/AMS Server to bring it to Offline State.

force:     Force the server to go Offline immediately, even though it is busy.

Returns:   ngamsSTAT_SUCCESS or one of the error codes returned
           by the function ngamsHttpGet().
*/
ngamsSTAT ngamsOffline(const char          *host,
		       const int           port,
		       const float         timeoutSecs,
		       const int           force,
		       const int           wait,
		       ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef; 
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsOffline() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?wait=\"%d\"", ngamsCMD_OFFLINE_STR, wait);
    if (force) strcat(tmpUrl, "&force=1");
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsOffline()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status); 
    ngamsLogDebug("Leaving ngamsOffline()/FAILURE. Status: %d", retCode);
    return retCode;
}


/**
ngamsSTAT ngamsParseSrvList(const char *servers)
Parse a comma separated list of server nodes and port numbers of the form:

  "<Host 1>:<Port 1>,<Host 2>:<Port 2>,..."

The server/port pairs are kept in an internal registry to be used when 
communicating with the remote server(s).

servers:   List of servers/ports.

Returns:   ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/  
ngamsSTAT ngamsParseSrvList(const char *servers)
{
    char           *nextTok, *strPtr;
    int            port;
    ngamsMED_BUF   host;

    ngamsLogDebug("Entering ngamsParseSrvList() ...");
    memset(_ngamsSrvInfo(), 0, sizeof(ngamsSRV_INFO));
    _ngamsSrvInfo()->srvIdx = -1;
    nextTok = strtok_r((char *)servers, ",", &strPtr);
    while (nextTok != NULL)
	{
	if (ngamsSplitSrvAddr(nextTok, host, &port) != ngamsSTAT_SUCCESS)
	    {  
	    ngamsLogDebug("Leaving ngamsParseSrvList()/FAILURE");
	    return ngamsSTAT_FAILURE;
	    }
	strcpy(_ngamsSrvInfo()->hosts[_ngamsSrvInfo()->numberOfSrvs], host);
	_ngamsSrvInfo()->ports[_ngamsSrvInfo()->numberOfSrvs] = port;

	_ngamsSrvInfo()->numberOfSrvs += 1;
	nextTok = strtok_r(NULL, ",", &strPtr);
	}
    ngamsLogDebug("Leaving ngamsParseSrvList()");
    return ngamsSTAT_SUCCESS;
}


/**
ngamsSTAT ngamsRegister(const char          *host,
                        const int           port,
                        const float         timeoutSecs,
                        const char          *path,
                        const int           wait,
                        ngamsSTATUS         *status)
Send an REGISTER command to the NG/AMS Server to make it register
a file or set of files.

path:      Path starting point from which the scanning for files to
           register will be initiated. Only files with a known mime-type
           is taken into account.

Returns:   ngamsSTAT_SUCCESS or one of the error codes returned
           by the function ngamsHttpGet().           
*/  
ngamsSTAT ngamsRegister(const char          *host,
			const int           port,
			const float         timeoutSecs,
			const char          *path,
			const int           wait,
			ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef; 
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsRegister() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?wait=\"%d\"&path=\"%s\"", ngamsCMD_REGISTER_STR, 
	    wait, path);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsRegister()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status); 
    ngamsLogDebug("Leaving ngamsRegister()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsRemDisk(const char        *host,
                       const int         port,
                       const float       timeoutSecs,
                       const char        *diskId,
                       const int         execute,
                       ngamsSTATUS       *status)
Send a REMDISK command to the NG/AMS Server. If execute is 0 the 
disk information will not be deleted from the NGAS DB and from the
the disk itself. Otherwise, if 1 is specified, this information will
will be deleted.

diskId:      ID of disk to remove.

execute:     If set to 1 the command will be executed and the disk
             removed from the system (if possible). Otherwise a report
             will be send back indicating if it is possible to
             remove the disk referenced.

Returns:     ngamsSTAT_SUCCESS or one of the error codes returned
             by the function ngamsHttpGet().
*/
ngamsSTAT ngamsRemDisk(const char        *host,
		       const int         port,
		       const float       timeoutSecs,
		       const char        *diskId,
		       const int         execute,
		       ngamsSTATUS       *status)
{
    int              retCode, repDataLen;
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;

    ngamsLogDebug("Entering ngamsRemDisk() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s?disk_id=\"%s\"&execute=%d", ngamsCMD_REMDISK_STR, 
	    ngamsEncodeUrlVal(diskId, 1), execute);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsRemDisk()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsRemDisk()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsRemFile(const char        *host,
                       const int         port,
                       const float       timeoutSecs,
                       const char        *diskId,
                       const char        *fileId,
                       const int         fileVersion,
                       const int         execute,
                      ngamsSTATUS       *status)
Send a REMFILE command to the NG/AMS Server. If execute is 0 the 
disk information will not be deleted from the NGAS DB and from the
the disk itself. Otherwise, if 1 is specified, this information will
will be deleted. For the interpretation of various combinations of
the parameters diskId, fileId, fileVersion and execute consult the
man-page of the Python module "ngamsRemoveCmd", function remFile().

diskId:        ID of disk hosting the file(s) to be removed.

fileId:        ID of file(s) to be removed.

fileVersion:   Version of file(s) to be removed.

execute:       If set to 1 the files will be removed (if possible),
               otherwise a report will be send back indicating what
               would be removed if the command is executed.

Returns:       ngamsSTAT_SUCCESS or one of the error codes returned
               by the function ngamsHttpGet().
*/
ngamsSTAT ngamsRemFile(const char        *host,
		       const int         port,
		       const float       timeoutSecs,
		       const char        *diskId,
		       const char        *fileId,
		       const int         fileVersion,
		       const int         execute,
		       ngamsSTATUS       *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl, fileIdTmp;

    ngamsLogDebug("Entering ngamsRemFile() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    strcpy(fileIdTmp, ngamsEncodeUrlVal(fileId, 1));
    sprintf(tmpUrl, "%s?disk_id=\"%s\"&file_id=\"%s\"&file_version=%d"
	    "&execute=%d", ngamsCMD_REMFILE_STR, ngamsEncodeUrlVal(diskId, 1),
	    fileIdTmp, fileVersion, execute);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsRemFile()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status); 
    ngamsLogDebug("Leaving ngamsRemFile()/FAILURE. Status: %d", retCode);
    return retCode;
}

ngamsSTAT _ngamsRetrieve2File(const char          *host,
			      const int           port,
			      const float         timeoutSecs,
			      const char          *fileRef,
			      const int           fileVersion,
			      const char          *processing,
			      const char          *processingPars,
			      const char          *targetFile,
			      ngamsSTATUS         *status,
			      const int           internal,
			      const char          *hostId)
{	
    char             tmpBuf[10000], *tmpP;
    int              retCode, repDataLen;
    int              fd = 0, bytesRd, bytesRead = 0;
    time_t           timeOut, timeLastRec;
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsBIG_BUF     tmpUrl;
    ngamsMED_BUF     finalTargetFile, tmpTargetFilename;

    ngamsLogDebug("Entering _ngamsRetrieve2File() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    memset(&repDataRef, 0, sizeof(repDataRef));

    /* Perform the query */
    if (strcmp(fileRef, ngamsNG_LOG_REF) == 0)
        sprintf(tmpUrl, "%s?ng_log", ngamsCMD_RETRIEVE_STR);
    else if (strcmp(fileRef, ngamsCFG_REF) == 0)
        sprintf(tmpUrl, "%s?cfg", ngamsCMD_RETRIEVE_STR);
    else if (internal)
	sprintf(tmpUrl, "%s?internal=\"%s\"", 
		ngamsCMD_RETRIEVE_STR, ngamsEncodeUrlVal(fileRef, 1));
    else
        sprintf(tmpUrl, "%s?file_id=\"%s\"", ngamsCMD_RETRIEVE_STR, 
		ngamsEncodeUrlVal(fileRef, 1));
    if ((processing != NULL) && (*processing != '\0'))
        {
	sprintf(tmpBuf, "&processing=\"%s\"", ngamsEncodeUrlVal(processing,1));
	strcat(tmpUrl, tmpBuf);
	if ((processingPars != NULL) && (*processingPars != '\0'))
	    {
	    sprintf(tmpBuf, "&processing_pars=\"%s\"", 
		    ngamsEncodeUrlVal(processingPars, 1));
	    strcat(tmpUrl, tmpBuf);
	    }
	}

    if (fileVersion != -1)
	{
	sprintf(tmpBuf, "&file_version=\"%d\"", fileVersion);
	strcat(tmpUrl, tmpBuf);
	}

    if (*hostId != '\0') 
	{
	sprintf(tmpBuf, "&host_id=%s", hostId);
	strcat(tmpUrl, tmpBuf);
	}

    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl,
				0, &repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;

    /* If the specified filename is a file, store the data into a file
     * with that name. Othwerwise if a directory is specified, receive
     * the data into a file in that directory, with the name indicated
     * in the HTTP header.
     */
    if (ngamsIsDir(targetFile))
	{
	if ((retCode = ngamsGetHttpHdrEntry(httpHdr, "content-disposition", 
					    "filename", tmpTargetFilename)) 
	    != ngamsSTAT_SUCCESS)
	    goto errExit;
	strcpy(finalTargetFile, targetFile);
	if (*(finalTargetFile + strlen(finalTargetFile) - 1) != '/')
	    strcat(finalTargetFile, "/");
	strcat(finalTargetFile, tmpTargetFilename);
	}
    else if (*targetFile != '\0')
	strcpy(finalTargetFile, targetFile);
    else
	{
	if ((retCode = ngamsGetHttpHdrEntry(httpHdr, "content-disposition", 
					    "filename", tmpTargetFilename)) 
	    != ngamsSTAT_SUCCESS)
	    goto errExit;
	if ((tmpP = getenv("PWD")) != NULL)
	    sprintf(finalTargetFile, "%s/", tmpP);
	else
	    *finalTargetFile = '\0';
	strcat(finalTargetFile, tmpTargetFilename);
	}

    /* Receive the data */
    if ((fd = creat(finalTargetFile, 0664)) == -1)
	{
	retCode = ngamsERR_INV_TARG_FILE;	
	goto errExit;
	}
    timeOut = _getTimeOut();
    timeLastRec = time(NULL);
    while (bytesRead < repDataLen)
	{ 
	if ((retCode = _pollFd(repDataRef.fd, 
			       _getTimeOut())) != ngamsSTAT_SUCCESS)
	    goto errExit;
	bytesRd = read(repDataRef.fd, tmpBuf, 10000);
	if (bytesRd > 0)
	    {
	    write(fd, tmpBuf, bytesRd);
	    bytesRead += bytesRd;
	    timeLastRec = time(NULL);
	    }
	else
	    {
	    if ((time(NULL) - timeLastRec) >= timeOut)
		{
		retCode = ngamsERR_TIMEOUT;
		goto errExit;
		}
	    }
	}
    close(fd);

    /* Query went OK - for the moment no XML status document from RETRIEVE */
    strcpy(status->message, "Successfully handled RETRIEVE command");
    strcpy(status->status, "OK");
    status->errorCode = ngamsSTAT_SUCCESS;

    if (repDataRef.fd > 0) close(repDataRef.fd);

    ngamsLogDebug("Leaving _ngamsRetrieve2File()");
    return ngamsSTAT_SUCCESS;

  errExit:
    /* Flush the connection for data waiting to be picked up. */
    if ((retCode != ngamsSRV_INV_QUERY) && (retCode != ngamsERR_TIMEOUT) && 
	(repDataRef.pdata == NULL))
	{
	while (bytesRead < repDataLen)
	    { 
	    bytesRd = read(repDataRef.fd, tmpBuf, 10000);
	    bytesRead += bytesRd;
	    }
	}
    if (repDataRef.pdata)
	ngamsUnpackStatus(repDataRef.pdata, status);
    else
	{
	ngamsStat2Str(retCode, status->message);
	strcpy(status->status, "FAILURE");
	status->errorCode = retCode;
	}
    ngamsLogDebug("Leaving _ngamsRetrieve2File()/FAILURE. Status: %d",retCode);
    return retCode;
}

/**
ngamsSTAT ngamsRetrieve2File(const char          *host,
                             const int           port,
                             const float         timeoutSecs,
                             const char          *fileId,
                             const int           fileVersion,
                             const char          *processing,
                             const char          *processingPars,
                             const char          *targetFile,
                             ngamsSTATUS         *status);

Send a RETRIEVE command to the NG/AMS Server to retrieve a
data file, and store this in a file on the local disk.

fileId:           ID of the file to retrieve.   

fileVersion:      Specific version of file to retrieve. If set to -1 the
                  latest version will be retrieved.

processing:       Name of DPPI to be invoked by NG/AMS when data is 
                  retrieved.

processingPars:   Optional list of parameters to hand over to the DPPI.

targetFile:       If a valid filename is specified the data retrieved
                  will be stored in a file with that name. If a directory
                  is given, the data file retrieved will be stored in that
                  directory with the name under which it is stored in 
                  NGAS. If this parameter is an empty string, it will be 
                  tried to stored the file retrieved under the NGAS 
                  archive name in the current working directory.

Returns:          ngamsSTAT_SUCCESS or one of the error codes returned
                  by the function ngamsHttpGet().
*/
ngamsSTAT ngamsRetrieve2File(const char          *host,
			     const int           port,
			     const float         timeoutSecs,
			     const char          *fileId,
			     const int           fileVersion,
			     const char          *processing,
			     const char          *processingPars,
			     const char          *targetFile,
			     ngamsSTATUS         *status)
{
    int stat;

    ngamsLogDebug("Entering ngamsRetrieve2File() ...");    
    stat = _ngamsRetrieve2File(host, port, timeoutSecs, fileId, fileVersion,
			       processing, processingPars, targetFile, 
			       status, 0, "");
    ngamsLogDebug("Leaving ngamsRetrieve2File()");    
    return stat;
}


ngamsSTAT ngamsGenRetrieve2File(const char            *host,
				const int             port,
				const float           timeoutSecs,
				const ngamsCMD        cmdCode,
				const ngamsPAR_ARRAY  *parArray,
				char                  *targetFile,
				ngamsSTATUS           *status)
{	
    char             tmpBuf[10000], *tmpP;
    int              retCode, repDataLen, i;
    int              fd = 0, bytesRd, bytesRead = 0;
    time_t           timeOut, timeLastRec;
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsBIG_BUF     url;
    ngamsMED_BUF     finalTargetFile, tmpTargetFilename, cmd;
    
    ngamsLogDebug("Entering ngamsGenRetrieve2File() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    memset(&repDataRef, 0, sizeof(repDataRef));
    ngamsCmd2Str(cmdCode, cmd);
    strcpy(url, cmd);
    for (i = 0; i < parArray->idx; i++)
	{
	if (i == 0)
	    sprintf(tmpBuf, "?%s=\"%s\"", parArray->parArray[i], 
		    ngamsEncodeUrlVal(parArray->valArray[i], 1));
	else
	    sprintf(tmpBuf, "&%s=\"%s\"", parArray->parArray[i], 
		    ngamsEncodeUrlVal(parArray->valArray[i], 1));
	strcat(url, tmpBuf);
	}
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, url,
				0, &repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;

    /* If the specified filename is a file, store the data into a file with 
     * that name. Othwerwise if a directory is specified, receive the data into
     * a file in that directory, with the name indicated in the HTTP header.
     */
    if (ngamsIsDir(targetFile))
	{
	if ((retCode = ngamsGetHttpHdrEntry(httpHdr, "content-disposition", 
					    "filename", tmpTargetFilename)) 
	    != ngamsSTAT_SUCCESS)
	    goto errExit;
	strcpy(finalTargetFile, targetFile);
	if (*(finalTargetFile + strlen(finalTargetFile) - 1) != '/')
	    strcat(finalTargetFile, "/");
	strcat(finalTargetFile, tmpTargetFilename);
	}
    else if (*targetFile != '\0')
	strcpy(finalTargetFile, targetFile);
    else
	{
	if ((retCode = ngamsGetHttpHdrEntry(httpHdr, "content-disposition", 
					    "filename", tmpTargetFilename)) 
	    != ngamsSTAT_SUCCESS)
	    goto errExit;
	if ((tmpP = getenv("PWD")) != NULL)
	    sprintf(finalTargetFile, "%s/", tmpP);
	else
	    *finalTargetFile = '\0';
	strcat(finalTargetFile, tmpTargetFilename);
	}

    /* Receive the data */
    if ((fd = creat(finalTargetFile, 0664)) == -1)
	{
	retCode = ngamsERR_INV_TARG_FILE;	
	goto errExit;
	}
    timeOut = _getTimeOut();
    timeLastRec = time(NULL);
    while (bytesRead < repDataLen)
	{ 
	if ((retCode = _pollFd(repDataRef.fd, 
			       _getTimeOut())) != ngamsSTAT_SUCCESS)
	    goto errExit;
	bytesRd = read(repDataRef.fd, tmpBuf, 10000);
	if (bytesRd > 0)
	    {
	    write(fd, tmpBuf, bytesRd);
	    bytesRead += bytesRd;
	    timeLastRec = time(NULL);
	    }
	else
	    {
	    if ((time(NULL) - timeLastRec) >= timeOut)
		{
		retCode = ngamsERR_TIMEOUT;
		goto errExit;
		}
	    }
	}
    close(fd);

    /* Query went OK - for the moment no XML status document from RETRIEVE */
    strcpy(status->message, "Successfully handled RETRIEVE command");
    strcpy(status->status, "OK");
    status->errorCode = ngamsSTAT_SUCCESS;

    if (repDataRef.fd > 0) close(repDataRef.fd);

    ngamsLogDebug("Leaving ngamsGenRetrieve2File()");
    return ngamsSTAT_SUCCESS;

  errExit:
    /* Flush the connection for data waiting to be picked up. */
    if ((retCode != ngamsSRV_INV_QUERY) && (retCode != ngamsERR_TIMEOUT) && 
	(repDataRef.pdata == NULL))
	{
	while (bytesRead < repDataLen)
	    { 
	    bytesRd = read(repDataRef.fd, tmpBuf, 10000);
	    bytesRead += bytesRd;
	    }
	}
    if (repDataRef.pdata)
	ngamsUnpackStatus(repDataRef.pdata, status);
    else
	{
	ngamsStat2Str(retCode, status->message);
	strcpy(status->status, "FAILURE");
	status->errorCode = retCode;
	}
    ngamsLogDebug("Leaving ngamsGenRetrieve2File()/FAILURE. Status: %d",
		  retCode);
    return retCode;
}

/**
ngamsSTAT ngamsStatus(const char          *host,
                      const int           port,
                      const float         timeoutSecs,
                      ngamsSTATUS         *status)
Send a STATUS command to NG/AMS to query the current status of the
system. No parameters are defined at present.

Returns:      ngamsSTAT_SUCCESS or one of the error codes returned
              by the function ngamsHttpGet().
*/
ngamsSTAT ngamsStatus(const char          *host,
		      const int           port,
		      const float         timeoutSecs,
		      ngamsSTATUS         *status)	  
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef; 
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl;
			 
    ngamsLogDebug("Entering ngamsStatus() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(tmpUrl, "%s", ngamsCMD_STATUS_STR);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, tmpUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsStatus()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsStatus()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
ngamsSTAT ngamsSubscribe(const char          *host,
                         const int           port,
                         const float         timeoutSecs,
                         const char          *url,
                         const int           priority,
                         const char          *startDate,
                         const char          *filterPlugIn,
                         const char          *filterPlugInPars,
                         ngamsSTATUS         *status)
Send a SUBSCRIBE to NG/AMS to subscribe for data or a specific type of data.

url:                 Subscriber URL to where data is pushed.

priority:            Priority of the Subscriber (low number = high 
                     priority). Default value 10.

startDate:           Start date defining which data the subscription
                     should take into account.

filterPlugIn:        Optional Filter Plug-In to apply when selecting
                     which data files to deliver to the clients.

filterPlugInPars:    Optional Filter Plug-In Parameters to transfer
                     to the Filter Plug-In.

Returns:             ngamsSTAT_SUCCESS or one of the error codes returned
                     by the function ngamsHttpGet().
*/
ngamsSTAT ngamsSubscribe(const char          *host,
			 const int           port,
			 const float         timeoutSecs,
			 const char          *url,
			 const int           priority,
			 const char          *startDate,
			 const char          *filterPlugIn,
			 const char          *filterPlugInPars,
			 ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl, reqUrl;
    
    ngamsLogDebug("Entering ngamsSubscribe() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(reqUrl, "%s?url=\"%s\"&priority=%d", ngamsCMD_SUBSCRIBE_STR,
	    ngamsEncodeUrlVal(url, 1), priority);
    if ((startDate != NULL) && (*startDate != '\0'))
        {
	sprintf(tmpUrl, "&start_date=\"%s\"", ngamsEncodeUrlVal(startDate, 1));
	strcpy(reqUrl, tmpUrl);
	}
    if ((filterPlugIn != NULL) && (*filterPlugIn != '\0'))
        {
	sprintf(tmpUrl, "&filter_plug_in=\"%s\"", 
		ngamsEncodeUrlVal(filterPlugIn, 1));
	strcpy(reqUrl, tmpUrl);
	}
    if ((filterPlugInPars != NULL) && (*filterPlugInPars != '\0'))
        {
	sprintf(tmpUrl, "&plug_in_pars=\"%s\"",
		ngamsEncodeUrlVal(filterPlugInPars, 1));
	strcpy(reqUrl, tmpUrl);
	}
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, reqUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsSubscribe()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsSubscribe()/FAILURE. Status: %d", retCode);
    return retCode;
  
}

/**
ngamsSTAT ngamsUnsubscribe(const char          *host,
			   const int           port,
			   const float         timeoutSecs,
			   const char          *url,
			   ngamsSTATUS         *status)
Send an UNSUBSCRIBE Command to the given NG/AMS Server.

Returns:    ngamsSTAT_SUCCESS or one of the error codes returned
            by the function ngamsHttpGet().
*/
ngamsSTAT ngamsUnsubscribe(const char          *host,
			   const int           port,
			   const float         timeoutSecs,
			   const char          *url,
			   ngamsSTATUS         *status)
{
    int              retCode, repDataLen; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     reqUrl;

    ngamsLogDebug("Entering ngamsUnsubscribe() ...");
    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);
    sprintf(reqUrl, "%s?url=\"%s\"", ngamsCMD_UNSUBSCRIBE_STR,
	    ngamsEncodeUrlVal(url, 1));
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, reqUrl, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsUnsubscribe()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsUnsubscribe()/FAILURE. Status: %d", retCode);
    return retCode;
  
}

/**
*******************************************************************************
*/

/**
*******************************************************************************
The following functions are used to handle the HTTP communication.
These could be moved to a separate module for reusage.
*******************************************************************************
*/

#define	ngamsMAXLINE    16384
#define	ngamsBUFSIZE    65536

/**
ngamsSTAT ngamsGetHttpHdrEntry(ngamsHTTP_HDR   httpHdr,
			       const char      *hdrName,
			       const char      *fieldName,
			       ngamsMED_BUF    value)
Get entry from HTTP headers.

httpHdr:      Instance of ngamsHTTP_HDR structure containing the HTTP
              headers.

hdrName:      Name of the HTTP header entry to consider.

fieldName:    Fieldname within the HTTP header. If this is not relevant,
              this should be specified as '\0'.

value:        Value of the HTTP header or field.

Returns:      Execution status (ngamsSTAT_SUCCESS|ngamsSTAT_FAILURE).
*/
ngamsSTAT ngamsGetHttpHdrEntry(ngamsHTTP_HDR   httpHdr,
			       const char      *hdrName,
			       const char      *fieldName,
			       ngamsMED_BUF    value)
{
    char           *fieldPtr, *valPtr;
    int            found = 0, idx = 0, stat;
    ngamsMED_BUF   fieldLoc;

    ngamsLogDebug("Entering ngamsGetHttpHdrEntry() ...");
    sprintf(fieldLoc, "%s=", fieldName);
    
    /* An HTTP header could be something like:
     * Content-disposition: attachment; 
     * filename="ISAAC.1999-04-11T05:03:21.035.fits.Z"
     */

    /* Find the right HTTP header */
    while (!found && (idx < ngamsHTTP_MAX_HDRS))
	{
	if (strstr(httpHdr[idx], hdrName) != NULL)
	    found = 1;
	else
	    idx++;
	}
    
    /* Find the field - if fieldName was set to "" we just take
     * the whole value of the header.
     */
    if (*fieldName != '\0')
	{
	if ((fieldPtr = strstr(httpHdr[idx], fieldLoc)) == NULL)
	    {
	    stat = ngamsERR_INV_REPLY;
	    goto errExit;
	    }
	valPtr = (fieldPtr + strlen(fieldLoc)); 
	}
    else
	{
	valPtr = strchr(httpHdr[idx], ':');
	valPtr++;
	while (*valPtr == ' ') valPtr++;
	}
    
    /* Get the value */
    memset(value, 0, sizeof(ngamsMED_BUF));
    if (*valPtr == '"') valPtr++;
    while ((*valPtr != '"') && (*valPtr != '\0') && (*valPtr != '\r') &&
	   (*valPtr != '\n'))
	strncat(value, valPtr++, sizeof(char));

    ngamsLogDebug("Leaving ngamsGetHttpHdrEntry()");
    return ngamsSTAT_SUCCESS;

 errExit:
    ngamsLogDebug("Leaving ngamsGetHttpHdrEntry()/FAILURE. Status: %d", stat);
    return stat;
}


/**
int ngamsReadLine(int	*fd,
		  char	*ptr,
		  int	maxlen)
Read a line from a HTTP (socket) connection

fd:        File descriptor for HTTP socket connection.

ptr:       Pointer to string buffer in which the line will be written.

maxlen:    Maximum number of bytes to read.

Returns:   Bytes read.
*/
int ngamsReadLine(int	*fd,
		  char	*ptr,
		  int	maxlen)
{
    char    c;
    int	    n, rc;

    for (n = 1; n < maxlen; n++) 
	{
	/* IMPL: If the remote server writes a line, which does not contain
	 *       a newline character, read() blocks. Could maybe be improved
	 *       by probing (using select with timeout defined) before
	 *       trying to read from the socket.
	 */
	if ((rc = read(*fd, &c, 1)) == 1) 
	    {
	    *ptr++ = c;
	    if (c == '\n')
		break;
	    } 
	else if (rc == 0) 
	    {
	    if (n == 1)
		return 0;	/* EOF, no data read */
	    else
		break;		/* EOF, some data was read */
	    } 
	else
	    return -1;	/* error */
	}
    
    *ptr = 0;
    return n;
}

/**
int ngamsRecvData(char    **data,
		  int     dataLen,
		  int     *sockFd)
Receive data of a given size from a socket connection.

data:      Pointer pointer to string buffer, which will be allocated by 
           this function and which will contain the data read.

dataLen:   Length of data read.

sockFd:    Socket file descriptor to read from.

Returns:   Execution status (ngamsSTAT_SUCCESS|ngamsERR_RD_DATA|
                             ngamsERR_ALLOC_MEM).
*/
int ngamsRecvData(char    **data,
		  int     dataLen,
		  int     *sockFd)
{
    int attempt = 0, dataRead = 0, dataRem, stat = ngamsSTAT_SUCCESS;

    ngamsLogDebug("Entering ngamsRecvData() ...");
    if (!(*data = malloc(dataLen + 1)))
	{
	stat = ngamsERR_ALLOC_MEM;
	goto errExit;
	}
    while (1)
	{
	attempt += 1;
	dataRem = (dataLen - dataRead);
	dataRead += read(*sockFd, (*data + dataRead), dataRem);
	if (dataRead < dataLen)
	    {
	    if (attempt <= 10)
		_pollFd(*sockFd, _getTimeOut());
	    else
		{
		stat = ngamsERR_RD_DATA;
		break;
		}
	    }
	else
	    break;
	}
    *(*data + dataRead) = '\0';
    close(*sockFd);
    *sockFd = 0;

    ngamsLogDebug("Leaving ngamsRecvData(). Status: %d", stat);
    return stat;

 errExit:
    ngamsLogDebug("Leaving ngamsRecvData()/FAILURE. Status: %d", stat);
    return stat;
}

/* The _ngamsConnect() function (and _connectThread() + _ngamsTHREAD_INFO) 
 * are used as a work-around for a problem encountered whereby the connect() 
 * socket system calls blocks an arbitrarily long period of time if the 
 * specified target host is defined in the DNS domain but is not accessible.
 *  
 * The actual connect() is executed in a sub-thread, monitored by the calling
 * thread, which does not block and which can take the necessary actions if
 * connect() blocks.
 */
typedef struct {
    struct sockaddr  *servAddr;
    int              sockFd;
    int              syncPipe[2];
} _ngamsTHREAD_INFO;


void *_connectThread(void *ptr)
{
    int                stat, connected = 0;;
    _ngamsTHREAD_INFO  *thrInfo;

    ngamsLogDebug("Entering _connectThread() ...");
    thrInfo = (_ngamsTHREAD_INFO *)ptr;
    stat = connect(thrInfo->sockFd, thrInfo->servAddr, 
		   sizeof(struct sockaddr_in));
    if (stat >= 0) connected = 1; 
    write(thrInfo->syncPipe[1], &connected, sizeof(int));
    pthread_exit(NULL);
    ngamsLogDebug("Leaving _connectThread()");
    return NULL;
}


ngamsSTAT _ngamsConnect(const int                 sockFd,  
			const struct sockaddr_in  *servAddr)
{
    int                  stat, creStat = -1, retStat, connected = 0;
    pthread_t            thrHandle;
    pthread_attr_t       thrAttr;
    _ngamsTHREAD_INFO    thrInfo;

    ngamsLogDebug("Entering _ngamsConnect() ...");
    memset(&thrInfo, 0, sizeof(_ngamsTHREAD_INFO));
    thrInfo.servAddr = (struct sockaddr *)servAddr;
    thrInfo.sockFd = sockFd;
    if (pipe(thrInfo.syncPipe) != 0)
	{
	retStat = ngamsSTAT_FAILURE;
	goto fctExit;
	}
    pthread_attr_init(&thrAttr);
    pthread_attr_setdetachstate(&thrAttr, PTHREAD_CREATE_DETACHED);
    creStat = pthread_create(&thrHandle, &thrAttr, _connectThread, &thrInfo);
    if (creStat != 0) 
	{
	retStat = ngamsSTAT_FAILURE;
	goto fctExit;
	}
    stat = _pollFd(thrInfo.syncPipe[0], 0.5);
    if (stat == ngamsSTAT_SUCCESS)
	read(thrInfo.syncPipe[0], &connected, sizeof(int));
    if (connected)
	retStat = ngamsSTAT_SUCCESS;
    else
	retStat = ngamsSTAT_FAILURE;
    goto fctExit;
	    
 fctExit:
    if (thrInfo.syncPipe[0])
	{
	close(thrInfo.syncPipe[0]);
	close(thrInfo.syncPipe[1]);
	}
    if (!connected && (creStat == 0)) pthread_cancel(thrHandle);
    pthread_attr_destroy(&thrAttr);
    if (retStat == ngamsSTAT_SUCCESS)
	ngamsLogDebug("Leaving _ngamsConnect()");
    else
	ngamsLogDebug("Leaving _ngamsConnect()/FAILURE. Status: %d", retStat);
    return retStat;
}


/**
int ngamsPrepSock(const char      *host,
		  const int       port)
Prepare a socket connection.

host:     Host to which to build up socket connection.

port:     Port number of listen socket.

Returns:  Socket file descriptor in case a socket connection was 
          successfully created. Otherwise one of the following errors
          might be returned:

            ngamsERR_HOST, ngamsERR_SOCK, ngamsERR_CON
 */
int ngamsPrepSock(const char      *host,
		  const int       port)
{
    int                    sockFd = 0, stat;
    struct hostent         *hostRef;
    struct sockaddr_in     servAddr;

    ngamsLogDebug("Entering ngamsPrepSock() ...");

    if (!(hostRef = gethostbyname(host)))
	{
	stat = ngamsERR_HOST;
	goto errExit;
	}

    /* Fill in the structure "servAddr" with the address of the
     * server that we want to connect with.
     */
    memset((char *)&servAddr, '\0', sizeof(servAddr));
    memcpy(&servAddr.sin_addr, hostRef->h_addr_list[0], hostRef->h_length);
    servAddr.sin_family = AF_INET;
    servAddr.sin_port   = htons(port);
     
    /* Open a TCP socket (an Internet stream socket).
     */
    if ((sockFd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	stat = ngamsERR_SOCK;
	goto errExit;
	}
    stat = _ngamsConnect(sockFd, &servAddr);
    if (stat == ngamsSTAT_FAILURE) 
	{
	if (sockFd) close(sockFd);
	stat = ngamsERR_CON;
	goto errExit;
	}

    ngamsLogDebug("Leaving ngamsPrepSock()");
    return sockFd;

 errExit:
    ngamsLogDebug("Leaving ngamsPrepSock()/FAILURE. Status: %d", stat);
    return stat;
}

/**
int ngamsRecvHttpHdr(int             *sockFd,
		     ngamsHTTP_HDR   httpHdr,
		     ngamsHTTP_RESP  *httpResp,
		     ngamsHTTP_DATA  *repDataRef,
		     int             *dataLen)
Receive HTTP headers from the given HTTP socket file descriptor and
store them in the instance of the ngamsHTTP_HDR structure.

sockFd:        Socket file descriptor.

httpHdr:       Instance of ngamsHTTP_HDR structure.

httpResp:      Contains the decoded HTTP response, which is of the form:

                 "HTTP/1.0 200 OK"

repDataRef:    In case of an error this is used to retrieve the
               XML status document from the NG/AMS Server. This is 
               possibly a minor violation of the HTTP protocol, but
               practicle for displaying a better error message for the
               user. However, this should not make the HTTP protocol 
               implemented incompatible with the definition of the HTTP
               protocol.

dataLen:       Length of the data (XML Status Document) possibly received.

Returns:       Execution status (ngamsSTAT_SUCCESS|ngamsERR_TIMEOUT|
               ngamsERR_INV_REPLY|ngamsSRV_INV_QUERY).
*/
int ngamsRecvHttpHdr(int             *sockFd,
		     ngamsHTTP_HDR   httpHdr,
		     ngamsHTTP_RESP  *httpResp,
		     ngamsHTTP_DATA  *repDataRef,
		     int             *dataLen)
{
    char     recvLine[ngamsMAXLINE + 1], *strP1, *strP2;
    char     *tmpData, *chr;
    int      bytesRead, hdrIdx = 0;
    int      retCode = 0;

    ngamsLogDebug("Entering ngamsRecvHttpHdr() ...");

    tmpData = NULL;
    memset(httpResp, 0, sizeof(ngamsHTTP_RESP));
    
    /* Wait for the reply from the server.
     */
    if ((retCode = _pollFd(*sockFd, _getTimeOut())) != ngamsSTAT_SUCCESS)
	goto errExit;

    /* Read the HTTP header: ends with a blank line.
     */
    while ((bytesRead = ngamsReadLine(sockFd, recvLine, ngamsMAXLINE)) > 0)
	{
	if ((bytesRead == 1) || 
	    (strstr(recvLine, "\015\012") == recvLine)) break;
	for ((chr = recvLine); (*chr && (*chr != ':')); chr++) 
	    *chr = (char)tolower((int)*chr);
	strcpy(httpHdr[hdrIdx], recvLine);
	hdrIdx++;
	if (strstr(recvLine, "content-length:")) 
	    *dataLen = atoi(&recvLine[15]);
	}

    /* In case -1 is returned by ngamsReadLine() a problem occurred
     * with the communication.
     */
    if (bytesRead < 0)
	{
	retCode = ngamsERR_COM;
	goto errExit;
	}
    
    /* Extract the info from the first line of the HTTP response. 
     * Is something like: "HTTP/1.0 200 OK".
     */
    strP1 = httpHdr[0];
    strP2 = strchr(strP1, ' ');
    if ((strP2 - strP1) >= sizeof(ngamsSMALL_BUF))
	{
	retCode = ngamsERR_INV_REPLY;
	goto errExit;
	}
    strncpy(httpResp->version, strP1, (strP2 - strP1));
    strP1 = (strP2 + 1);
    httpResp->status = atoi(strP1);
    strP2 = strchr(strP1, ' ');
    strcpy(httpResp->message, (strP2 + 1));
    ngamsTrimString(httpResp->message, "\015\012");

    /* If the request was invalid, we try to receive possible data 
     * contained in the HTTP reply.
     */
    if (httpResp->status == ngamsSRV_INV_QUERY)
	{
	if (ngamsRecvData(&tmpData, *dataLen, sockFd) == ngamsSTAT_SUCCESS)
	    {
	    repDataRef->pdata = calloc((*dataLen + 1), sizeof(char));
	    memcpy(repDataRef->pdata, tmpData, *dataLen);
	    free(tmpData);
	    tmpData = NULL;
	    }
	retCode = httpResp->status;
	goto errExit;
	}
    else if (httpResp->status == ngamsSRV_REDIRECT)
	{
	if (ngamsRecvData(&tmpData, *dataLen, sockFd) == ngamsSTAT_SUCCESS)
	    free(tmpData);
	}

    ngamsLogDebug("Leaving ngamsRecvHttpHdr()");
    return ngamsSTAT_SUCCESS;

  errExit:
    if (tmpData) free(tmpData);
    ngamsLogDebug("Leaving ngamsRecvHttpHdr()/FAILURE. Status: %d", retCode); 
    return retCode;
}

/**
int _ngamsHttpGet(const char      *host,
	 	  const int       port,
		  const char      *userAgent,
		  const char      *path,
		  const int       receiveData,
		  ngamsHTTP_DATA  *repDataRef,
		  int             *dataLen,
		  ngamsHTTP_RESP  *httpResp,
		  ngamsHTTP_HDR   httpHdr)
Execute an HTTP GET.

See ngamsHttpGet() for parameters and return value.
*/
int _ngamsHttpGet(const char      *host,
		  const int       port,
		  const char      *userAgent,
		  const char      *path,
		  const int       receiveData,
		  ngamsHTTP_DATA  *repDataRef,
		  int             *dataLen,
		  ngamsHTTP_RESP  *httpResp,
		  ngamsHTTP_HDR   httpHdr)
{
    char              *strP;
    int               sockFd = 0, n;
    int               retCode = 0;
    ngamsMED_BUF      redirectUrl, altHost, altPort;
    ngamsHUGE_BUF     sendLine, authHdr;

    ngamsLogDebug("Entering _ngamsHttpGet() ...");
    
    *dataLen = 0;
    memset(repDataRef, 0, sizeof(ngamsHTTP_DATA));
    memset(httpResp, 0, sizeof(ngamsHTTP_RESP));

    if ((sockFd = ngamsPrepSock(host, port)) < 0)
	{
	retCode = sockFd;
	goto errExit;
	}

    /* Send the GET method */
    if (ngamsGetAuthorization())
	{
	memset(authHdr, 0, sizeof(ngamsHUGE_BUF));
	sprintf(authHdr, "\015\012Authorization: Basic%%20%s", 
		ngamsGetAuthorization());
	}
    else
	*authHdr = '\0';
    sprintf(sendLine, "GET %s HTTP/1.0\nUser-Agent: %s%s\015\012\012", 
	    path, userAgent, authHdr);
    ngamsLogDebug("Submitting HTTP header: %s to host/port: %s/%d", 
		  sendLine, host, port);
    n = strlen(sendLine);
    if (write(sockFd, sendLine, n) != n) 
	{
	retCode = ngamsERR_WR_HD;
	goto errExit;
	}

    /* Receive the reply for the request */
    if ((retCode = ngamsRecvHttpHdr(&sockFd, httpHdr, httpResp, repDataRef,
				    dataLen)) != ngamsSTAT_SUCCESS)
        goto errExit;

    /* If an HTTP Redirection Response has been returned, resend the 
     * query to the alternative URL.
     */
    if (httpResp->status == ngamsSRV_REDIRECT)
	{
	/* Find the "Location" HTTP header and get the alternative URL.
	 */
	if ((retCode = ngamsGetHttpHdrEntry(httpHdr, "location", "", 
					    redirectUrl)) != ngamsSTAT_SUCCESS)
	    goto errExit;

	/* Get out the host + port from the URL, e.g.: http://host:1234/...
	 */
	memset(altHost, 0, sizeof(ngamsMED_BUF));
	memset(altPort, 0, sizeof(ngamsMED_BUF));
	strP = strstr(redirectUrl, "//");
	strP += 2;
	while (*strP != ':') strncat(altHost, strP++, 1);
	strP++;
	while (*strP != '/') strncat(altPort, strP++, 1);

	/* This function again (recursively).
	 */
	close(sockFd);
	sockFd = 0;
	if (repDataRef->pdata) free(repDataRef->pdata);
	memset(repDataRef, 0, sizeof(ngamsHTTP_DATA));
	return ngamsHttpGet(altHost, atoi(altPort), userAgent, path,
			    receiveData, repDataRef, dataLen,httpResp,httpHdr);
	}

    /* If specified, read the data from the socket and store it into
     * a buffer of memory allocated.
     */
    if (receiveData && *dataLen)
	{
	if ((retCode = ngamsRecvData(&(repDataRef->pdata), *dataLen, 
				     &sockFd)) != ngamsSTAT_SUCCESS)
	    {
	    if (repDataRef->pdata) free(repDataRef->pdata);
	    memset(repDataRef, 0, sizeof(ngamsHTTP_DATA));
	    goto errExit;
	    }
	}
    else
	repDataRef->fd = sockFd;
    
    ngamsLogDebug("Leaving _ngamsHttpGet()");
    return ngamsSTAT_SUCCESS;
    
  errExit:
    if (sockFd > 0) close(sockFd);
    ngamsLogDebug("Leaving _ngamsHttpGet()/FAILURE. Status: %d", retCode);
    return retCode;
}


/**
int ngamsHttpGet(const char      *host,
		 const int       port,
		 const char      *userAgent,
		 const char      *path,
		 const int       receiveData,
		 ngamsHTTP_DATA  *repDataRef,
		 int             *dataLen,
		 ngamsHTTP_RESP  *httpResp,
		 ngamsHTTP_HDR   httpHdr)
Execute an HTTP GET.

If server multiplexing is used, remember to set the value of the host input
parameter to '\0'.

userAgent:     User agent to put in the HTTP query string.

path:          Path (=URL) for the HTTP GET.

receiveData:   Indicates if possibly data contained in the HTTP 
               response should be received automatically by the
               function (0|1).

repDataRef:    Pointer to instance of ngamsHTTP_DATA structure containing
               possible response data received.

dataLen:       Length of possible response data received.

httpResp:      Instance of ngamsHTTP_RESP structure containing information
               about the HTTP response.

httpHdr:       Instance of the ngamsHTTP_HDR structure containing the
               HTTP headers received in the HTTP response.

Returns:       ngamsSTAT_SUCCESS or error code returned by the functions:

                 ngamsPrepSock(), ngamsRecvHttpHdr(), ngamsGetHttpHdrEntry(),
                 ngamsHttpGet(), ngamsRecvData()

               - or the following error coded: ngamsERR_WR_HD.
*/
int ngamsHttpGet(const char      *host,
		 const int       port,
		 const char      *userAgent,
		 const char      *path,
		 const int       receiveData,
		 ngamsHTTP_DATA  *repDataRef,
		 int             *dataLen,
		 ngamsHTTP_RESP  *httpResp,
		 ngamsHTTP_HDR   httpHdr)
{ 
    char  *locHost;
    int   locPort, maxTries = _ngamsSrvInfo()->numberOfSrvs, tryCount = 0;
    int   stat;

    ngamsLogDebug("Entering ngamsHttpGet() ...");

    if ((*host != '\0') || !_ngamsSrvInfo()->numberOfSrvs)
	{
	 stat = _ngamsHttpGet(host, port, userAgent, path, receiveData,
			      repDataRef, dataLen, httpResp, httpHdr);
	 ngamsLogDebug("Leaving ngamsHttpGet(). Status: %d", stat);
	 return stat;
	}
    else
	{
	/* Simple implementation: Just try the next server if the present
	 * one failed. Could be refined.
	 */
	/* For now we shuffle the servers for simple load balancing. */
	_ngamsShuffleSrvs();
	while (tryCount < maxTries)
	    {
	    _ngamsGetNextSrv(&locHost, &locPort);
	    stat = _ngamsHttpGet(locHost, locPort, userAgent, path, 
				 receiveData, repDataRef, dataLen, httpResp,
				 httpHdr);
	    if (stat == ngamsSTAT_SUCCESS) break;
	    tryCount += 1;
	    }
	if (tryCount == maxTries)
	    {
	    /* IMPL: Should probably make a summarizing response somehow
	       containing info from all attempts . */	 
	    ngamsLogDebug("Leaving ngamsHttpGet()/FAILURE");
	    return ngamsSTAT_FAILURE;
	    }
	}

    ngamsLogDebug("Leaving ngamsHttpGet()");
    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsGenSendCmd(const char            *host,
			  const int             port,
			  const float           timeoutSecs,
			  const ngamsCMD        cmdCode,
			  const ngamsPAR_ARRAY  *parArray,
			  ngamsSTATUS           *status)
Generic function to send a command to the NG/AMS Server specified
by issuing an HTTP GET request.

Returns:       ngamsSTAT_SUCCESS or one of the error codes returned
               by the function ngamsHttpGet().
*/
ngamsSTAT ngamsGenSendCmd(const char            *host,
			  const int             port,
			  const float           timeoutSecs,
			  const ngamsCMD        cmdCode,
			  const ngamsPAR_ARRAY  *parArray,
			  ngamsSTATUS           *status)
{
    int              retCode, repDataLen, i; 
    ngamsHTTP_DATA   repDataRef;
    ngamsHTTP_RESP   httpResp;
    ngamsHTTP_HDR    httpHdr;
    ngamsMED_BUF     tmpUrl, tmpPar, tmpVal;
    ngamsHUGE_BUF    url, urlRaw;

    ngamsLogDebug("Entering ngamsGenSendCmd() ...");

    _setReplyTimeoutFromFloat(timeoutSecs);
    ngamsInitStatus(status);

    /* Prepare the URL */
    ngamsCmd2Str(cmdCode, url);
    strcpy(urlRaw, url);
    for (i = 0; i < parArray->idx; i++)
	{
	strcpy(tmpPar, ngamsEncodeUrlVal(parArray->parArray[i], 1));
	if (parArray->valArray[i][0] != '\0')
	    {
	    sprintf(tmpUrl, "%s=\"%s\"", tmpPar, parArray->valArray[i]);
	    if (i)
		strcat(urlRaw, "&");
	    else
		strcat(urlRaw, "?");
	    strcat(urlRaw, tmpUrl);

	    strcpy(tmpVal, ngamsEncodeUrlVal(parArray->valArray[i], 1));
	    sprintf(tmpUrl, "%s=\"%s\"", tmpPar, tmpVal);
	    }
	else
	    strcpy(tmpUrl, tmpPar);
	if (i)
	    strcat(url, "&");
	else
	    strcat(url, "?");
	strcat(url, tmpUrl);
	}
    ngamsLogInfo(LEV4, "Issuing command with URL: %s ...", urlRaw);
    if ((retCode = ngamsHttpGet(host, port, ngamsUSER_AGENT, url, 1,
				&repDataRef, &repDataLen, &httpResp, 
				httpHdr)) != ngamsSTAT_SUCCESS)
	goto errExit;
    ngamsHandleStatus(retCode, &repDataRef, status);

    ngamsLogDebug("Leaving ngamsGenSendCmd()");
    return ngamsSTAT_SUCCESS;

  errExit: 
    ngamsHandleStatus(retCode, &repDataRef, status);
    ngamsLogDebug("Leaving ngamsGenSendCmd()/FAILURE. Status: %d", retCode);
    return retCode;
}


/**
int _ngamsHttpPost(const char      *host,
	 	   const int       port,
		   const char      *userAgent,
		   const char      *path,
		   const char      *mimeType,
		   const char      *contentDisp,
		   const char      *srcFilename,
		   const char      *data,
		   const int       dataLen,
		   ngamsHTTP_DATA  *repDataRef,
		   int             *repDataLen,
		   ngamsHTTP_RESP  *httpResp,
		   ngamsHTTP_HDR   httpHdr)
Execute an HTTP POST and send possible data referred to in the body
of the HTTP request.

See man-page for ngamsHttpPost() for a description of the parameters.
*/
int _ngamsHttpPost(const char      *host,
		   const int       port,
		   const char      *userAgent,
		   const char      *path,
		   const char      *mimeType,
		   const char      *contentDisp,
		   const char      *srcFilename,
		   const char      *data,
		   const int       dataLen,
		   ngamsHTTP_DATA  *repDataRef,
		   int             *repDataLen,
		   ngamsHTTP_RESP  *httpResp,
		   ngamsHTTP_HDR   httpHdr)
{ 
    char         header[ngamsBUFSIZE], inBuf[1024];
    int          sockFd = 0, fileFd = 0, contLen = 0;
    int          retCode = 0;
    int          bytesRead, hdrLen;
    struct stat  statBuf;
    ngamsBIG_BUF authHdr;

    ngamsLogDebug("Entering _ngamsHttpPost() ...");
    memset(repDataRef, 0, sizeof(ngamsHTTP_DATA));
    *repDataLen = 0;

    /* Figure out the length of the data to send */
    if (*srcFilename != '\0')
	{
	if (stat(srcFilename, &statBuf) == -1)
	    {
	    retCode = ngamsERR_FILE;
	    goto errExit;
	    }
	contLen = statBuf.st_size;
	}
    else if (dataLen >= 0)
        contLen = dataLen;

    /* Open the socket connection to the server */
    if ((sockFd = ngamsPrepSock(host, port)) < 0)
	{
	retCode = sockFd;
	goto errExit;
	}

    /* Prepare and send the HTTP headers */
    if (ngamsGetAuthorization())
	{
	memset(authHdr, 0, sizeof(ngamsHUGE_BUF));
	sprintf(authHdr, "\015\012Authorization: Basic %s", 
		ngamsGetAuthorization());
	}
    else
	*authHdr = '\0';
    sprintf(header,
	    "POST /%.256s HTTP/1.0\015\012"
	    "User-agent: %s\015\012"
	    "Content-type: %s\015\012"
	    "Content-length: %d\015\012"
	    "Content-disposition: %s%s\015\012\012",
	    path, ngamsUSER_AGENT, mimeType, contLen, contentDisp, authHdr);
    hdrLen = strlen(header);
    if (write(sockFd, header, hdrLen) != hdrLen)
	{
	retCode = ngamsERR_WR_HD;
	goto errExit;
	}

    /* Send the data if any */
    if (contLen)
	{
	if (*srcFilename != '\0')
	    {
	    if ((fileFd = open(srcFilename, O_RDONLY)) == -1)
		{
		retCode = ngamsERR_FILE;
		goto errExit;
		}
	    while ((bytesRead = read(fileFd, inBuf, 1024)) > 0)
		{
		if (write(sockFd, inBuf, bytesRead) != bytesRead)
		    {
		    retCode = ngamsERR_WR_DATA;
		    goto errExit;
		    }
		}
	    if (bytesRead < 0)
		{
		retCode = ngamsERR_FILE;
		goto errExit;
		}
	    close(fileFd);
	    }
	else
	    {
	    if (write(sockFd, data, contLen) != contLen)
		{
		retCode = ngamsERR_WR_DATA;
		goto errExit;
		}
	    }
	}

    /* Receive the reply for the request */
    if ((retCode = ngamsRecvHttpHdr(&sockFd, httpHdr, httpResp, repDataRef, 
				    repDataLen)) != ngamsSTAT_SUCCESS)
	goto errExit;

    /* If specified, read the data from the socket and store it into
     * a buffer of memory allocated.
     */
    if (repDataLen)
        {
        if ((retCode = ngamsRecvData(&(repDataRef->pdata), *repDataLen, 
	 			     &sockFd)) != ngamsSTAT_SUCCESS)
	    {
	    if (repDataRef->pdata) free(repDataRef->pdata);
	    memset(repDataRef, 0, sizeof(ngamsHTTP_DATA));
	    goto errExit;
	    }
        }
    if (sockFd) close(sockFd);

    ngamsLogDebug("Leaving _ngamsHttpPost()");
    return ngamsSTAT_SUCCESS;

  errExit:
    if (sockFd) close(sockFd);
    if (fileFd) close(fileFd);
    ngamsLogDebug("Leaving _ngamsHttpPost()/FAILURE. Status: %d", retCode);
    return retCode;
}

/**
int ngamsHttpPost(const char      *host,
	 	  const int       port,
		  const char      *userAgent,
		  const char      *path,
		  const char      *mimeType,
		  const char      *contentDisp,
		  const char      *srcFilename,
		  const char      *data,
		  const int       dataLen,
		  ngamsHTTP_DATA  *repDataRef,
		  int             *repDataLen,
		  ngamsHTTP_RESP  *httpResp,
		  ngamsHTTP_HDR   httpHdr)
Execute an HTTP POST and send possible data referred to in the body
of the HTTP request.

userAgent:     User agent to put in the HTTP query string.

path:          Path (=URL) for the HTTP GET.

mimeType:      Mime-type of the possible data contained in the body of
               the HTTP request.

contentDisp:   Value of Content-disposition HTTP header.

srcFilename:   If different from '\0', the data contained in this file
               will be send in the body of the HTTP request.

data:          String buffer containing data to be send in the body of
               the HTTP request. This is mutual exclusive with possible
               data given in the file referred to by srcFilename.

dataLen:       Length of possible data given in string buffer.

repDataRef:    Pointer to instance of ngamsHTTP_DATA structure containing
               possible response data received.

repDataLen:    Length of possible response data received.

httpResp:      Instance of ngamsHTTP_RESP structure containing information
               about the HTTP response.

httpHdr:       Instance of the ngamsHTTP_HDR structure containing the
               HTTP headers received in the HTTP response.

Returns:       ngamsSTAT_SUCCESS or error code returned by the functions:

                 ngamsPrepSock(), ngamsRecvHttpHdr(), ngamsGetHttpHdrEntry(),
                 ngamsRecvData()

               - or one of the following error codes:

                 ngamsERR_FILE, ngamsERR_WR_HD, ngamsERR_WR_DATA.
*/
int ngamsHttpPost(const char      *host,
		  const int       port,
		  const char      *userAgent,
		  const char      *path,
		  const char      *mimeType,
		  const char      *contentDisp,
		  const char      *srcFilename,
		  const char      *data,
		  const int       dataLen,
		  ngamsHTTP_DATA  *repDataRef,
		  int             *repDataLen,
		  ngamsHTTP_RESP  *httpResp,
		  ngamsHTTP_HDR   httpHdr)
{ 
    char  *locHost;
    int   locPort, maxTries = _ngamsSrvInfo()->numberOfSrvs, tryCount = 0;
    int   stat;

    ngamsLogDebug("Entering ngamsHttpPost() ...");
    if (!_ngamsSrvInfo()->numberOfSrvs)
	{
	stat = _ngamsHttpPost(host, port, userAgent, path, mimeType,
			      contentDisp, srcFilename, data, dataLen,
			      repDataRef, repDataLen, httpResp, httpHdr);
	ngamsLogDebug("Leaving ngamsHttpPost(). Status: %d", stat);
	if (stat != ngamsSTAT_SUCCESS) goto errExit;
	}
    else
	{
	/* Simple implementation: Just try the next server if the present
	 * one failed. Could be refined.
	 */
	/* For now we shuffle the servers for simple load balancing. */
	_ngamsShuffleSrvs();
	while (tryCount < maxTries)
	    {
	    _ngamsGetNextSrv(&locHost, &locPort);
	    stat = _ngamsHttpPost(locHost, locPort, userAgent, path, mimeType,
				  contentDisp, srcFilename, data, dataLen,
				  repDataRef, repDataLen, httpResp, httpHdr);
	    if (stat == ngamsSTAT_SUCCESS) break;
	    tryCount += 1;
	    }
	if (tryCount == maxTries)
	    {
	    /* IMPL: Should probably make a summarizing response somehow
	       containing info from all attempts . */
	    stat = ngamsSTAT_FAILURE;
	    goto errExit;
	    }
	}

    ngamsLogDebug("Leaving ngamsHttpPost()");
    return ngamsSTAT_SUCCESS;

 errExit:
    ngamsLogDebug("Leaving ngamsHttpPost()/FAILURE. Status: %d", stat);
    return stat;
}


/**
*******************************************************************************
*/


/**
*******************************************************************************
Functions to Handle Logging.
*******************************************************************************
*/

/**
void ngamsLog_v(const char            *type,
		const ngamsLOG_LEVEL  level,
		const char            *format,
		va_list               vaParList)
Generic log function taken a variable list of input parameters.

type:        Type of log.

level:       Log Level (intensity, [0; 5]).

format:      Format of log string in C format style.

vaParList:   Compiled list of variable input parameters. This must be
             generated with the system call va_start(). After this function
             returns, this must be freed with va_end() to free allocated
             memory.

Returns:     Void.
*/
void ngamsLog_v(const char            *type,
		const ngamsLOG_LEVEL  level,
		const char            *format,
		va_list               vaParList)
{
    int           fd = 0;        
    ngamsHUGE_BUF logMsg, tmpLogMsg;
    ngamsMED_BUF  isoTime;
    ngamsSTAT     stat;

    /* Create log line */
    ngamsGenIsoTime(3, isoTime);
    vsprintf((char *)tmpLogMsg, format, vaParList);
    sprintf(logMsg, "%s [%s] %s\n%s", isoTime, type, tmpLogMsg, "\0");

    /* Log in log file */
    if (level <= _logLevel)
	{
	/* IMPL: Support buffering to avoid opening and closing the file all
	 *       the time 
	 */
	if ((fd = open(_logFile, O_WRONLY | O_CREAT, 0644)) == -1)
	    {
	    stat = ngamsERR_OPEN_LOG_FILE;
	    goto errExit;
	    }
	lseek(fd, 0, SEEK_END);
	write(fd, logMsg, strlen(logMsg));
	close(fd);
	}

    /* Log on stdout */
    if (level <= _verboseLevel) printf(logMsg);

 errExit:
    if (fd) close(fd);
}


/**
ngamsSTAT ngamsPrepLog(const ngamsMED_BUF   logFile,
		       const ngamsLOG_LEVEL logLevel,
		       const int            logRotate,
		       const int            logHistory)
Set up the logging environment for the NG/AMS C-Client functions.

logFile:         Name of log file.

logLevel:        Log level (intensity, [0; 5]).

logRotate:       Seconds since midnight when the log file should be rotated.

logHistory:      Number of log files (=days) to keep.

Returns:         ngamsSTAT_SUCCESS or ngamsERR_OPEN_LOG_FILE.
*/
ngamsSTAT ngamsPrepLog(const ngamsMED_BUF   logFile,
		       const ngamsLOG_LEVEL logLevel,
		       const int            logRotate,
		       const int            logHistory)
{
    int fd;

    strcpy(_logFile, logFile);
    _logLevel = logLevel;
    _logRotate = logRotate;
    _logHistory = logHistory;

    /* Create the log file, RW for owner, R for group/others */
    if ((fd = open(logFile, (O_RDWR | O_CREAT), ngamsSTD_PERMS)) == -1)
	return ngamsERR_OPEN_LOG_FILE;
    close(fd);
    return ngamsSTAT_SUCCESS;
}

/**
void ngamsSetVerboseLevel(const ngamsLOG_LEVEL level)
Set the Verbose Level to be used. Verbose Logs are written to stdout.

level:    Verbose Level (intensity, [0; 5]).

Returns:  Void.
*/
void ngamsSetVerboseLevel(const ngamsLOG_LEVEL level)
{
    _verboseLevel = level;
}

/**
void ngamsInitLogLevel()
Set the Verbose Level to the value of the environment variable NGAMS_LOG_LEVEL
(if defined).

Returns:  Void.
*/
void ngamsInitLogConds()
{
    static int initialized;
    char       *charP;

    if (!initialized)
	{
	if ((charP = getenv("NGAMS_LOG_FILE")) != NULL)
	    strcpy(_logFile, charP);
	if ((charP = getenv("NGAMS_LOG_LEVEL")) != NULL)
	    _logLevel = atoi(charP);
	if ((charP = getenv("NGAMS_VERBOSE_LEVEL")) != NULL)
	    ngamsSetVerboseLevel(atoi(charP));
	initialized = 1;
	}
}

/**
void ngamsLogEmerg(const char  *format, ...)
Generate an Emergency Log entry (EMERG).

format:   Format of log entry.

...:      Values to be written in log format.

Returns:  Void.
*/
void ngamsLogEmerg(const char  *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);
    ngamsLog_v("EMERG", LEV0, format, vaParList);
    va_end(vaParList);
}

/**
void ngamsLogAlert(const char  *format, ...)
Generate an Alert Log entry (ALERT).

See ngamsLogEmerg().
*/
void ngamsLogAlert(const char  *format, ...)
{ 
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);
    ngamsLog_v("ALERT", LEV0, format, vaParList);
    va_end(vaParList);
}

/**
void ngamsLogCrit(const char  *format, ...)
Generate a Critical Log entry (CRIT).

See ngamsLogEmerg().
*/
void ngamsLogCrit(const char  *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);
    ngamsLog_v("CRIT", LEV0, format, vaParList); 
    va_end(vaParList);
}

/**
void ngamsLogError(const char  *format, ...)
Generate an Error Log entry (ERROR).

See ngamsLogEmerg().
*/
void ngamsLogError(const char  *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);
    ngamsLog_v("ERROR", LEV0, format, vaParList);
    va_end(vaParList);
}

/**
void ngamsLogWarning(const char  *format, ...)
Generate a Warning Log entry (WARNING).

See ngamsLogEmerg().
*/
void ngamsLogWarning(const char  *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);   
    ngamsLog_v("WARNING", LEV0, format, vaParList); 
    va_end(vaParList);
}


/**
void ngamsLogNotice(const char  *format, ...)
Generate a Notice Log entry (NOTICE).

See ngamsLogEmerg().
*/
void ngamsLogNotice(const char  *format, ...)
{
    va_list        vaParList;
    
    ngamsInitLogConds();
    va_start(vaParList, format);      
    ngamsLog_v("NOTICE", LEV0, format, vaParList);
    va_end(vaParList);
}


/**
void ngamsLogInfo(const char  *format, ...)
Generate an Info Log entry (INFO).

See ngamsLogEmerg().
*/
void ngamsLogInfo(const ngamsLOG_LEVEL level,
		  const char           *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    va_start(vaParList, format);
    ngamsLog_v("INFO", level, format, vaParList);
    va_end(vaParList);
}

/**
void ngamsLogDebug(const char  *format, ...)
Generate an Debug Log entry (DEBUG).

See ngamsLogEmerg().
*/
void ngamsLogDebug(const char  *format, ...)
{
    va_list        vaParList;

    ngamsInitLogConds();
    if ((_logLevel >= LEV5) || (_verboseLevel >= LEV5))
	{ 
	va_start(vaParList, format);
	ngamsLog_v("DEBUG", LEV5, format, vaParList);
	va_end(vaParList);
	}
}

/**
ngamsSTAT ngamsLogFileRotate(const ngamsLOG_LEVEL  tmpLogLevel,
			     const ngamsMED_BUF    systemId,
			     ngamsMED_BUF          rotatedLogFile)
Check if the log file should rotated and rotate it if needed.

tmpLogLevel:     Temporary Log Level to apply during the process of
                 rotating the log file.

systemId:        ID of the system.

rotatedLogFile:  Name of the rotated log file.

Returns:         ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsLogFileRotate(const ngamsLOG_LEVEL  tmpLogLevel,
			     const ngamsMED_BUF    systemId,
			     ngamsMED_BUF          rotatedLogFile)
{
    char                      *chrP;
    struct timeval            timeEpoch;
    struct tm                 *timeDay;
    unsigned long int         secsOfDay, secsLastRot;
    ngamsMED_BUF              path, logFilename, tmpLogBuf;

    if (_logRotate == -1) return ngamsSTAT_SUCCESS;
    *rotatedLogFile = '\0';

    gettimeofday(&timeEpoch, NULL); 
    timeDay = localtime((time_t *)&timeEpoch.tv_sec);
    secsOfDay = ((3600 * timeDay->tm_hour) + (60 * timeDay->tm_min) +
		 timeDay->tm_sec);
    
    /* Get time for last rotatation and determine from this if we should 
     * rotate now.
     */
    ngamsLoadFile(_logFile, tmpLogBuf, sizeof(ngamsMED_BUF));
    if ((chrP = strstr(tmpLogBuf, ngamsLOG_ROT_PREFIX)) == NULL)
	secsLastRot = 0;
    else
	{
	/* The log entry has a format like, e.g.:
	 *
	 * 2003-12-24T08:36:18.133 [INFO] LOG-ROTATE: 1072254978 \
	 * - SYSTEM-ID: <ID>
	 */
	sscanf((chrP + strlen(ngamsLOG_ROT_PREFIX) + 2), "%ld", &secsLastRot);
	}

    if (((timeEpoch.tv_sec - secsLastRot) > (24 * 3600)) &&
	(secsOfDay >= _logRotate))
	{
	ngamsLogInfo(tmpLogLevel, "Rotating log file ...");
	ngamsSplitFilename(_logFile, path, logFilename);
	sprintf(rotatedLogFile, "%s/%s_%ld_%s", path, ngamsLOG_ROT_PREFIX,
		(long int)timeEpoch.tv_sec, logFilename);
	if (rename(_logFile, rotatedLogFile) == -1)
	    {
	    ngamsLogError("Serious error ocurred rotating log file: %s "
			  "- rotated log file: %s!", _logFile, rotatedLogFile);
	    return ngamsSTAT_FAILURE;
	    }
	ngamsPrepLog(_logFile, _logLevel, _logRotate, _logHistory);
	ngamsLogInfo(tmpLogLevel, "%s: %ld - SYSTEM-ID: %s", 
		     ngamsLOG_ROT_PREFIX, timeEpoch.tv_sec, systemId);
	}

    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsCleanUpRotLogFiles(const ngamsLOG_LEVEL  tmpLogLevel)
Check if there are rotated log files to remove and remove them if this is
the case.

tmpLogLevel:    Temporary Log Level to apply during the remove process.

Returns:        ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsCleanUpRotLogFiles(const ngamsLOG_LEVEL  tmpLogLevel)
{
    DIR                *dirPtr = NULL;
    struct dirent      *dirEnt = NULL;
    struct timeval     timeEpoch;
    ngamsMED_BUF       tmpLogFilename, logFilePath, logFilename;
    unsigned long int  secsLastRot, daysSinceLastRot;

    if (_logHistory == -1) return ngamsSTAT_SUCCESS;

    ngamsSplitFilename(_logFile, logFilePath, logFilename);
    if ((dirPtr = opendir(logFilePath)) == NULL)
	{    
	ngamsLogError("Error opening Log Files Directory: %s", logFilePath);
	goto errExit;
	}
    while ((dirEnt = readdir(dirPtr)) != NULL)
	{
	if (strstr(dirEnt->d_name, ngamsLOG_ROT_PREFIX) != NULL)
	    {
	    sprintf(tmpLogFilename, "%s/%s", logFilePath, dirEnt->d_name);
	    
	    /* The format of the names of the rotated log files is, e.g.:
	     *
	     * LOG_ROTATE_<Secs/Epoch Rotation>_ngamsArchiveClient.log
	     *
	     * Get the seconds since epoch for when the log file was rotated.
	     */
	    sscanf((dirEnt->d_name + strlen(ngamsLOG_ROT_PREFIX) + 1), "%ld", 
		   &secsLastRot);
	    gettimeofday(&timeEpoch, NULL); 
	    daysSinceLastRot = ((timeEpoch.tv_sec - secsLastRot)/(3600 * 24));
	    if (daysSinceLastRot >= _logHistory)
		{
		ngamsLogInfo(tmpLogLevel, "Removing rotated log file: %s", 
			     tmpLogFilename);
		remove(tmpLogFilename);
		}
	    }
	}
    closedir(dirPtr);

    return ngamsSTAT_SUCCESS;

 errExit:
    if (dirPtr != NULL) closedir(dirPtr);
    return ngamsSTAT_FAILURE;
}

/**
*******************************************************************************
*/

/**
*******************************************************************************
Various Utility Functions.
*******************************************************************************
*/

/**
void _ngamsSetAuthorization(const char   *authUserPass)
Function to set autorization user/password, which will subsequently be
sent with all requests automatically if defined.

authUserPass:   Encrypted user/password. Should be provided by the responsible
                of the NG/AMS Server side (encrypted).              

Returns:        Void.
*/
void ngamsSetAuthorization(const char   *authUserPass)
{
    _authorization = malloc(strlen(authUserPass) + 1);
    strcpy(_authorization, authUserPass);
}


/**
void _ngamsGetAuthorization(void)
Function to retrieve the autorization user/password defined in the module.

Returns:        Pointer to authorization user/password (encrypted).
*/
char *ngamsGetAuthorization(void)
{
    return _authorization;
}


/**
void ngamsAddParAndVal(ngamsPAR_ARRAY  *parArray,
		       const char      *par,
		       const char      *val)
Add a parameter and a corresponding value in the instance of the 
ngamsPAR_ARRAY structure.

parArray:    Instance of the ngamsPAR_ARRAY structure.

par:         Name of parameter.

val:         Value in connection with paameter. Can be '\0'.

Returns:     Void.
*/
void ngamsAddParAndVal(ngamsPAR_ARRAY  *parArray,
		       const char      *par,
		       const char      *val)
{
    strcpy(parArray->parArray[parArray->idx], par);
    strcpy(parArray->valArray[parArray->idx], val);
    parArray->idx++;
}

/**
ngamsSTAT ngamsCmd2No(const ngamsSMALL_BUF    cmdStr,
                      ngamsCMD                *cmdCode)
Convert a command given as string into the corresponding code (integer).

cmdStr:    Command name as string.

cmdCode:   Command code as defined by the enumerated type ngamsCMD.

Returns:   Command code or ngamsERR_UNKNOWN_CMD.
*/
ngamsSTAT ngamsCmd2No(const ngamsSMALL_BUF    cmdStr,
		      ngamsCMD                *cmdCode)
{
    if (strcmp(cmdStr, ngamsCMD_ARCHIVE_STR) == 0)
	*cmdCode = ngamsCMD_ARCHIVE;
    else if (strcmp(cmdStr, ngamsCMD_CHECKFILE_STR) == 0)
	*cmdCode = ngamsCMD_CHECKFILE;
    else if (strcmp(cmdStr, ngamsCMD_CLONE_STR) == 0)
	*cmdCode = ngamsCMD_CLONE;
    else if (strcmp(cmdStr, ngamsCMD_DISCARD_STR) == 0)
        *cmdCode = ngamsCMD_DISCARD;
    else if (strcmp(cmdStr, ngamsCMD_EXIT_STR) == 0)
	*cmdCode = ngamsCMD_EXIT;
    else if (strcmp(cmdStr, ngamsCMD_INIT_STR) == 0)
	*cmdCode = ngamsCMD_INIT;
    else if (strcmp(cmdStr, ngamsCMD_LABEL_STR) == 0)
	*cmdCode = ngamsCMD_LABEL;
    else if (strcmp(cmdStr, ngamsCMD_ONLINE_STR) == 0)
	*cmdCode = ngamsCMD_ONLINE;
    else if (strcmp(cmdStr, ngamsCMD_OFFLINE_STR) == 0)
	*cmdCode = ngamsCMD_OFFLINE;
    else if (strcmp(cmdStr, ngamsCMD_REGISTER_STR) == 0)
	*cmdCode = ngamsCMD_REGISTER;
    else if (strcmp(cmdStr, ngamsCMD_REMDISK_STR) == 0)
	*cmdCode = ngamsCMD_REMDISK;
    else if (strcmp(cmdStr, ngamsCMD_REMFILE_STR) == 0)
	*cmdCode = ngamsCMD_REMFILE;
    else if (strcmp(cmdStr, ngamsCMD_RETRIEVE_STR) == 0)
	*cmdCode = ngamsCMD_RETRIEVE;
    else if (strcmp(cmdStr, ngamsCMD_STATUS_STR) == 0)
	*cmdCode = ngamsCMD_STATUS;
    else if (strcmp(cmdStr, ngamsCMD_SUBSCRIBE_STR) == 0)
	*cmdCode = ngamsCMD_SUBSCRIBE;
    else if (strcmp(cmdStr, ngamsCMD_UNSUBSCRIBE_STR) == 0)
	*cmdCode = ngamsCMD_UNSUBSCRIBE;
    else
	{
	*cmdCode = 0;
	return ngamsERR_UNKNOWN_CMD;
	}

    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsCmd2Str(const ngamsCMD    cmdCode,
                       ngamsSMALL_BUF    cmdStr)
Convert an NG/AMS command given as a code (integer) to a string.

cmdCode,
cmdStr:    See ngamsCmd2No().

Returns:   Command as string.
*/
ngamsSTAT ngamsCmd2Str(const ngamsCMD    cmdCode,
		       ngamsSMALL_BUF    cmdStr)
{
    switch (cmdCode)
	{
	case ngamsCMD_ARCHIVE:
	    strcpy(cmdStr, ngamsCMD_ARCHIVE_STR);
	    break;
	case ngamsCMD_CHECKFILE:
	    strcpy(cmdStr, ngamsCMD_CHECKFILE_STR);
	    break;
	case ngamsCMD_CLONE:
	    strcpy(cmdStr, ngamsCMD_CLONE_STR);
	    break;
	case ngamsCMD_DISCARD:
	    strcpy(cmdStr, ngamsCMD_DISCARD_STR);
	    break;
	case ngamsCMD_EXIT:
	    strcpy(cmdStr, ngamsCMD_EXIT_STR);
	    break;
	case ngamsCMD_INIT:
	    strcpy(cmdStr, ngamsCMD_INIT_STR);
	    break;
	case ngamsCMD_LABEL:
	    strcpy(cmdStr, ngamsCMD_LABEL_STR);
	    break;
	case ngamsCMD_ONLINE:
	    strcpy(cmdStr, ngamsCMD_ONLINE_STR);
	    break;
	case ngamsCMD_OFFLINE:
	    strcpy(cmdStr, ngamsCMD_OFFLINE_STR);
	    break;
	case ngamsCMD_REGISTER:
	    strcpy(cmdStr, ngamsCMD_REGISTER_STR);
	    break;
	case ngamsCMD_REMDISK:
	    strcpy(cmdStr, ngamsCMD_REMDISK_STR);
	    break;
	case ngamsCMD_REMFILE:
	    strcpy(cmdStr, ngamsCMD_REMFILE_STR);
	    break;
	case ngamsCMD_RETRIEVE:
	    strcpy(cmdStr, ngamsCMD_RETRIEVE_STR);
	    break;
	case ngamsCMD_STATUS:
	    strcpy(cmdStr, ngamsCMD_STATUS_STR);
	    break;
	case ngamsCMD_SUBSCRIBE:
	    strcpy(cmdStr, ngamsCMD_SUBSCRIBE_STR);
	    break;
	case ngamsCMD_UNSUBSCRIBE:
	    strcpy(cmdStr, ngamsCMD_UNSUBSCRIBE_STR);
	    break;
	default:
	    *cmdStr = '\0';
	    return ngamsERR_UNKNOWN_CMD;
	    break;
	}   /* end-switch (cmdCode) */

    return ngamsSTAT_SUCCESS;
}

/**
void ngamsDumpStatStdout(const ngamsSTATUS  *status)
Dump information contained in the instance of the ngamsSTATUS structure
to stdout.

Returns:     Void.
*/
void ngamsDumpStatStdout(const ngamsSTATUS  *status)
{
    if (*status->date != '\0') 
	printf("\nDate:           %s", status->date);
    printf("\nError Code:     %d", status->errorCode);
    if (*status->hostId != '\0')
	printf("\nHost ID:        %s", status->hostId);
    if (*status->message != '\0')
	printf("\nMessage:        %s", status->message);
    if (*status->status != '\0')
	printf("\nStatus:         %s", status->status);
    if (*status->state != '\0')
	printf("\nState:          %s", status->state);
    if (*status->subState != '\0')
	printf("\nSub-State:      %s", status->subState);
    if (*status->version != '\0')
	printf("\nNG/AMS Version: %s", status->version);
    printf("\n\n");
}

/**
char *ngamsEncodeUrlVal(const char *urlVal,
                        const int  skipScheme)
Encode the value given as input parameter to replace special
characters to make the value suitable for usage in a URL.

urlVal:            Value to be encoded.

skipScheme:        If the value is initiated with an HTTP scheme
                   (ftp:, http:, file:), this will not be encoded
                   if this flag is set to 1.

Returns:           Encoded value. This is stored in an internal, static
                   buffer and a local copy should be made before invoking
                   the function again (char *).
*/
char *ngamsEncodeUrlVal(const char *urlVal,
			const int  skipScheme)
{
    static ngamsMED_BUF  encVal;
    char                 *strP;
    int                  idx = 0;

    *encVal = '\0';
    if (skipScheme)
	{
	if ((strstr(urlVal, "http:") == urlVal) ||
	    (strstr(urlVal, "file:") == urlVal))
	    idx = 5;
	else if (strstr(urlVal, "ftp:") == urlVal)
	    idx = 4;
	}

    strP = (char *)(urlVal + idx);
    strncpy(encVal, urlVal, idx);
    *(encVal + idx) = '\0';
    while (*strP != '\0')
	{
	if (*strP == ':')
	    strcat(encVal, "%3A");
	else if (*strP == '?')
	    strcat(encVal, "%3F");
	else if (*strP == '=')
	    strcat(encVal, "%3D");
	else if (*strP == '&')
	    strcat(encVal, "%26");
	else
	    strncat(encVal, strP, 1);
	strP++;
	}
    return encVal;
}

/**
void ngamsFreeStatus(ngamsSTATUS         *status)

Free the memory occupied by the status object.

status:     Pointer to instance of the ngamsSTATUS structure.

Returns:    Void.
*/
void ngamsFreeStatus(ngamsSTATUS         *status)
{
    int i = 0;
    while (status->replyData[i] != NULL)
      {
      free(status->replyData[i]);
      status->replyData[i] = NULL;
      }
}

/**
void ngamsGenIsoTime(const int      prec,
		     ngamsMED_BUF   dayTime)
Generate an ISO8601 time string of the format:

  <YYYY>-<MM><DD>T<HH>:<MM>:<SS>[.<ss>]

The time stamp will be generated at the time of invocation of the function.

prec:        Precision (number of decimals).

dayTime:     Generated ISO8601 time stamp.

Returns:     Void.
*/
void ngamsGenIsoTime(const int      prec,
		     ngamsMED_BUF   dayTime)
{
    char            day[16], month[16], date[16], timeStr[16], year[16];
    char            date2[16];
    struct timeval  timeEpoch;
    ngamsMED_BUF    format;
    struct tm       *timeNow;
    ngamsMED_BUF    tmpBuf, tmpBuf2;
    int             intVal;

    gettimeofday(&timeEpoch, NULL); 
    timeNow = gmtime((time_t *)&timeEpoch.tv_sec);
    strcpy(tmpBuf, asctime(timeNow));

    /* Now, reformat:
     *
     * "Sun Sep 16 01:03:52 1973\n"
     *
     * into:
     *
     * "1973-09-16T01:03:52"
     */ 
    sscanf(tmpBuf, "%s %s %s %s %s", day, month, date, timeStr, year);

    /* Convert Month as string to number */
    if (strstr(month, "Jan") != NULL)
	sprintf(month, "01");
    else if (strstr(month, "Feb") != NULL)
	sprintf(month, "02");
    else if (strstr(month, "Mar") != NULL)
	sprintf(month, "03");
    else if (strstr(month, "Apr") != NULL)
	sprintf(month, "04");
    else if (strstr(month, "May") != NULL)
	sprintf(month, "05");
    else if (strstr(month, "Jun") != NULL)
	sprintf(month, "06");
    else if (strstr(month, "Jul") != NULL)
	sprintf(month, "07");
    else if (strstr(month, "Aug") != NULL)
	sprintf(month, "08");
    else if (strstr(month, "Sep") != NULL)
	sprintf(month, "09");
    else if (strstr(month, "Oct") != NULL)
	sprintf(month, "10");
    else if (strstr(month, "Nov") != NULL)
	sprintf(month, "11");
    else 
	sprintf(month, "12");
    
    /* If date is less than 10 put a 0 in front */
    intVal = atoi(date);
    if (intVal < 10)
	sprintf(date2, "0%d", intVal);
    else
	sprintf(date2, "%d", intVal);
   
    /* Remove possible newline character in the year */
    if (*(year + strlen(year) - 1) == '\n')
	*(year + strlen(year) - 1) = '\0';

    /* Print the DAYTIM: "1973-09-16T01:03:52[.SSS]" */
    sprintf(dayTime, "%s-%s-%sT%s", year, month, date2, timeStr); 

    /* Add fractions of seconds if specified */
    if (prec)
	{
	sprintf(format, "%%.%df", prec);
	sprintf(tmpBuf, format, (timeEpoch.tv_usec / 1e6));
	strcpy(tmpBuf2, (tmpBuf + 1));
	strcat(dayTime, tmpBuf2);
	}
}

/**
ngamsSTAT ngamsGetHostName(ngamsSMALL_BUF hostName)
Return the host name of the local host.

hostName:    Buffer which will contain the hostname.

Returns:     ngamsSTAT_SUCCESS.
*/
ngamsSTAT ngamsGetHostName(ngamsSMALL_BUF hostName)
{
    char    *chrP;

    if ((chrP = getenv("HOSTNAME")) == NULL)
	if ((chrP = getenv("HOST")) == NULL)
	    return ngamsSTAT_FAILURE;
    strcpy(hostName, chrP);

    return ngamsSTAT_SUCCESS;
}


/**
char *ngamsGetParVal(ngamsPAR_ARRAY  *parArray,
	             const char      *par)
Return the value associated to a parameter. If the parameter is not defined,
NULL is returned.

parArray:    Instance of the ngamsPAR_ARRAY structure.

par:         Name of parameter.

val:         Value in connection with paameter. Can be '\0'.

Returns:     Void.
*/
char *ngamsGetParVal(ngamsPAR_ARRAY  *parArray,
		     const char      *par)
{
    char  *val = NULL;
    int   idx = 0;

    while (idx < parArray->idx)
	{
	if (strcmp(par, parArray->parArray[idx]) == 0)
	    {
	    val = parArray->valArray[idx];
	    break;
	    }
	idx++;
	}

    return val;
}

/**
ngamsSTAT ngamsGetXmlAttr(const char    *xmlDoc,
			  const char    *pt,
			  const char    *attr,
			  const int     maxValLen,
			  char          *value)
Get the value of an XML attribute contained in an NG/AMS XML Status Document.

xmlDoc:       Buffer containing the NT/AMS XML Status Document.

pt:           Name of point in which the attribute is contained.

attr:         Name of the attribute to retrieve the value for.

maxValLen:    Maximum length of the attribute value to read.

value:        Buffer allocated by the calling function containing the
              value of the attribute.

Returns:      The status code: ngamsSTAT_SUCCESS or ngamsERR_INV_REPLY.
*/
ngamsSTAT ngamsGetXmlAttr(const char    *xmlDoc,
			  const char    *pt,
			  const char    *attr,
			  const int     maxValLen,
			  char          *value)
{
    char            *ptP, *attrP, *valP;
    int             idx = 0;
    ngamsSMALL_BUF  tmpPt, tmpAttr;

    sprintf(tmpPt, "<%s", pt);
    sprintf(tmpAttr, "%s=", attr);

    /* Find first the element then the attribute */
    if ((ptP = strstr(xmlDoc, tmpPt)) == NULL) return ngamsERR_INV_REPLY;
    ptP += strlen(tmpPt);
    if ((attrP = strstr(ptP, tmpAttr)) == NULL) return ngamsERR_INV_REPLY;

    /* Find the start of the attribute value. Expect something like: 
     * <Status Date="2001-08-30T16:56:01.101" HostId="arcus2" ...
     */
    valP = attrP;
    while ((*valP != '\0') && (*valP != '"')) valP++;
    if (*valP == '\0') return ngamsERR_INV_REPLY;
    valP++;

    /* Get out the value */
    while ((idx < maxValLen) && (*valP != '\0') && (*valP != '"'))
	{
	value[idx++] = *valP;
	valP++;
	}
    if ((*valP == '\0') || (idx == maxValLen)) return ngamsERR_INV_REPLY;
    value[idx] = '\0';

    return ngamsSTAT_SUCCESS;
}

/**
void ngamsInitStatus(ngamsSTATUS         *status)

Initialize the ngamsSTATUS structure, making it ready for usage within
the NG/AMS C-API functions.

status:   Pointing to instance of the ngamsSTATUS structure containing 
          the information to print out.
 
Returns:  Void.
*/
void ngamsInitStatus(ngamsSTATUS         *status)
{
    memset(status, 0, sizeof(ngamsSTATUS));
}

/**
int ngamsIsDir(const char  *filename)

Return 1 if the filename given is a directory, otherwise 0 is returned.

filename:    Name of directory to probe.

Returns:     1 = directory, 0 = not a directory.
*/
int ngamsIsDir(const char  *filename)
{
    struct stat      statBuf;
    
    if (stat(filename, &statBuf) == -1)
	return 0;
    if ((statBuf.st_mode & S_IFMT) == S_IFDIR)
	return 1;
    return 0;
}

/**
char *ngamsLicense(void)

Return pointer to buffer containing the NG/AMS License Agreement.

Returns:   Pointer to static buffer containing the license agreement
           for NG/AMS.
*/
char *ngamsLicense(void)
{
    return _ngamsLicense();
}

/**
int ngamsLogCodeInStatus(const ngamsSTATUS     *status,
			 const char            *errId)
Probe if the given log/status code is contained in the ngamsSTAT object.

errId:     ID of the error, e.g. NGAMS_ER_DAPI_BAD_FILE.

Returns:   1 if specified log code is contained in the status object.
*/
int ngamsLogCodeInStatus(const ngamsSTATUS     *status,
			 const char            *errId)
{
    ngamsSMALL_BUF  codeBuf;
    sprintf(codeBuf, "%s:", errId);
    if (strstr(status->message, codeBuf) != NULL)
	return 1;
    else
	return 0;
}


/**
ngamsSTAT ngamsLoadFile(const char    *filename,
			char          *buf,
			const int     maxSize)
Load the given file into the buffer allocated by the calling function.

filename:      Name of file to load.

buf:           Buffer in which the contents of the file will be written.
	
maxSize:       Maximum size to read.

Returns:       ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsLoadFile(const char    *filename,
			char          *buf,
			const int     maxSize)
{
    int fd, size;

    ngamsLogInfo(LEV4, "Opening/loading file: %s ...", filename);
    if ((fd = open(filename, O_RDONLY, 0)) == -1)
	{
	ngamsLogError("Error ocurred opening file: %s", filename);
	return ngamsSTAT_FAILURE;
	}
    size = read(fd, buf, maxSize);
    buf[size] = '\0';
    close(fd);
    ngamsLogInfo(LEV4, "Opened/loaded file: %s", filename);

    return ngamsSTAT_SUCCESS;
}

/**
char *ngamsManPage(void)
Return reference to the man-page of the NG/AMS C-Client.

Returns:    Pointer to static buffer containing the man-page
            for the NG/AMS C-Client.
*/
char *ngamsManPage(void)
{
    return _ngamsManPage();
}

/**
void ngamsResetParArray(ngamsPAR_ARRAY  *parArray)
Initialize the instance of the ngamsPAR_ARRAY structure.

parArray:  Instance of the ngamsPAR_ARRAY structure.

Returns:   Void.
*/
void ngamsResetParArray(ngamsPAR_ARRAY  *parArray)
{
    memset(parArray, 0, sizeof(ngamsPAR_ARRAY));
}

/**
ngamsSTAT ngamsSaveInFile(const char *filename,
			  const char *buf)
Save the contents of the buffer in the given file.

filename:   Name of file in which to store the contents of buffer.

buf:        Buffer containing the information to be stored in the file.

Returns:    ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsSaveInFile(const char *filename,
			  const char *buf)
{
    int fd;

    if ((fd = open(filename, (O_WRONLY | O_CREAT), ngamsSTD_PERMS)) == -1)
	{
	ngamsLogError("Error ocurred creating file: %s", filename);
	return ngamsSTAT_FAILURE;
	}
    write(fd, buf, strlen(buf));
    close(fd);

    return ngamsSTAT_SUCCESS;
}


/**
ngamsSTAT ngamsSplitFilename(const char     *complPath,
			     ngamsMED_BUF   path,
			     ngamsMED_BUF   filename)
Split a filename is path and actual filename.

complPath:   Complete path of filename.

path:        Path of filename.

filename:    Filename.

Returns:     ngamsSTAT_SUCCESS.
*/
ngamsSTAT ngamsSplitFilename(const char     *complPath,
			     ngamsMED_BUF   path,
			     ngamsMED_BUF   filename)
{
    char  *chrP;

    memset(path, 0, sizeof(ngamsMED_BUF));
    memset(filename, 0, sizeof(ngamsMED_BUF));
    chrP = ngamsStrRStr(complPath, "/");
    strncpy(path, complPath, (chrP - complPath));
    strcpy(filename, (chrP + 1));

    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsSplitSrvAddr(const char  *srvAddr,
                            char        *host,
                            int         *port)
Split a server address given as 'node:port' into its components.

srvAddr:   Server address.

host:      Host name of the address.

port:      Port number of the address.

Returns:   ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsSplitSrvAddr(const char  *srvAddr,
                            char        *host,
                            int         *port)
{
    char      *chrPtr;

    if ((chrPtr = strchr(srvAddr, ':')) == NULL) return ngamsSTAT_FAILURE;
    strncpy(host, srvAddr, (chrPtr - srvAddr));
    *(host + (chrPtr - srvAddr)) = '\0';
    *port = atoi(chrPtr + 1);
    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsSplitParVal(const char  *parVal,
                           char        *par,
                           char        *val)
Split a server address given as 'node:port' into its components.

parVal:     Parameter value set (<Par>=<Val>).

par:        Parameter name.

val:        Value.

Returns:   ngamsSTAT_SUCCESS or ngamsSTAT_FAILURE.
*/
ngamsSTAT ngamsSplitParVal(const char  *parVal,
                           char        *par,
                           char        *val)
{
    char      *chrPtr;

    if ((chrPtr = strchr(parVal, '=')) == NULL) return ngamsSTAT_FAILURE;
    strncpy(par, parVal, (chrPtr - parVal));
    *(par + (chrPtr - parVal)) = '\0';
    strcpy(val, (chrPtr + 1));
    return ngamsSTAT_SUCCESS;
}

/**
ngamsSTAT ngamsStat2Str(const ngamsSTAT  statNo,
                        ngamsMED_BUF     statStr)

Convert a status code (ngamsSTAT) to a readable string.

statNo:    Status as code.

statStr:   Status as string.

Returns:   ngamsSTAT_SUCCESS or ngamsERR_UNKNOWN_STAT.
*/
ngamsSTAT ngamsStat2Str(const ngamsSTAT  statNo,
			ngamsMED_BUF     statStr)
{
    switch (statNo)
	{
	case ngamsSTAT_SUCCESS:
	    strcpy(statStr, "Status OK");
	    break;
	case ngamsERR_UNKNOWN_CMD:
	    strcpy(statStr, "Unknown command issued");
	    break;
	case ngamsERR_INV_TARG_FILE:
	    strcpy(statStr, "Invalid target filename specified");
	    break;
	case ngamsERR_UNKNOWN_STAT:
	    strcpy(statStr, "Unknown status code");
	    break;
	case ngamsERR_INV_PARS:
	    strcpy(statStr, "Illegal parameters given");
	    break;
	case ngamsERR_HOST:
	    strcpy(statStr, "No such host");
	    break;
	case ngamsERR_SOCK:
	    strcpy(statStr, "Cannot create socket");
	    break;
	case ngamsERR_CON:
	    strcpy(statStr, "Cannot connect to host/server");
	    break;
	case ngamsERR_COM:
	    strcpy(statStr, "Problem communicating with server");
	    break;
	case ngamsERR_TIMEOUT:
	    strcpy(statStr, "Timeout encountered while communicating with "
		   "server");
	    break;
	case ngamsERR_WR_HD:
	    strcpy(statStr, "Write error on socket while writing header");
	    break;
	case ngamsERR_WR_DATA:
	    strcpy(statStr, "Write error on socket while writing data");
	    break;
	case ngamsERR_INV_REPLY:
	    strcpy(statStr, "Invalid reply from data server");
	    break;
	case ngamsERR_ALLOC_MEM:
	    strcpy(statStr, "Cannot allocate memory");
	    break;
	case ngamsERR_RD_DATA:
	    strcpy(statStr, "Read error while reading data");
	    break;
	case ngamsERR_FILE:
	    strcpy(statStr, "Invalid filename specified");
	    break;
	case ngamsERR_OPEN_LOG_FILE:
	    strcpy(statStr, "Could not open specified log file");
	    break;
	case ngamsSRV_OK:
	    strcpy(statStr, "Request sucessfully handled by server");
	    break;
	case ngamsSRV_INV_QUERY:
	    strcpy(statStr, "Invalid query");
	    break;
	default:
	    return ngamsERR_UNKNOWN_STAT;
	    break;
	}   /* end-switch (statNo) */
    return ngamsSTAT_SUCCESS;
}

/**
int ngamsSuccess(const ngamsSTAT *status)

Return 1 if the status object contains a response indicating success.

Returns:   1 = SUCCESS, 0 = FAILURE.
*/
int ngamsSuccess(const ngamsSTATUS *status)
{
    if (strcmp(status->status, NGAMS_SUCCESS) == 0)
	return 1;
    else
	return 0;
}

/**
char *ngamsStrRStr(const char *haystack, 
		   const char *needle)
Carry out a reverse string search.

haystack:   String to search.

needle:     Sub-string to look for.

Returns:    Pointer to first reverse occurrence of needle or NULL if not
            found.
*/
char *ngamsStrRStr(const char *haystack, 
		   const char *needle)

{
    int i = strlen(haystack);
    while (i >= 0)
	if (strstr((haystack + --i), needle) != NULL) break;
    if (i >= 0)
	return (char *)(haystack + i);
    else
	return NULL;
}

/**
void ngamsToUpper(char *str)

Convert a string to upper characters.
   
str:      Pointer to string to convert.

Returns:  Upper-cased string.
*/
void ngamsToUpper(char *str)
{
    int i;
    for (i = 0; i < strlen(str); i++)
        *(str + i) = (char)toupper((int)*(str + i));
}

/**
void ngamsTrimString(char  *str,
		     char  *trimChars)
Trim a string according to the given trim characters.

str:          String to be trimmed.

trimChars:    String buffer with the preceeding/trailing characters
              to trim away.

Returns:      Trimmed string.
*/
void ngamsTrimString(char  *str,
		     char  *trimChars)
{
    char  *tmpStr = NULL, *strPtr;

    /* Trim from left */
    strPtr = str;
    while ((*strPtr != '\0') && (strchr(trimChars, *strPtr) != NULL)) strPtr++;
    tmpStr = malloc(strlen(strPtr) + 1);
    strcpy(tmpStr, strPtr);

    /* Trim from right */
    strPtr = (tmpStr + strlen(tmpStr) - 1);
    while ((strPtr != tmpStr) && (strchr(trimChars, *strPtr) != NULL))
	*strPtr-- = '\0';
    strcpy(str, tmpStr);
    free(tmpStr);
}

/**
char *ngamsVersion(void)

Return pointer to buffer with the NG/AMS version information.

Returns:   Pointer to internal, static buffer containing the version
           for the NG/AMS C-API.
*/
char *ngamsVersion(void)
{
    static char  verBuf[64];
    /*char         buf1[16], buf2[16], buf3[16], buf4[16], date[16], time[16];
      char         *chrP = date;
    */

    /*
    sscanf(ngamsNGAMS_CVS_ID, "%s %s %s %s %s %s", 
	   buf1, buf2, buf3, buf4, date, time);
    while (*chrP != '\0') 
        {
        if (*chrP == '/') *chrP = '-';
	chrP++;
	}
    sprintf(verBuf, "%s/%sT%s", ngamsNGAMS_SW_VER, date, time);
    */
    sprintf(verBuf, "%s/%s", ngamsNGAMS_SW_VER, ngamsVER_DATE);
    return verBuf;
}


/**
*******************************************************************************
*/


/* EOF */

