#******************************************************************************
# ESO/DMD
#
# "@(#) $Id: ngamsDb.py,v 1.291 2005/09/23 14:38:08 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  07/05/2001  Created
#

"""
Contains the ngamsDb class.

Variables are defined to query all columns of the tables in a specific order.
These variables are defined below. An example of this is:


  from ngams import *
  import ngamsDb

  ...
  diskId = diskInfo[ngamsDb.NGAS_DISKS_DISK_ID]


In the docmentation of this class the following aliases are used for
the tables:

  nd = NGAS Disks Table (ngas_disks).
  nf = NGAS Files Table (ngas_files).
  nh = NGAS Hosts Table (ngas_hosts).
  ns = NGAS Subscribers Table (ngas_subscribers).
"""

# IMPL: Avoid usage of "noholdlock", this is not ANSI SQL.

# IMPL: - Split this module up in smaller modules.
#
#       - Review methods to query file info - possible to merge? Methods:
#           - dumpFileInfo()
#           - dumpFileInfoList()
#           - dumpFileSummary1()
#           - getFileInfoFromFileId()
#           - getFileInfoFromFileIdHostId()
#           - getFileInfoList()
#           - getFileSummary1()
#           - getFileSummary2()
#           - getFileSummarySpuriousFiles1()

import os, sys, string, time, types, cPickle, base64
import pcc, PccUtTime
from   ngams import *
import ngamsLib, ngamsDbm

# IMPL: Don't use these classes within this module. They are at a higher
#       level. See IMPL remark below.
import ngamsFileInfo, ngamsFileList, ngamsDiskInfo, ngamsStatus


# Global DB Semaphore to protect critical, global DB interaction.
_globalDbSem = threading.Semaphore(1)


# Define lay-out of ngas_disks table
_ngasDisksDef = [["nd.disk_id",               "NGAS_DISKS_DISK_ID"],
                 ["nd.archive",               "NGAS_DISKS_ARCHIVE"],
                 ["nd.logical_name",          "NGAS_DISKS_LOG_NAME"],
                 ["nd.host_id",               "NGAS_DISKS_HOST_ID"],
                 ["nd.slot_id",               "NGAS_DISKS_SLOT_ID"],
                 ["nd.mounted",               "NGAS_DISKS_MOUNTED"],
                 ["nd.mount_point",           "NGAS_DISKS_MT_PT"],
                 ["nd.number_of_files",       "NGAS_DISKS_NO_OF_FILES"],
                 ["nd.available_mb",          "NGAS_DISKS_AVAIL_MB"],
                 ["nd.bytes_stored",          "NGAS_DISKS_BYTES_STORED"],
                 ["nd.type",                  "NGAS_DISKS_TYPE"],
                 ["nd.manufacturer",          "NGAS_DISKS_MANUFACTURER"],
                 ["nd.installation_date",     "NGAS_DISKS_INST_DATE"],
                 ["nd.checksum",              "NGAS_DISKS_CHECKSUM"],
                 ["nd.total_disk_write_time", "NGAS_DISKS_TOT_WR_TIME"],
                 ["nd.completed",             "NGAS_DISKS_COMPLETED"],
                 ["nd.completion_date",       "NGAS_DISKS_COMPL_DATE"],
                 ["nd.last_check",            "NGAS_DISKS_LAST_CHECK"],
                 ["nd.last_host_id",          "NGAS_DISKS_LAST_HOST_ID"]]
_ngasDisksCols = ""
idx = 0
for colDef in _ngasDisksDef:
    if (_ngasDisksCols): _ngasDisksCols += ", "
    _ngasDisksCols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    idx += 1

# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._ngasDisksCols defines all the " +\
           "columns of the NGAS Disks Table. It has the following " +\
           "value:\n\n" + _ngasDisksCols
__doc__ += "\n\nWhen quering ngas_disks using ngamsDb._ngasDisksCol, " +\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _ngasDisksDef:
    __doc__ += "  %s\n" % colDef[1]


# Define lay-out of ngas_files table.
#
# NOTE: DON'T CHANGE THE SEQUENCE ALLOCATED - A NEW ENTRY CAN BE ADDED AT THE
#       END OF THE LIST, BUT CHANGING THE EXISTING ENTRIES WILL BREAK THE DB
#       SNAPSHOT FEATURE.
_ngasFilesDef = [["nf.disk_id",                "NGAS_FILES_DISK_ID"],
                 ["nf.file_name",              "NGAS_FILES_FILE_NAME"],
                 ["nf.file_id",                "NGAS_FILES_FILE_ID"],
                 ["nf.file_version",           "NGAS_FILES_FILE_VER"],
                 ["nf.format",                 "NGAS_FILES_FORMAT"],
                 ["nf.file_size",              "NGAS_FILES_FILE_SIZE"],
                 ["nf.uncompressed_file_size", "NGAS_FILES_UNCOMPR_FILE_SIZE"],
                 ["nf.compression",            "NGAS_FILES_COMPRESSION"],
                 ["nf.ingestion_date",         "NGAS_FILES_INGEST_DATE"],
                 ["nf.ignore",                 "NGAS_FILES_IGNORE"],
                 ["nf.checksum",               "NGAS_FILES_CHECKSUM"],
                 ["nf.checksum_plugin",        "NGAS_FILES_CHECKSUM_PI"],
                 ["nf.file_status",            "NGAS_FILES_FILE_STATUS"],
                 ["nf.creation_date",          "NGAS_FILES_CREATION_DATE"]]
_ngasFilesCols = ""
_ngasFilesNameMap = {}
idx = 0
for colDef in _ngasFilesDef:
    if (_ngasFilesCols): _ngasFilesCols += ", "
    _ngasFilesCols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    colName = colDef[0].split(".")[1]
    _ngasFilesNameMap[idx] = colName
    _ngasFilesNameMap[colName] = idx
    idx += 1

# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._ngasFilesCols defines all the " +\
           "columns of the NGAS Files Table. It has the following " +\
           "value:\n\n" + _ngasFilesCols
__doc__ += "\n\nWhen quering ngas_disks using ngamsDb._ngasFilesCol, " +\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _ngasFilesDef:
    __doc__ += "  %s\n" % colDef[1]


# Define lay-out of ngas_hosts table.
_ngasHostsDef = [["nh.host_id",              "NGAS_HOSTS_HOST_ID"],
                 ["nh.domain",               "NGAS_HOSTS_DOMAIN"],
                 ["nh.ip_address",           "NGAS_HOSTS_ADDRESS"],
                 ["nh.mac_address",          "NGAS_HOSTS_MAC_ADDRESS"],
                 ["nh.n_slots",              "NGAS_HOSTS_N_SLOTS"],
                 ["nh.cluster_name",         "NGAS_HOSTS_CLUSTER_NAME"],
                 ["nh.installation_date",    "NGAS_HOSTS_INST_DATE"],
                 ["nh.srv_version",          "NGAS_HOSTS_SRV_VER"],
                 ["nh.srv_port",             "NGAS_HOSTS_SRV_PORT"],
                 ["nh.srv_archive",          "NGAS_HOSTS_SRV_ARCHIVE"],
                 ["nh.srv_retrieve",         "NGAS_HOSTS_SRV_RETRIEVE"],
                 ["nh.srv_process",          "NGAS_HOSTS_SRV_PROCESS"],
                 ["nh.srv_remove",           "NGAS_HOSTS_SRV_REMOVE"],
                 ["nh.srv_data_checking",    "NGAS_HOSTS_SRV_DATA_CHECK"],
                 ["nh.srv_state",            "NGAS_HOSTS_SRV_STATE"],
                 ["nh.srv_suspended",        "NGAS_HOSTS_SRV_SUSP"],
                 ["nh.srv_req_wake_up_srv",  "NGAS_HOSTS_SRV_REQ_WAKE_UP_SRV"],
                 ["nh.srv_req_wake_up_time", "NGAS_HOSTS_SRV_REQ_WAKE_UP_TIM"]]
_ngasHostsCols = ""
idx = 0
for colDef in _ngasHostsDef:
    if (_ngasHostsCols): _ngasHostsCols += ", "
    _ngasHostsCols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    idx += 1

# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._ngasHostsCols defines all the " +\
           "columns of the NGAS Hosts Table. It has the following " +\
           "value:\n\n" + _ngasHostsCols
__doc__ += "\n\nWhen quering ngas_disks using ngamsDb._ngasHostsCols, " +\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _ngasHostsDef:
    __doc__ += "  %s\n" % colDef[1]


# Define lay-out of ngas_subscribers table.
_NGS = "NGAS_SUBSCR_"
_ngasSubscribersDef = [["ns.host_id",                   _NGS + "HOST_ID"],
                       ["ns.srv_port",                  _NGS + "SRV_PORT"],
                       ["ns.subscr_prio",               _NGS + "PRIO"],
                       ["ns.subscr_id",                 _NGS + "ID"],
                       ["ns.subscr_url",                _NGS + "URL"],
                       ["ns.subscr_start_date",         _NGS + "START"],
                       ["ns.subscr_filter_plugin",      _NGS + "FILT_PI"],
                       ["ns.subscr_filter_plugin_pars", _NGS + "FILT_PI_PARS"],
                       ["ns.last_file_ingestion_date",  _NGS + "ING_DATE"]]
_ngasSubscribersCols = ""
idx = 0
for colDef in _ngasSubscribersDef:
    if (_ngasSubscribersCols): _ngasSubscribersCols += ", "
    _ngasSubscribersCols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    idx += 1

# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._ngasSubscribersCols defines all the " +\
           "columns of the NGAS Subscribers Table. It has the following " +\
           "value:\n\n" + _ngasSubscribersCols
__doc__ += "\n\nWhen quering ngas_disks using ngamsDb._ngasSubscribersCols, "+\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _ngasSubscribersDef:
    __doc__ += "  %s\n" % colDef[1]


# Define lay-out of a Summary 1 Query.
_sum1Def = [["nd.slot_id",         "SUM1_SLOT_ID"],
            ["nd.mount_point",     "SUM1_MT_PT"],
            ["nf.file_name",       "SUM1_FILENAME"],
            ["nf.checksum",        "SUM1_CHECKSUM"],
            ["nf.checksum_plugin", "SUM1_CHECKSUM_PI"],
            ["nf.file_id",         "SUM1_FILE_ID"],
            ["nf.file_version",    "SUM1_VERSION"],
            ["nf.file_size",       "SUM1_FILE_SIZE"],
            ["nf.file_status",     "SUM1_FILE_STATUS"],
            ["nd.disk_id",         "SUM1_DISK_ID"],
            ["nf.ignore",          "SUM1_FILE_IGNORE"],
            ["nd.host_id",         "SUM1_HOST_ID"]]
_sum1Cols = ""
idx = 0
for colDef in _sum1Def:
    if (_sum1Cols): _sum1Cols += ", "
    _sum1Cols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    idx += 1
    
# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._sum1Cols defines the columns " +\
           "to query for s File Summary 1 Query. It has the following " +\
           "value:\n\n" + _sum1Cols
__doc__ += "\n\nWhen executing a File Summary 1 Query using " +\
           "ngamsDb._sum1Cols, "+\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _sum1Def:
    __doc__ += "  %s\n" % colDef[1]


# Define lay-out of a Summary 2 Query.
_sum2Def = [["nf.file_id",        "SUM2_FILE_ID"],
            ["nd.mount_point",    "SUM2_MT_PT"],
            ["nf.file_name",      "SUM2_FILENAME"],
            ["nf.file_version",   "SUM2_VERSION"],
            ["nf.ingestion_date", "SUM2_ING_DATE"],
            ["nf.format",         "SUM2_MIME_TYPE"]]
_sum2Cols = ""
idx = 0
for colDef in _sum2Def:
    if (_sum2Cols): _sum2Cols += ", "
    _sum2Cols += colDef[0]
    exec(colDef[1] + "=%d" % idx)
    idx += 1
  
# Update documentation.
__doc__ += "\n\nThe variable ngamsDb._sum2Cols defines the columns " +\
           "to query for s File Summary 2 Query. It has the following " +\
           "value:\n\n" + _sum2Cols
__doc__ += "\n\nWhen executing a File Summary 2 Query using " +\
           "ngamsDb._sum2Cols, "+\
           "the resulting (sub-)list can be accessed by means of the " +\
           "variables:\n\n"
for colDef in _sum2Def:
    __doc__ += "  %s\n" % colDef[1]


class ngamsDb:
    """
    Class to handle the connection to the NGAS DB.
    """
    
    # For test purposes. If set to 1 (e.g. in a DAPI) a DB exception will be
    # thrown. Can be used e.g. to provoke back-log buffering of data.
    provokeErr = 0

  
    def __init__(self,
                 server,
                 db,
                 user,
                 password,
                 createSnapshot = 1,
                 interface = "ngamsSybase",
                 tmpDir="/tmp"):
        """
        Constructor method.

        server:          DB server name (string).

        db:              DB name (string).
        
        user:            DB user (string).
        
        password:        DB password (string).

        createSnapshot:  Indicates if a DB Snapshot (temporary snapshot
                         files) should be created (integer/0|1).

        interface:       NG/AMS DB Interface Plug-In (string).
        
        tmpDir:          Name of NGAS Temporary Directory (string).
        """
        self.__dbDrv = None

        # Semaphore to protect critical DB interaction.
        self.__dbSem = threading.Lock() # threading.Semaphore(1)

        # Controls if the snapshot of the DB should be created.
        self.__createSnapshot = createSnapshot

        # List of Event Objects which are used to inform other instances
        # about changes in the DB.
        self.__dbChangeEvents = []

        # Timer for analyzing time spent for DB access
        self.__dbTimer      = PccUtTime.Timer()
        self.__dbAccessTime = 0.0
        self.resetTimers()

        # Import the DB Interface Plug-In + create connection.
        self.__dbServer    = server
        self.__dbName      = db
        self.__dbUser      = user
        self.__dbPassword  = password
        self.__dbSnapshot  = createSnapshot
        self.__dbInterface = interface
        self.connect(server, db, user, password, interface)

        # Verification/Auto Recover.
        self.__dbVerify      = 1
        self.__dbAutoRecover = 0
       
        self.__dbTmpDir      = tmpDir
       

    def takeDbSem(self):
        """
        Acquire access to a critical DB interaction.

        Returns:   Reference to object itself (ngamsDb).
        """
        info(5,"Taking DB Access Semaphore")
        self.__dbSem.acquire()


    def relDbSem(self):
        """
        Release acquired access to a critical DB interaction.

        Returns:   Reference to object itself (ngamsDb).
        """
        info(5,"Releasing DB Access Semaphore")
        self.__dbSem.release()

        
    def takeGlobalDbSem(self):
        """
        Acquire access to a critical, global DB interaction.

        Returns:   Reference to object itself (ngamsDb).
        """
        # IMPL: Check if really needed with a Global DB Semaphore (if yes
        #       write the reason in the documentation).
        info(5,"Taking Global DB Access Semaphore")
        global _globalDbSem
        _globalDbSem.acquire()


    def relGlobalDbSem(self):
        """
        Release acquired access to a critical, global DB interaction.

        Returns:   Reference to object itself (ngamsDb).
        """
        # IMPL: Check if really needed with a Global DB Semaphore (if yes
        #       write the reason in the documentation).
        info(5,"Releasing Global DB Access Semaphore")
        global _globalDbSem
        _globalDbSem.release()


    def resetTimers(self):
        """
        Reset internal read and write timers.

        Returns:   Reference to object itself (ngamsDb).
        """
        self.__rdTime = 0.0
        self.__wrTime = 0.0
        return self


    def getRdWrTime(self):
        """
        Return time spent for reading/writing DB.

        Returns:  Tuple with read/write time in seconds (tuple/floats)
        """
        return (self.__rdTime, self.__wrTime)
    

    def _startTimer(self):
        """
        Start the DB read timer.

        Returns:   Reference to object itself (ngamsDb).
        """
        if (getVerboseLevel() < 3): return self
        self.__dbAccessTime = 0.0
        self.__dbTimer.start()
        return self
    

    def _storeRdTime(self):
        """
        Get the time since the DB access timer was reset the last time
        and add this to the DB Read Time timer.

        Returns:    Total read time in seconds (or 0) (float).
        """
        if (getVerboseLevel() < 3): return 0
        self.__dbAccessTime = self.__dbTimer.stop()
        self.__rdTime += self.__dbAccessTime
        return self.__rdTime
    

    def _storeWrTime(self):
        """
        Get the time since the DB access timer was reset the last time
        and add this to the DB Write Time timer.

        Returns:    Total write time in seconds (or 0) (float).
        """
        if (getVerboseLevel() < 3): return 0
        self.__dbAccessTime = self.__dbTimer.stop()
        self.__wrTime += self.__dbAccessTime
        return self.__wrTime


    def getDbTime(self):
        """
        Return the time spent for the last DB access.

        Returns:    Last DB access time in seconds (float).
        """
        return self.__dbAccessTime
    

    def getDbTimeStr(self,
                     prec = 3):
        """
        Return the time spent for the last DB access as a string.

        Returns:    Last DB access time in seconds as a string (string).
        """
        return str('%.' + str(prec) + 'fs') % self.__dbAccessTime


    def getDbServer(self):
        """
        Returns the name of the DB server.

        Returns:  Name of DB server (string).
        """
        return self.__dbServer


    def setDbVerify(self,
                    verify):
        """
        Enable/disable DB Verification.

        verify:   0=off, 1=on (integer/0|1).

        Returns:  Reference to object itself (ngamsDb).
        """
        self.__dbVerify = int(verify)
        return self


    def setDbAutoRecover(self,
                         autoRecover):
        """
        Enable/disable DB Auto Recovering.

        autoRecover:   0=off, 1=on (integer/0|1).

        Returns:       Reference to object itself (ngamsDb).        
        """
        self.__dbAutoRecover = int(autoRecover)
        return self


    def setDbTmpDir(self,
                    tmpDir):
        """
        Set the DB temporary directory.

        tmpDir:        Temporary directory (string).

        Returns:       Reference to object itself (ngamsDb).        
        """
        self.__dbTmpDir = tmpDir
        return self


    def connect(self,
                server,
                db,
                user,
                password,
                interface = "ngamsSybase"):
        """
        Connect to the DB.

        server:     DB server name (string).

        db:         DB name (string).
        
        user:       DB user (string).
        
        password:   DB password (base 64 encoded) (string).

        interface:  NG/AMS DB Interface Plug-In (string).

        Returns:    Void.
        """
        try:
            self.takeDbSem()
            self._connect(server, db, user, password, interface, 0)
            self.relDbSem()
        except Exception, e:
            self.relDbSem()
            raise Exception, e
        

    def _connect(self,
                 server,
                 db,
                 user,
                 password,
                 interface,
                 recurseLevel):
        """
        Connect to the DB. The method may do this in a recursive manner
        (re-trying) in case it fails in connecting.

        NOTE: The invocation of this method should be semaphore protected.

        server:       DB server name (string).

        db:           DB name (string).
        
        user:         DB user (string).
        
        password:     DB password (base 64 encoded) (string).

        interface:    NG/AMS DB Interface Plug-In (string).

        recurseLevel: Recursive depth level (integer).

        Returns:      Void.
        """
        info(4,"Entering _connect() ...")
        try:
            if (self.__dbDrv != None):
                del self.__dbDrv
                self.__dbDrv = None

            info(4,"Importing DB Driver Interface: %s" % interface)
            exec "import " + interface
            try:
                password = base64.decodestring(password)
            except Exception, e:
                errMsg = "Incorrect, encrypted DB password given. Error: " +\
                         str(e)
                raise Exception, errMsg
            creStat = "%s.%s('%s', '%s', '%s', '%s', '%s')" %\
                      (interface, interface, server, db, user, password,
                       "NG/AMS:" + getThreadName())
            info(4,"Creating instance of DB Driver Interface/connecting ...")
            self.__dbDrv = eval(creStat)            
            info(3,"DB Driver Interface ID: " + self.__dbDrv.getDriverId())
        except Exception, e:
            errMsg = genLog("NGAMS_ER_DB_COM", ["Problem setting up " +\
                                                "DB connection: " + str(e)])
            errMsg = errMsg.replace("\n", "")

            # Retry up to 5 times.
            if (recurseLevel < 5):
                warning(errMsg)
                msg = "Reconnecting to DB server. Recursive depth: %d/" +\
                      "Test Mode: %d ..."
                notice(msg % (recurseLevel, getTestMode()))
                if (getTestMode()):
                    time.sleep(0.5)
                else:
                    time.sleep(5.0)
                recurseLevel += 1
                self._connect(self.__dbServer, self.__dbName, self.__dbUser,
                              self.__dbPassword, self.__dbInterface,
                              recurseLevel)
                notice("Reconnected to DB server")
            else:
                notice("Abandonned reconnectiong attempt. Recurse level: %d" %\
                       recurseLevel)
                raise Exception, errMsg
        info(4,"Leaving _connect()")


    def checkCon(self):
        """
        Method that check if the DB connection seems to be OK. If it is
        not, it will be tried to reconnect. If the connection is down, and
        it cannot reconnect, an exception is raised.

        Returns:    Void.
        """
        # Perform a simply query, the ngamsDb.query() method will reconnect
        # automatically if the DB connection is lost
        info(4, "Entering checkCon() ...")
        sqlQuery = "SELECT * from ngas_hosts"
        res = self.query(sqlQuery)
        self._storeRdTime()
        info(4, "Leaving checkCon(). DB Time: " + self.getDbTimeStr())


    def close(self):
        """
        Close the DB connection.

        Returns:    Void.
        """
        try:
            self.takeDbSem()
            if (self.__dbDrv): self.__dbDrv.close()
            self.relDbSem()
        except Exception, e:
            self.relDbSem()
            raise Exception, e            
        

    def addDbChangeEvt(self,
                       evtObj):
        """
        Add an Event Object (threading.Event) that will be triggered to
        indicate other threads that DB changes where introdiced.

        evtObj:        Event object (threading.Event).

        Returns:       Reference to object itself (ngamsDb).
        """
        self.__dbChangeEvents.append(evtObj)
        return self


    def triggerEvents(self,
                      eventInfo = None):
        """
        Set the Event Objects to inform other threads about DB changes.

        eventInfo:    Piece of information to be transferred from one
                      thread to another (free format).

        Returns:      Reference to object itself (ngamsDb).
        """
        try:
            self.takeDbSem()
            for evtObj in self.__dbChangeEvents:
                if (eventInfo): evtObj.addEventInfo(eventInfo)
                if (not evtObj.isSet()): evtObj.set()
            self.relDbSem()
            return self
        except Exception, e:
            self.relDbSem()
            raise Exception, e     


    def query(self,
              sqlQuery,
              ignoreEmptyRes = 1):
        """
        Perform a query in the DB and return the result.

        sqlQuery:         SQL query (string).

        ignoreEmptyRes:   If set to 1, no error will be assumed if a
                          completely empty result is returned (integer/0|1).

        Returns:          Result of SQL query (list).
        """
        try:
            self.takeDbSem()
            res = self._query(sqlQuery, ignoreEmptyRes, 0)
            self.relDbSem()
            return res
        except Exception, e:
            self.relDbSem()
            raise Exception, e


    def _query(self,
               sqlQuery,
               ignoreEmptyRes,
               recurseLevel):
        """
        Perform a query in the DB and return the result. The method may do
        this in a recursive manner (re-trying) in case it fails in executing
        the query. Also, if the connection is lost, it will be tried
        automatically to reconnect.

        NOTE: The invocation of this method should be semaphore protected.

        sqlQuery:         SQL query (string).

        ignoreEmptyRes:   If set to 1, no error will be assumed if a
                          completely empty result is returned (integer/0|1).

        recurseLevel:     Indicates the number of times this method was
                          called recursively (integer).

        Returns:          Result of SQL query (list).
        """
        info(4,"Entering _query() ...")
        info(5,"Performing SQL query: " + str(sqlQuery))

        #####################################################################
        # Make it possible to control a simulated DB error externally if
        # Test Mode is enabled.
        #####################################################################
        if (getTestMode()):
            fo = None
            try:
                fo = open("/tmp/ngamsDbError.tmp")
                provokeDbErr = int(fo.read())
                fo.close()
            except:
                try:
                    if (fo): fo.close()
                except:
                    pass
                provokeDbErr = 0
            # Simulate a DB communication problem if requested.
            if (self.provokeErr or provokeDbErr):
                errMsg = genLog("NGAMS_ER_DB_COM",
                                ["Error: connection is not open"])
                raise Exception, errMsg
        #####################################################################
            
        res = []
        self._startTimer()
        try:
            res = self.__dbDrv.query(sqlQuery)
        except Exception, e:
            errMsg = genLog("NGAMS_ER_DB_COM", [str(e)])
            error(errMsg)

            # In case of an exception: Try to reconnect if the Recursive
            # Level is less than 10.
            if (recurseLevel < 10):
                msg = "Reconnecting to DB server. Recursive depth: %d/" +\
                      "Test Mode: %d ..."
                notice(msg % (recurseLevel, getTestMode()))
                if (getTestMode()):
                    time.sleep(0.2)
                else:
                    time.sleep(1.0)
                recurseLevel += 1
                self._connect(self.__dbServer, self.__dbName, self.__dbUser,
                              self.__dbPassword, self.__dbInterface, 0)
                notice("Reconnected to DB server")
                res = self._query(sqlQuery, ignoreEmptyRes, recurseLevel)
            else:
                notice("Abandonned reconnectiong attempt. Recurse level: %d" %\
                       recurseLevel)
                raise Exception, errMsg

        # Return the query result, or try up to 10 times to perform the query
        # with a 1s delay between each attempt. If all attempts to query the DB
        # fail, return an error in case the result was empty and such a result
        # should be reported as an error.
        info(5,"Result of SQL query (" + str(sqlQuery) + "): " + str(res))
        if ((res == []) and (not ignoreEmptyRes) and
            (sqlQuery.find("SELECT ") != -1)):
            msg = "Unexpected result returned from SQL query: %s: %s"
            warning(msg % (str(sqlQuery), str(res)))
            if (recurseLevel < 10):
                msg = "Retrying query: %s/Recursive depth: %d/" +\
                      "Test Mode: %d ..."
                notice(msg % (sqlQuery, recurseLevel, getTestMode()))
                if (getTestMode()):
                    time.sleep(0.1)
                else:
                    time.sleep(0.5)
                recurseLevel += 1
                res = self._query(sqlQuery, ignoreEmptyRes, recurseLevel)
            else:
                raise Exception, msg % (str(sqlQuery), str(res))
        info(4,"Leaving _query()")
        return res


    def dbCursor(self,
                 sqlQuery):
        """
        Create a DB Cursor Object (defined in the NG/AMS DB Interface Plug-In)
        on the given query and return the cursor object.

        sqlQuery:        SQL query for the cursor (string).

        Return:          Cursor object instance (<Cursor Object>).
        """
        self._startTimer()
        try:
            self.takeDbSem()
            curObj = self.__dbDrv.cursor(sqlQuery)
            self.relDbSem()
        except Exception, e:
            self.relDbSem()
            errMsg = genLog("NGAMS_ER_DB_COM", [str(e)])
            error(errMsg)
            raise Exception, errMsg
        return curObj


    def diskInDb(self,
                 diskId):
        """
        Check if disk with the given Disk ID is available in the DB.

        diskId:    Disk ID (string).
        
        Returns:   1 = Disk ID was found, 0 = Disk ID not found (integer).
        """
        info(4, "Entering diskInDb() ...")
        sqlQuery = "SELECT disk_id FROM ngas_disks WHERE disk_id='"+diskId +"'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving diskInDb(). DB Time: " + self.getDbTimeStr())
        if (res == [[]]):
            return 0
        else:
            if (res[0][0][0] == diskId):
                return 1
            else:
                return 0


    def fileInDb(self,
                 diskId,
                 fileId,
                 fileVersion = -1):
        """
        Check if file with the given File ID is registered in NGAS DB
        in connection with the given Disk ID.
        
        diskId:        Disk ID (string)

        fileId:        File ID (string).
        
        fileVersion:   Version of the file. If -1 version is not taken
                       into account (integer).

        Returns:       1 = file found, 0 = file no found (integer).
        """
        info(4, "Entering fileInDb() ...")
        if (fileVersion != -1):
            sqlQuery = "SELECT file_id FROM ngas_files WHERE file_id='" +\
                       fileId + "' AND disk_id='" + diskId + "' AND " +\
                       "file_version=" + str(fileVersion)
        else:
            sqlQuery = "SELECT file_id FROM ngas_files WHERE file_id='" +\
                       fileId + "' AND disk_id='" + diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving fileInDb(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0] == fileId):
                return 1
            else:
                return 0
        else:
            return 0
        

    def subscriberInDb(self,
                       subscrId):
        """
        Check if the Subscriber with the given ID is registered in the DB.
    
        subscrId:    Subscriber ID (string).
        
        Returns:     1 = Subscriber registered, 0 = Subscriber not
                     registered (integer).
        """
        info(4, "Entering subscriberInDb() ...")
        sqlQuery = "SELECT subscr_id FROM ngas_subscribers WHERE " +\
                   "subscr_id='" + subscrId + "'"
        res = self.query(sqlQuery)
        self._storeRdTime()
        info(4, "Leaving subscriberInDb(). DB Time: " + self.getDbTimeStr())
        if (res == [[]]):
            return 0
        else:
            if (res[0][0][0] == subscrId):
                return 1
            else:
                return 0


    def setLogicalNameForDiskId(self,
                                diskId,
                                logicalName):
        """
        Change the Logical Name of the disk with the given Disk ID.

        diskId:        Disk ID (string).
        
        logicalName:   New Logical Name (string).

        Returns:       Void.
        """
        info(4, "Entering setLogicalNameForDiskId() ...")
        sqlQuery = "UPDATE ngas_disks SET logical_name='" + logicalName +"' "+\
                   "WHERE disk_id='" + diskId + "'"
        self.query(sqlQuery)
        self._storeWrTime()
        info(4, "Leaving setLogicalNameForDiskId() ")


    def getLogicalNameFromDiskId(self,
                                 diskId):
        """
        Query the Logical Name of a disk from the DB, based on
        the Disk ID of the disk.
        
        diskId:      Disk ID (string).    
    
        Returns:     Logical Name or None if not found (string | None).
        """
        info(4, "Entering getLogicalNameFromDiskId() ...")
        sqlQuery = "SELECT logical_name FROM ngas_disks WHERE disk_id='" +\
                   diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getLogicalNameFromDiskId(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 1):
            return res[0][0][0]
        else:
            return None


    def getDiskCompleted(self,
                         diskId):
        """
        Check if a disk is marked in the NGAS DB as completed.

        diskId:    ID of the disk (string).

        Returns:   1 = completed, 0 = not completed. If the disk
                   is not registered None is returned (integer).
        """
        info(4, "Entering getDiskCompleted() ...")
        sqlQuery = "SELECT completed FROM ngas_disks WHERE " +\
                   "disk_id = '" + diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskCompleted(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 0):
            return None
        else:
            return res[0][0][0]


    def getSlotIdFromDiskId(self,
                            diskId):
        """
        Get the Slot ID for a disk, given by the Disk ID for the disk.

        diskId:    ID of the disk (string).

        Returns:   Slot ID of disk. If disk is not found None is
                   returned (string | None).
        """
        info(4, "Entering getSlotIdFromDiskId() ...")
        sqlQuery = "SELECT slot_id FROM ngas_disks WHERE disk_id='" +\
                   diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSlotIdFromDiskId(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 0):
            return None
        else:
            return res[0][0][0]


    def getDiskInfoFromDiskId(self,
                              diskId):
        """
        Query the information for one disk (referred to by its ID), and
        return this in raw format.

        The information about the disk is stored in a list with the
        the lay-out following the sequence as listed in the overall man-page
        for the ngamsDb class.

        diskId:   ID of the disk (string).

        Returns:  Disk information for the disk or [] if disk was
                  not found (list).
        """
        info(4, "Entering getDiskInfoFromDiskId() ...")
        sqlQuery = "SELECT " + _ngasDisksCols + " " +\
                   "FROM ngas_disks nd WHERE disk_id='" + diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskInfoFromDiskId(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            return res[0][0]


    def getDiskInfoFromDiskIdList(self,
                                  diskIdList):
        """
        Get disk information from a list of Disk IDs given. The result is
        returned in a list containing again lists with the field as described
        in documentation for ngamsDb.getDiskInfoFromDiskId().

        diskIdList:   List with Disk IDs (list/string).

        Returns:      List with disk information (list/list).
        """
        info(4, "Entering getDiskInfoFromDiskIdList() ...")
        sqlQuery = "SELECT " + _ngasDisksCols + " " +\
                   "FROM ngas_disks nd WHERE disk_id IN ("
        if (len(diskIdList)):
            tmpDiskIdDic = {}
            for id in diskIdList:
                tmpDiskIdDic[id] = 0
            for diskId in tmpDiskIdDic.keys():
                sqlQuery += "'" + diskId + "', "
            sqlQuery = sqlQuery[0:-2]
        else:
            return []
        sqlQuery += ")"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskInfoFromDiskIdList(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            return res[0]


    def getMaxDiskNumber(self,
                         cat = None):
        """
        Get the maximum disk index (number) in connection with the
        Logical Disk Names in the DB.

        cat:       'M' for Main, 'R' for Replication (string).

        Returns:   The maximum disk number or None if this could not
                   be generated (integer).
        """
        info(4, "Entering getMaxDiskNumber() ...")
        sqlQuery = "SELECT max(right(logical_name, 6)) FROM ngas_disks"
        if (cat):
            sqlQuery += " WHERE logical_name LIKE '%" + cat + "-%'"
        else:
            sqlQuery += " WHERE logical_name LIKE '%M-%' or " +\
                        "logical_name LIKE '%R-%'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getMaxDiskNumber(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 0):
            return None
        else:
            if (res[0][0][0] == None):
                return None
            else:
                try:
                    return int(res[0][0][0])
                except:
                    errMsg = "Illegal index for Logical Name (Disk Label) "+\
                             "encountered: " + str(res[0][0][0])
                    raise Exception, errMsg


    def writeDiskEntry(self,
                       diskId,
                       archive,
                       installationDate,
                       type,
                       manufacturer,
                       logicalName,
                       hostId,
                       slotId,
                       mounted,
                       mountPoint,
                       numberOfFiles,
                       availableMb,
                       bytesStored,
                       completed,
                       completionDate,
                       checksum,
                       totalDiskWriteTime,
                       lastCheck,
                       lastHostId):
        """
        The method writes the information in connection with a disk in the
        NGAS DB. If an entry already exists for that disk, it is updated
        with the information contained in the Disk Info Object. Otherwise,
        a new entry is created.

        diskId
        ...
        lastHostId:      Values for the columns in the ngas_disks
                         table (use values returned from ngamsDiskInfo).

        Returns:         Returns 1 if a new entry was created in the DB
                         and 0 if an existing entry was updated (integer/0|1).
        """
        try:
            info(4, "Entering writeDiskEntry() ...")

            # Check if the entry already exists. If yes update it, otherwise
            # insert a new element.
            try:
                self.takeDbSem()
                instDate = self.__dbDrv.convertTimeStamp(installationDate)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e

            #bytesStored = string.split(str(bytesStored), ".")[0]
            bytesStored = "%.0f" % float(bytesStored)
            if (self.diskInDb(diskId)):
                sqlQuery = "UPDATE ngas_disks SET " +\
                           "archive='" + archive + "', " +\
                           "installation_date='" + instDate + "', " +\
                           "type='" + type + "', " +\
                           "manufacturer='" + manufacturer + "', " +\
                           "logical_name='" + logicalName + "', " +\
                           "host_id='" + hostId + "', " +\
                           "slot_id='" + slotId + "', " +\
                           "mounted=" + str(mounted) + ", " +\
                           "mount_point='" + mountPoint + "', " +\
                           "number_of_files=" + str(numberOfFiles) + ", "+\
                           "available_mb=" + str(availableMb) + ", " +\
                           "bytes_stored=" + str(bytesStored) + ", " +\
                           "completed=" + str(completed) + ", " +\
                           "checksum='" + checksum + "', " +\
                           "total_disk_write_time=" + str(totalDiskWriteTime)

                if (lastCheck != ""):
                    try:
                        self.takeDbSem()
                        lastCheckTmp = self.__dbDrv.convertTimeStamp(lastCheck)
                        self.relDbSem()
                    except Exception, e:
                        self.relDbSem()
                        raise Exception, e      
                        
                    sqlQuery += ", last_check='" + lastCheckTmp + "'"

                if (lastHostId != None):
                    sqlQuery += ", last_host_id='" + lastHostId + "'"

                sqlQuery += " WHERE disk_id='" + diskId + "'"
                addDiskHistEntry = 0
            else:
                sqlQuery = "INSERT INTO ngas_disks " +\
                           "(disk_id, archive, installation_date, " +\
                           "type, manufacturer, logical_name, " +\
                           "host_id, slot_id, mounted, mount_point, "+\
                           "number_of_files, available_mb, bytes_stored, " +\
                           "completed, checksum, total_disk_write_time, " +\
                           "last_host_id) VALUES " +\
                           "('" + diskId + "', " +\
                           "'" + archive + "', " +\
                           "'" + instDate + "', " +\
                           "'" + type + "', " +\
                           "'" + manufacturer + "', " +\
                           "'" + logicalName + "', " +\
                           "'" + hostId + "', " +\
                           "'" + slotId + "', " +\
                           str(mounted) + ", " +\
                           "'" + mountPoint + "', " +\
                           str(numberOfFiles) + ", " +\
                           str(availableMb) + ", " +\
                           str(bytesStored) + ", " +\
                           str(completed) + ", " +\
                           "'" + checksum + "', " +\
                           str(totalDiskWriteTime) + ", " +\
                           "'" + lastHostId + "')"
                addDiskHistEntry = 1
            res = self.query(sqlQuery)
            self._storeWrTime()

            if (completionDate != ""):
                try:
                    self.takeDbSem()
                    complDate = self.__dbDrv.convertTimeStamp(completionDate)
                    self.relDbSem()
                except Exception, e:
                    self.relDbSem()
                    raise Exception, e                   
                sqlQuery = "UPDATE ngas_disks SET completion_date='" +\
                            complDate + "' " + "WHERE disk_id='" + diskId + "'"
                self.query(sqlQuery)
                self._storeWrTime()

            self.triggerEvents([diskId, mountPoint])
            info(4, "Leaving writeDiskEntry(). DB Time: "+self.getDbTimeStr())
            return addDiskHistEntry
        except Exception, e:   
            raise e

            
    def getLogicalNamesMountedDisks(self,
                                    host):
        """
        Get the Logical Names for the disks mounted on the given host.
        A list is returned, which contains the Logical Names of the disks
        mounted.

        host:      Name of host where the disk must be mounted (string).

        Returns:   List with Logical Names (list).
        """
        info(4, "Entering getLogicalNamesMountedDisks() ...")
        sqlQuery = "SELECT logical_name FROM ngas_disks WHERE host_id='" +\
                   host + "' AND mounted=1"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getLogicalNamesMountedDisks(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 0):
            return []
        else:
            logNames = []
            for disk in res[0]:
                logNames.append(disk[0])
            return logNames


    def getMtPtFromDiskId(self,
                          diskId):
        """
        Get the mount point for the disk referred to.

        diskId:      ID of the disk (string).

        Returns:     Mount point of disk or None if not mounted or
                     not found (string|None)
        """
        info(4, "Entering getMtPtFromDiskId() ...")
        sqlQuery = "SELECT mount_point FROM ngas_disks WHERE disk_id='" +\
                   diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getMtPtFromDiskId(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 0):
            return None
        else:
            return res[0][0][0]


    def getDiskIdsMtPtsMountedDisks(self,
                                    host):
        """
        Get the mount points for the disks mounted on the given host.
        A list is returned, which contains the Logical Names of the disks
        mounted.

        host:      Name of host where the disk must be mounted (string).

        Returns:   List with tuples containing Disk IDs and Mount
                   Points (list/tuple).
        """
        info(5, "Entering getDiskIdsMtPtsMountedDisks() ...")
        sqlQuery = "SELECT disk_id, mount_point FROM ngas_disks " +\
                   "WHERE host_id='" + host + "' AND mounted=1"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(5, "Leaving getDiskIdsMtPtsMountedDisks(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 0):
            return []
        else:
            return res[0]


    def getSlotIdsMountedDisks(self,
                               host):
        """
        Get the Slot IDs for the disks mounted on the given host. A list is
        returned, which contains the Slot IDs of the disks mounted.

        host:      Name of host where the disk must be mounted (string).

        Returns:   List with Slot IDs (list).
        """
        info(4, "Entering getSlotIdsMountedDisks() ...")
        sqlQuery = "SELECT slot_id FROM ngas_disks WHERE host_id='" +\
                   host + "' AND mounted=1"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSlotIdsMountedDisks(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 0):
            return []
        else:
            slotIds = []
            for disk in res[0]:
                slotIds.append(disk[0])
            return slotIds


    def getDiskIdsMountedDisks(self,
                               host,
                               mtRootDir):
        """
        Get the Disk IDs for the disks mounted on the given host. A list is
        returned, which contains the Disk IDs of the disks mounted.

        host:        Name of host where the disk must be mounted (string).

        mtRootDir:   Base directory for NG/AMS (string).

        Returns:     List with Disk IDs (list).
        """
        info(4, "Entering getDiskIdsMountedDisks() ...")
        if (mtRootDir[-1] != "/"): mtRootDir += "/"
        mtRootDir += "%"
        sqlQuery = "SELECT disk_id FROM ngas_disks WHERE host_id='" +\
                   host + "' AND mounted=1 AND mount_point LIKE '" +\
                   mtRootDir + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskIdsMountedDisks(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            diskIds = []
            for disk in res[0]:
                diskIds.append(disk[0])
            return diskIds


    def getDiskIdFromSlotId(self,
                            host,
                            slotId):
        """
        Get a Disk ID for corresponding Slot ID and host name.

        host:     Host name (string).

        slotId:   ID of slot (string).

        Returns:  Disk ID or None if no match found (string).
        """
        info(4, "Entering getDiskIdFromSlotId() ...")
        sqlQuery = "SELECT disk_id FROM ngas_disks WHERE slot_id='" +\
                   slotId + "' AND host_id='" + host + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskIdFromSlotId(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return None
        else:
            return res[0][0][0]


    def getDiskIds(self):
        """
        Query the Disk IDs contained in the NGAS DB and return these in a list.

        Returns:    List with Disk IDs (list).
        """
        info(4, "Entering getDiskIds() ...")
        sqlQuery = "SELECT disk_id FROM ngas_disks"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskIds(). DB Time: " + self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            retList = []
            for diskId in res[0]:
                retList.append(diskId[0])
            return retList


    def getDiskInfoForSlotsAndHost(self,
                                   host,
                                   slotIdList):
        """
        From a given host and a given list of Slot IDs, the method
        returns a list with the disk info for the disks matching these.

        host:         Host name where the disks considered must be
                      mounted (string).

        slotIdList:   List of Slot IDs for the disk considered (list).

        Returns:      List with Disk Info objects  or [] if no matches
                      were found (list/ngamsDiskInfo).
        """
        info(4, "Entering getDiskInfoForSlotsAndHost() ...")
        if (slotIdList):
            slotIds = "("
            for id in slotIdList: slotIds = slotIds + "'" + id + "',"
            slotIds = slotIds[0:-1] + ")"
        sqlQuery = "SELECT " + _ngasDisksCols + " FROM ngas_disks nd WHERE "
        if (slotIdList): sqlQuery += " nd.slot_id IN " + slotIds + " AND "
        sqlQuery += "nd.host_id='" + host + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getDiskInfoForSlotsAndHost(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            return res[0]


    def getBestTargetDisk(self,
                          diskIds,
                          mtRootDir):
        """
        Find the best suitable target disk among a set of disks referred
        to by their Disk IDs. The condition is:

        Get ID for disk that is most full, which is Main Disk, mounted, has
        a Host ID corresponding to this host ID, which is not completed,
        which has the lowest Slot ID and which has the Disk ID among the set
        of disks defined for the given mime-type.

        diskIds:     List with Disk IDs to probe for (list).

        mtRootDir:   Base diretory for the NG/AMS Server (string).

        Returns:     Disk ID or None if no matches were found (string)
        """
        info(4,"Entering getBestTargetDisk() ...")
        # Create a list of IDs for the SQL statement.
        ids = "("
        for id in diskIds:
            ids = ids + "'" + id + "',"
        ids = ids[0:-1] + ")"

        if (mtRootDir[-1] != "/"): mtRootDir += "/"
        mtRootDir += "%"

        # Get ID for best suitable disk.
        sqlQuery = "SELECT disk_id FROM ngas_disks WHERE " +\
                   "completed=0 AND disk_id IN " + ids + " " +\
                   "AND mount_point LIKE '" + mtRootDir + "' " +\
                   "ORDER BY bytes_stored desc, installation_date asc"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime() 
        info(4, "Leaving getBestTargetDisk(). DB Time: " + self.getDbTimeStr())
        if (res == [[]]):
            return None
        else:
            return res[0][0][0]
        

    def getFileInfoFromFileIdHostId(self,
                                    hostId,
                                    fileId,
                                    fileVersion = 1,
                                    diskId = None,
                                    ignore = None):
        """
        Return list with information about a certain file referenced
        by its File ID. A list is returned with the following elements:

          [<Disk ID>, <Filename>, <File ID>, <File Version>, <Format>,
           <File Size>, <Uncompressed File Size>, <Compression>, 
           <Ingestion Date>, <Ignore>, <Checksum>, <Checksum Plug-In>,
           <File Status>]

        hostId:           Name of host where the disk is mounted on
                          which the file is stored (string).

        fileId:           ID for file to acquire information for (string).

        fileVersion:      Version of the file to query information
                          for (integer).

        diskId:           Used to refer to a specific disk (string).

        ignore:           If set to 0 or 1, this value of ignore will be
                          queried for. If set to None, ignore is not
                          considered (None|0|1).

        Returns           List with information about file, or [] if
                          no file(s) was found (list).
        """
        info(4, "Entering getFileInfoFromFileIdHostId() ...")
        sqlQuery = "SELECT " + _ngasFilesCols + " " +\
                   "FROM ngas_files nf, ngas_disks nd WHERE " +\
                   "nf.file_id='" + fileId + "' AND " +\
                   "nf.disk_id=nd.disk_id AND " +\
                   "nd.host_id='" + hostId + "' AND nd.mounted=1 AND " +\
                   "nf.file_version=" + str(fileVersion)
        if (diskId): sqlQuery += " AND nd.disk_id='%s'" % diskId
        if (ignore != None): sqlQuery += " AND nf.ignore=%d" % int(ignore)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getFileInfoFromFileIdHostId(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 0):
            return []
        else:
            return res[0][0]

    
    def writeFileEntry(self,
                       diskId,
                       filename,
                       fileId,
                       fileVersion,
                       format,
                       fileSize,
                       uncompressedFileSize,
                       compression,
                       ingestionDate,
                       ignore,
                       checksum,
                       checksumPlugIn,
                       fileStatus,
                       creationDate,
                       genSnapshot = 1,
                       updateDiskInfo = 0):
        """
        The method writes the information in connection with a file in the
        NGAS DB. If an entry already exists for that file, it is updated
        with the information contained in the File Info Object. Otherwise,
        a new entry is created.

        diskId           Values for the columns in the ngas_disks
        ...              table (use values returned from ngamsFileInfo).

        genSnapshot:     Generate a snapshot file (integer/0|1).

        updateDiskInfo:  Update automatically the disk info for the
                         disk hosting this file (integer/0|1).

        Returns:         Void.
        """
        try:
            info(5, "Entering writeFileEntry() ...")
 
            # Check if the entry already exists. If yes update it, otherwise
            # insert a new element.
            if (ignore == -1): ignore = 0
            try:
                self.takeDbSem()
                ingDate = self.__dbDrv.convertTimeStamp(ingestionDate)
                creDate = self.__dbDrv.convertTimeStamp(creationDate)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e
            if (self.fileInDb(diskId, fileId, fileVersion)):
                # We only allow to modify a limited set of columns.
                sqlQuery = "UPDATE ngas_files SET " +\
                           "file_name='" + filename + "', " +\
                           "format='" + format + "', " +\
                           "file_size=" + str(fileSize) + ", " +\
                           "uncompressed_file_size=" +\
                           str(uncompressedFileSize) + ", " +\
                           "compression='" + compression + "', " +\
                           "ingestion_date='" + ingDate + "', " +\
                           "ignore=" + str(ignore) + ", " +\
                           "checksum='" + checksum + "', " +\
                           "checksum_plugin='" + checksumPlugIn + "', " +\
                           "file_status='" + fileStatus + "', " +\
                           "creation_date='" + creDate + "' " +\
                           "WHERE file_id='" + fileId + "' AND " +\
                           "disk_id='" + diskId + "'"
                if (int(fileVersion) != -1):
                    sqlQuery += " AND file_version=" + str(fileVersion)
                dbOperation = NGAMS_DB_CH_FILE_UPDATE
            else:
                sqlQuery = "INSERT INTO ngas_files " +\
                           "(disk_id, file_name, file_id, file_version, " +\
                           "format, file_size, " +\
                           "uncompressed_file_size, compression, " +\
                           "ingestion_date, ignore, checksum, " +\
                           "checksum_plugin, file_status, creation_date) "+\
                           "VALUES " +\
                           "('" + diskId + "', " +\
                           "'" + filename + "', " +\
                           "'" + fileId + "', " +\
                           "" + str(fileVersion) + ", " +\
                           "'" + format + "', " +\
                           str(fileSize) + ", " +\
                           str(uncompressedFileSize) + ", " +\
                           "'" + compression + "', " +\
                           "'" + ingDate + "', " +\
                           str(ignore) + ", " +\
                           "'" + checksum + "', " +\
                           "'" + checksumPlugIn + "', " +\
                           "'" + fileStatus + "', " +\
                           "'" + creDate + "')"
                dbOperation = NGAMS_DB_CH_FILE_INSERT

            # Perform the main query.
            self.query(sqlQuery)
            self._storeWrTime()

            # Update the Disk Info of the disk concerned if requested and
            # if a new entry was added.
            #
            # Note: In case of an update the columns ngas_disks.avail_mb
            #       and ngas_disks.bytes_stored should in principle be
            #       updated according to the actual size of the new
            #       version of the file.
            if (updateDiskInfo and (dbOperation == NGAMS_DB_CH_FILE_INSERT)):
                self._updateDiskFileStatus(diskId, fileSize)

            # Create a Temporary DB Change Snapshot Document if requested.
            if (self.__createSnapshot and genSnapshot):
                tmpFileObj = ngamsFileInfo.ngamsFileInfo().\
                             unpackSqlResult([diskId, filename, fileId,
                                              fileVersion, format, fileSize,
                                              uncompressedFileSize,compression,
                                              ingestionDate, ignore, checksum,
                                              checksumPlugIn, fileStatus,
                                              creationDate])
                self.createDbFileChangeStatusDoc(dbOperation, [tmpFileObj])
                del tmpFileObj

            self.triggerEvents([diskId, None])

            info(5, "Leaving writeFileEntry(). DB Time: "+self.getDbTimeStr())
        except Exception, e:   
            raise e


    def getFileInfoFromFileId(self,
                              fileId,
                              fileVersion = -1,
                              diskId = None,
                              ignore = None,
                              dbCursor = 1):
        """
        The method queries the file information for the files with the given
        File ID and returns the information found in a list containing
        sub-lists each with a list with the information for the file from the
        ngas_files table, host ID and mount point. The following rules are
        applied when determining which files to return:

          o All files are considered, also files which are Offline.

 	  o Files marked to be ignored are ignored.
            
          o Latest version - first priority.

        It is possible to indicate if files marked as being 'bad' in the
        DB should be taken into account with the 'ignoreBadFiles' flag.

        If a specific File Version is specified only that will be
        taken into account.

        The data can be retrieved via the DB Cursor returned by this object.
        The format of each sub-result is:

          [<see getFileInfoFromFileIdHostId()>, <host ID>, <mnt pt>]
        

        fileId:          File ID for file to be retrieved (string).

        fileVersion:     If a File Version is specified only information
                         for files with that version number and File ID
                         are taken into account. The version must be a
                         number in the range [1; oo[ (integer).

        diskId:          ID of disk where file is residing. If specified
                         to None (or empty string) the Disk ID is not taken
                         into account (string).

        ignore:          If set to 0 or 1, this value of ignore will be
                         queried for. If set to None, ignore is not
                         considered (None|0|1).

        dbCursor:        If set to 1, a DB cursor is returned from which
                         the files can be retrieved. Otherwise the result
                         is queried and returned in a list (0|1/integer).
                         
        Returns:         Cursor object or list with results
                         (<NG/AMS DB Cursor Object API>|list).
        """
        info(4,"Entering getFileInfoFromFileId() ...")
        try:
            int(fileVersion)
        except:
            raise Exception, "Illegal value for File Version specified: " +\
                  str(fileVersion)

        # Query for files being online.
        sqlQuery = "SELECT " + _ngasFilesCols + ", nd.host_id, " +\
                   "nd.mount_point FROM ngas_files nf, ngas_disks nd, "+\
                   "ngas_hosts nh WHERE " +\
                   "nh.host_id=nd.host_id AND " +\
                   "nf.disk_id=nd.disk_id"
        if (ignore != None): sqlQuery != " AND nf.ignore=%d" % int(ignore)
        # File ID specified.
        if (fileId): sqlQuery += " AND nf.file_id='%s'" % fileId
        # Do we want a specific File Version?
        if (fileVersion != -1):
            sqlQuery += " AND nf.file_version=" + str(fileVersion)
        # Is a special disk referred?
        if (diskId): sqlQuery += " AND nf.disk_id='%s'" % diskId

        # Order the files according to the version.
        sqlQuery += " ORDER BY file_version desc"

        if (dbCursor):
            # Carry out query and return the DB cursor object.
            curObj = self.dbCursor(sqlQuery)
            info(4, "Leaving getFileInfoFromFileId(). DB Time: " +\
                 self.getDbTimeStr())
            return curObj
        else:
            # Execute the query directly and return the result.
            res = self.query(sqlQuery)
            self._storeRdTime()
            if (len(res) > 0):
                return res[0]
            else:
                return []


    def getFileInfoList(self,
                        diskId,
                        fileId = "",
                        fileVersion = -1,
                        ignore = None):
        """
        The function queries a set of files matching the conditions
        specified in the input parameters.

        diskId:        Disk ID of disk hosting the file(s) (string).
        
        fileId:        File ID of files to consider. Wildcards can be
                       used (string).
        
        fileVersion:   Version of file(s) to consider. If set to -1 this
                       is not taken into account (integer).

        ignore:        If set to 0 or 1, this value of ignore will be
                       queried for. If set to None, ignore is not
                       considered (None|0|1).

        Returns:       Cursor object (<NG/AMS DB Cursor Object API>).
        """
        info(4, "Entering getFileInfoList() ...")
        fileId = re.sub("\*", "%", fileId)
        sqlQuery = "SELECT " + _ngasFilesCols + " FROM ngas_files nf WHERE"

        if (fileId.find("%") != -1):
            fileIdStatement = " nf.file_id LIKE '" + fileId + "'"
        elif (fileId != ""):
            fileIdStatement = " nf.file_id='" + fileId + "'"
        else:
            fileIdStatement = ""

        if (diskId != ""):
            diskIdStatement = " nf.disk_id='" + diskId + "'"
        else:
            diskIdStatement = ""

        if (fileIdStatement): sqlQuery += fileIdStatement
        if (diskIdStatement and fileIdStatement):
            sqlQuery += " AND" + diskIdStatement
        elif (diskIdStatement):
            sqlQuery += diskIdStatement
        if (fileVersion != -1):
            sqlQuery += " AND nf.file_version=" + str(fileVersion)
        if (ignore != None): sqlQuery += " AND nf.ignore=%d" % int(ignore)

        # Create a cursor and perform the query.
        curObj = self.dbCursor(sqlQuery)
        
        info(4, "Leaving getFileInfoList(). DB Time: " + self.getDbTimeStr())
        return curObj


    def _dumpFileInfo(self,
                      fileId,
                      fileVersion,
                      diskId,
                      ignore,
                      fileInfoDbmName,
                      expMatches):
        """
        See ngamsDb.dumpFileInfo().
        """
        info(4,"Entering _dumpFileInfo() ...")
        if (not fileInfoDbmName):
            fileInfoDbmName = "/tmp/" +\
                              ngamsLib.genUniqueFilename("FILE_INFO_DB")
        rmFile(fileInfoDbmName + "*")
        fileInfoDbm = None
        dbCount = 0
        dbCur = None
        try:
            fileInfoDbm = ngamsDbm.ngamsDbm(fileInfoDbmName, cleanUpOnDestr=0,
                                            writePerm = 1)

            # If there are more than a 100 expected matches, a cursor is used
            # otherwise the result is queried directly.
            if (expMatches > 100):
                dbCur = self.getFileInfoFromFileId(fileId, fileVersion, diskId,
                                                   ignore)
                while (1):
                    tmpFileInfoList = dbCur.fetch(1000)
                    if (tmpFileInfoList == []): break
                    for tmpFileInfo in tmpFileInfoList:
                        tmpFileInfoObj =\
                                       ngamsFileInfo.ngamsFileInfo().\
                                       unpackSqlResult(tmpFileInfo)
                        newEl = [tmpFileInfoObj, tmpFileInfo[-2],
                                 tmpFileInfo[-1]]
                        fileInfoDbm.add(str(dbCount), newEl)
                        dbCount += 1
                    fileInfoDbm.sync()
                    time.sleep(0.050)
                del dbCur
            else:
                res = self.getFileInfoFromFileId(fileId,fileVersion,
                                                 diskId, ignore, dbCursor=0)
                for tmpFileInfo in res:
                    tmpFileInfoObj = ngamsFileInfo.ngamsFileInfo().\
                                     unpackSqlResult(tmpFileInfo)
                    newEl = [tmpFileInfoObj, tmpFileInfo[-2],
                             tmpFileInfo[-1]]
                    fileInfoDbm.add(str(dbCount), newEl)
                    dbCount += 1
                fileInfoDbm.sync()
            del fileInfoDbm
            info(4, "Leaving _dumpFileInfo()")
            return fileInfoDbmName
        except Exception, e:
            error("Exception: " + str(e))
            del fileInfoDbm
            del dbCur
            rmFile(fileInfoDbmName + "*")
            raise e


    def dumpFileInfo(self,
                     fileId,
                     fileVersion = -1,
                     diskId = "",
                     ignore = None,
                     fileInfoDbmName = ""):
        """
        The method queries the file information for the files with the given
        File ID and returns the information found in a list containing
        sub-lists each with a ngamsFileInfo object, host ID and mount point.
        The following rules are applied when determining which files to return:

            o Files marked to be ignored are ignored.
            o Only files that are marked as being hosted on hosts,
              which are Online or on suspended disks/hosts, are considered.
            o Latest version - first priority.

        It is possible to indicate if files marked as being 'bad' in the
        DB should be taken into account with the 'ignoreBadFiles' flag.

        The file information is referred to by a string key in the interval
        [0; oo[ (integer in string format).

        If a specific File Version is specified only that will be taken into
        account.

        fileId:          File ID for file to be retrieved (string).

        fileVersion:     If a File Version is specified only information
                         for files with that version number and File ID
                         are taken into account. The version must be a
                         number in the range [1; oo[ (integer).

        diskId:          ID of disk where file is residing. If specified
                         to '' (empty string) the Disk ID is not taken
                         into account (string).
                         
        ignore:          If set to 0 or 1, this value of ignore will be
                         queried for. If set to None, ignore is not
                         considered (None|0|1).

        fileInfoDbmName: If given, this will be used as name for the file info
                         DB (string).

        Returns:         Name of a BSD DB file in which the File Info Objects
                         are stored. This DB contains pickled objects pointed
                         to by an index number ([0; oo[). The contents of each
                         of these pickled objects is:

                           [<File Info Obj>, <Host ID>, <Mount Point>]

                         An element NGAMS_FILE_DB_COUNTER indicates the
                         number of files stored in the DB (string/filename).
        """
        info(4,"Entering dumpFileInfo() ...")

        # Try first to get the expected number of files, which will be returned
        expNoOfFiles = -1
        if (self.__dbVerify):
            for n in range(1):
                noOfFiles = self.getNumberOfFiles(diskId, fileId, fileVersion,
                                                  ignore, onlyOnlineFiles=1)
                if (noOfFiles > expNoOfFiles): expNoOfFiles = noOfFiles
                time.sleep(0.050)

        # Try up to 5 times to dump the files.
        for n in range(5):
            fileInfoDbmName = self._dumpFileInfo(fileId, fileVersion, diskId,
                                                 ignore, fileInfoDbmName,
                                                 expNoOfFiles)
            # Check the number of files for which info was dumped.
            tmpFileInfoDbm = ngamsDbm.ngamsDbm(fileInfoDbmName)
            dumpedNoOfFiles = tmpFileInfoDbm.getCount()
            del tmpFileInfoDbm

            # Print out DB Verification Warning if actual number of files
            # differs from expected number of files.
            if (self.__dbVerify and (dumpedNoOfFiles != expNoOfFiles)):
                errMsg = "Problem dumping file info! Expected number of "+\
                         "files: %d, actual number of files: %d"
                errMsg = errMsg % (expNoOfFiles, dumpedNoOfFiles)
                warning(errMsg)

            # Try to Auto Recover if requested.
            if ((self.__dbVerify and self.__dbAutoRecover) and
                (dumpedNoOfFiles != expNoOfFiles)):
                rmFile(fileInfoDbmName + "*")
                if (n < 4):
                    # We try again, after a small pause.
                    if (getTestMode()):
                        time.sleep(0.5)
                    else:
                        time.sleep(5)
                    notice("Retrying to dump file info ...")
                else:
                    errMsg = "Giving up to auto recover dumping of file info!"
                    error(errMsg)
                    raise Exception, errMsg
            else:
                # All files were dumped, we stop.
                break

        info(4,"Leaving dumpFileInfo()")
        return fileInfoDbmName


    def dumpFileSummary1(self,
                         fileInfoDbmName = None,
                         hostId = None,
                         diskIds = [],
                         fileIds = [],
                         ignore = None,
                         fileStatus = [NGAMS_FILE_STATUS_OK],
                         lowLimIngestDate = None,
                         order = 1):
        """
        Dump the summary of the files defined by the parameters. This is done
        in a safe manner (or at least attempted) such that in case of problems
        with the DB interaction it is retried to dump the info.

        The file info is dumped into a ngamsDbm DB.

        fileInfoDbmName: Name of DBM, which will contain the info about the
                         files (string).

        For the other parameters check man-page for: ngamsDb.getFileSummary1().

        Returns:         Name of the DBM DB containing the info about the files
                        (string).
        """
        info(4,"Entering dumpFileSummary1() ...")
    
        # Try first to get the expected number of files, which will be returned
        expNoOfFiles = -1
        if (self.__dbVerify):
            for n in range(1):
                noOfFiles = self.getSummary1NoOfFiles(hostId, diskIds, fileIds,
                                                      ignore, fileStatus,
                                                      lowLimIngestDate)
                if (noOfFiles > expNoOfFiles): expNoOfFiles = noOfFiles
                time.sleep(0.050)

        # Create a temporay File Info DBM.
        if (not fileInfoDbmName):
            fileInfoDbmName = "/tmp/" +\
                              ngamsLib.genUniqueFilename("FILE-SUMMARY1")

        # Try up to 5 times to dump the files. A retry is done if the number
        # of records dumped differs from the expected number.
        for n in range(5):
            fileInfoDbm = ngamsDbm.ngamsDbm(fileInfoDbmName, 0, 1)
            curObj = self.getFileSummary1(hostId, diskIds, fileIds, ignore,
                                          fileStatus, lowLimIngestDate, order)
            fileCount = 0
            while (1):
                fileList = curObj.fetch(1000)
                if (not fileList): break
                for fileInfo in fileList:
                    fileInfoDbm.add(str(fileCount), fileInfo)
                    fileCount += 1
            del curObj

            # Print out DB Verification Warning if actual number of files
            # differs from expected number of files.
            if (self.__dbVerify and (fileCount != expNoOfFiles)):
                errMsg = "Problem dumping file info! Expected number of "+\
                         "files: %d, actual number of files: %d"
                errMsg = errMsg % (expNoOfFiles, fileCount)
                warning(errMsg)            

            # Try to Auto Recover if requested.
            if ((self.__dbVerify and self.__dbAutoRecover) and
                (fileCount != expNoOfFiles)):
                del fileInfoDbm
                rmFile(fileInfoDbmName + "*")
                # Not all files were dumped.
                if (n < 4):
                    # - retry after a small pause.
                    if (getTestMode()):
                        time.sleep(0.5)
                    else:
                        time.sleep(5)
                    notice("Retrying to dump file info ...")
                else:
                    errMsg = "Giving up to auto recover dumping of file info!"
                    error(errMsg)
                    raise Exception, errMsg
            else:
                del fileInfoDbm
                break

        info(4,"Leaving dumpFileSummary1()")
        return fileInfoDbmName


    def dumpFileInfoList(self,
                         diskId,
                         fileId = "",
                         fileVersion = -1,
                         ignore = None,
                         fileListDbmName = ""):
        """
        The function queries the same info as getFileInfoList(). However,
        rathen than returning a cursor object, it retrieves the info itself
        and stores this in a DBM. This is done in a 'safe manner', whereby
        it is tried to recover from the situation where less files are returned
        than expected.

        ignore:             If set to 0 or 1, this value of ignore will be
                            queried for. If set to None, ignore is not
                            considered (None|0|1).
                       
        fileListDbmName:    Name of DBM where to store the queried info
                            (string).

        Returns:            Complete name of DBM containing the requested
                            info (string).
        """
        info(4,"Entering dumpFileInfoList() ...")

        # Retrieve the theoretical number of files the query should produce.
        # Try first to get the expected number of files, which will be returned
        expNoOfFiles = -1
        if (self.__dbVerify):
            for n in range(1):
                noOfFiles = self.getNumberOfFiles(diskId, fileId, fileVersion,
                                                  ignore)
                if (noOfFiles > expNoOfFiles): expNoOfFiles = noOfFiles
                time.sleep(0.050)

        # Generate final name of DBM + create DBM.
        if (not fileListDbmName):
            fileListDbmName = self.__dbTmpDir + "/" +\
                              ngamsLib.genUniqueFilename("_FILE_INFO")

        # Retrieve/dump the files (up to 5 times).
        curObj = None
        fileListDbm = None
        for n in range(5):
            try:
                fileListDbm = ngamsDbm.ngamsDbm(fileListDbmName,
                                                cleanUpOnDestr=0, writePerm=1)
                fileListDbmName = fileListDbm.getDbmName()
                dbCur = self.getFileInfoList(diskId, fileId, fileVersion,
                                             ignore)
                fileCount = 0
                while (1):
                    res = dbCur.fetch(100)
                    if (not res): break
                    for fileInfo in res:
                        fileListDbm.add(str(fileCount), fileInfo)
                        fileCount += 1
                    time.sleep(0.010)
                fileListDbm.sync()
            except Exception, e:
                rmFile(fileListDbmName + "*")
                if (curObj): del curObj
                if (fileListDbm): del fileListDbm
                raise Exception, e
                
            if (curObj): del dbCur

            # Print out DB Verification Warning if actual number of files
            # differs from expected number of files.
            if (self.__dbVerify and (fileCount != expNoOfFiles)):
                errMsg = "Problem dumping file info! Expected number of "+\
                         "files: %d, actual number of files: %d"
                errMsg = errMsg % (expNoOfFiles, fileCount)
                warning(errMsg)            
            
            # Try to Auto Recover if requested.
            if ((self.__dbVerify and self.__dbAutoRecover) and
                (fileCount != expNoOfFiles)):
                if (fileListDbm): del fileListDbm
                rmFile(fileListDbmName + "*")
                if (n < 4):
                    if (getTestMode()):
                        time.sleep(0.5)
                    else:
                        time.sleep(5)
                    notice("Retrying to dump file info ...")
                else:
                    errMsg = "Giving up to auto recover dumping of file info!"
                    error(errMsg)
                    raise Exception, errMsg
            else:
                break

        info(4,"Leaving dumpFileInfoList()")
        return fileListDbmName

    
    def getHostInfoFromHostIds(self,
                               hostList):
        """
        Return a dictionary with the information in connection with each host.
        If for a host ID given, no information is found in ngas_hosts, the
        value for this wil be None.

        hostList:    List of host IDs (list/string).

        Returns:     List with sub-lists containing the information about the
                     hosts from ngas_hosts. See ngamsDb._ngasHostsDef (list).
        """
        info(4,"Entering getHostInfoFromHostIds() ...")
        hostDic = {}
        sqlQuery = "SELECT " + _ngasHostsCols +\
                   " FROM ngas_hosts nh WHERE host_id IN ("
        for host in hostList:
            sqlQuery += "'" + host + "', "
        sqlQuery = sqlQuery[0:-2] + ")"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getHostInfoFromHostIds(). DB Time: " +\
             self.getDbTimeStr())
        return res[0]


    def getIpFromHostId(self,
                        hostId):
        """
        Get the IP Address of a host from its Host ID.

        hostId:     Host ID (string).

        Returns:    IP Address (string).
        """
        sqlQuery = "SELECT ip_address FROM ngas_hosts " +\
                   "WHERE host_id='" + hostId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (len(res[0]) == 1):
            return res[0][0][0]
        else:
            errMsg = "Error retrieving IP Address for host: " + hostId
            raise Exception, errMsg


    def getSrvSuspended(self,
                        contactAddr,
                        ngasHostId = None):
        """
        Return flag indicating if the server is suspended.

        contactAddr:  Host ID or IP address (string).

        ngasHostId:   NGAS Host ID, e.g. myhost:8888 (string).

        Returns:      Server suspension flag (integer/0|1).
        """
        if (ngasHostId):
            sqlQuery = "SELECT srv_suspended FROM ngas_hosts " +\
                       "WHERE host_id='" + ngasHostId + "'"
        else:
            sqlQuery = "SELECT srv_suspended FROM ngas_hosts " +\
                       "WHERE host_id='" + contactAddr + "' " +\
                       "OR ip_address='" + contactAddr + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (len(res[0]) == 1):
            try:
                return int(res[0][0][0])
            except:
                return 0
        else:
            errMsg = "Error retrieving Server Suspended Flag for host: "+hostId
            raise Exception, errMsg


    def getSrvDataChecking(self,
                           hostId):
        """
        Return flag indicating if server is executing a Data Consistency Check.

        hostId:     Host ID (string).

        Returns:    Server suspension flag (integer/0|1).
        """
        sqlQuery = "SELECT srv_data_checking FROM ngas_hosts " +\
                   "WHERE host_id='" + hostId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (len(res[0]) == 1):
            if (res[0][0][0]):
                return int(res[0][0][0])
            else:
                return 0
        else:
            # If we get to this point, the entry was not found.
            errMsg = "Error retrieving Data Checking Flag - host: " + hostId
            raise Exception, errMsg


    def getHostContactInfo(self,
                           hostId):
        """
        Get the information needed to be able to contact a host. If the
        host is located within a cluster the Host ID/IP + Port Number of the
        Cluster Main Node is returned. In addition is returned whether or not
        the contact host is suspended.

        hostId:   ID of host to analyze (string).
        
        Returns:  Returns a tuple with the information:

                  (<Contact Host ID>, <Contact Host IP>, <Contact Port>,
                   <Suspended>)  (tuple).
        """
        info(4,"Entering getHostContactInfo() ...")
        sqlQuery = "SELECT host_id, ip_address, srv_port, srv_suspended " +\
                   "FROM ngas_hosts WHERE host_id=(SELECT cluster_name " +\
                   "FROM ngas_hosts WHERE host_id='" + hostId + "')"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (len(res[0]) == 1):
            # If the contact host corresponds to the local host, the contact
            # information for the host specified in the method invocation
            # must be queried and returned (since it can contact the
            # specified host directly).
            if (res[0][0][0] == getHostId()):
                sqlQuery = "SELECT host_id, ip_address, srv_port, " +\
                           "srv_suspended FROM ngas_hosts " +\
                           "WHERE host_id='" + hostId + "'"
                res = self.query(sqlQuery, ignoreEmptyRes=0)
                self._storeRdTime()
                if (len(res[0]) != 1):
                    errMsg = "Error retrieving info for host: " + hostId
                    raise Exception, errMsg 
            return (res[0][0][0], res[0][0][1], int(res[0][0][2]),res[0][0][3])
        else:
            errMsg = "Error retrieving info for Cluster Main Node " +\
                     "in connection with host: " + hostId
            raise Exception, errMsg


    def updateSrvHostInfo(self,
                          hostId,
                          srvInfo,
                          ignoreErr = 0):
        """
        Update the information in the DB, which is managed by the server
        itself. All columns starting with 'srv_' in the ngas_hosts tables
        are defined. The values can be taken from an instance of the
        ngamsHostInfo class.

        srvInfo:    List containing all information about the host. These are
                    all fields starting with 'srv_' in the _ngasHostsDef list
                    from 'srv_version' to 'srv_state' (list).

        ignoreErr:  If set to 1, a possible exception thrown will be
                    caught, and this error ignored. Otherwise the
                    method will throw an exception itself (integer/0|1).

        Returns:    Void.
        """
        try:
            info(5, "Entering updateSrvHostInfo() ...")
            sqlQuery = "UPDATE ngas_hosts SET " +\
                       "srv_version='" + srvInfo[0] + "', " +\
                       "srv_port=" + str(srvInfo[1]) + ", " +\
                       "srv_archive=" + str(srvInfo[2]) + ", " +\
                       "srv_retrieve=" + str(srvInfo[3])+ ", " +\
                       "srv_process=" + str(srvInfo[4]) + ", " +\
                       "srv_remove=" + str(srvInfo[5]) + ", " +\
                       "srv_data_checking=" + str(srvInfo[6]) + ", " +\
                       "srv_state='" + srvInfo[7] + "' " +\
                       "WHERE host_id='" + hostId + "'"
            if (not ignoreErr):
                self.query(sqlQuery)
                self._storeWrTime()
            else:
                try:
                    self.query(sqlQuery)
                    self._storeWrTime()
                except:
                    pass
            self.triggerEvents()
            info(5, "Leaving updateSrvHostInfo(). DB Time: " +\
                 self.getDbTimeStr())
        except Exception, e:   
            raise e
            

    def getLatestFileVersion(self,
                             fileId):
        """
        The method queries the latest File Version for the file with the given
        File ID. If a file with the given ID does not exist, -1 is returned.

        fileId:    File ID (string).

        Returns:   Latest File Version (integer).
        """
        info(4, "Entering getLatestFileVersion() ...")
        sqlQuery = "SELECT max(file_version) FROM ngas_files WHERE "+\
                   "file_id='" + fileId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getLatestFileVersion(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0])):
            if (res[0][0][0] == None):
                return -1
            else:
                return int(res[0][0][0])
        else:
            return -1


    def buildFileSummary1Query(self,
                               hostId = None,
                               diskIds = [],
                               fileIds = [],
                               ignore = None,
                               fileStatus = [NGAMS_FILE_STATUS_OK],
                               lowLimIngestDate = None,
                               order = 1):
        """
        Builds the SQL query for a File Summary1 query. The fields to be
        selected are left open (specified as %s).

        For a description of the input parameters, check the man-page of
        ngamsDb.getFileSummary1().
        
        Returns:    SQL query for a File Summary 1 Query (string).
        """
        info(5,"Entering buildFileSummary1Query() ...")
        sqlQuery = "SELECT %s FROM ngas_disks nd noholdlock, " +\
                   "ngas_files nf noholdlock " +\
                   "WHERE nd.disk_id=nf.disk_id"

        # Additional WHERE clauses.
        if (ignore != None): sqlQuery += " AND nf.ignore=%d" % int(ignore)
        if (hostId): sqlQuery += " AND nd.host_id='" + hostId + "'"
        if (diskIds != []):
            sqlQuery += " AND nd.disk_id IN (" + str(diskIds)[1:-1] + ")"
        if (fileIds != []):
            sqlQuery += " AND nf.file_id IN (" + str(fileIds)[1:-1] + ")"
        if (fileStatus != []):
            sqlQuery += " AND nf.file_status IN (" + str(fileStatus)[1:-1]+ ")"
        if (lowLimIngestDate):
            try:
                self.takeDbSem()
                lowLimDate = self.__dbDrv.convertTimeStamp(lowLimIngestDate)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e
            sqlQuery += " AND nf.ingestion_date >= '" + lowLimDate + "'"
        if (order): sqlQuery += " ORDER BY nd.slot_id, nf.ingestion_date"

        info(5,"Leaving buildFileSummary1Query()")
        return sqlQuery


    def getSummary1NoOfFiles(self,
                             hostId = None,
                             diskIds = [],
                             fileIds = [],
                             ignore = None,
                             fileStatus = [NGAMS_FILE_STATUS_OK],
                             lowLimIngestDate = None):
        """
        Get the theoretical number of files that a Summary 1 Query would
        return with the given parameters.

        For a description of the input parameters, check the man-page of
        ngamsDb.getFileSummary1().

        Returns:   Number of files query would return (integer).
        """
        info(4,"Entering getSummary1NoOfFiles() ...")
        sqlQuery = self.buildFileSummary1Query(hostId, diskIds, fileIds,
                                               ignore, fileStatus,
                                               lowLimIngestDate,
                                               order=0) % "count(*)"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSummary1NoOfFiles(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0]):
                return res[0][0][0]
            else:
                return 0
        else:
            return 0


    def getFileSummary1(self,
                        hostId = None,
                        diskIds = [],
                        fileIds = [],
                        ignore = None,
                        fileStatus = [NGAMS_FILE_STATUS_OK],
                        lowLimIngestDate = None,
                        order = 1):
       
        """
        Return summary information about files. The information is returned
        in a list containing again sub-lists with contents as defined
        by ngamsDb._sum1Cols (see general documentation of the ngamsDb Class.

        hostId:            Name of NGAS host on which the files reside
                           (string).

        diskIds:           Used to limit the query to certain disks
                           (list/string).

        fileIds:           List of file IDs for which to query information.
                           If not specified, all files of the referenced
                           host will be chosen (list/string|[]).

        ignore:            If set to 0 or 1, this value of ignore will be
                           queried for. If set to None, ignore is not
                           considered (None|0|1).

        fileStatus:        With this parameter it is possible to indicate which
                           files to consider according to their File Status
                           (list).

        lowLimIngestDate:  Lower limit in time for which files are taken into
                           account. Only files with an Ingestion Date after
                           this date, are taken into account (string/ISO8601).

        order:             Used to trigger ordering by Slot ID + Ingestion Date
                           (integer/0|1).

        Returns:           Cursor object (<NG/AMS DB Cursor Object API>).
        """
        info(5,"Entering getFileSummary1() ...")
        sqlQuery = self.buildFileSummary1Query(hostId, diskIds, fileIds,
                                               ignore, fileStatus, 
                                               lowLimIngestDate,
                                               order) % _sum1Cols
        # Create a cursor and perform the query.
        curObj = self.dbCursor(sqlQuery)
        info(5, "Leaving getFileSummary1(). DB Time: " + self.getDbTimeStr())
        return curObj

    
    # IMPL: 1. Consider to merge this method together with
    #          ngamsDb.getFileSummary1 (i.e., to obtain one:
    #          ngamsDb.getFileSummary()).
    #       2. Make safe layer on top of this method like it is done for
    #          ngamsDb.getFileSummary1() (see ngamsb,umpFileSummary1()).
    def getFileSummary2(self,
                        hostId = None,
                        fileIds = [],
                        diskId = None,
                        ignore = None):
        """
        Return summary information about files. An NG/AMS DB Cursor Object
        is created, which can be used to query the information sequentially.

        The information is returned in a list containing again sub-lists
        with contents as defined by ngamsDb._sum2Cols (see general
        documentation of the ngamsDb Class.

        This method returns all the files stored on an NGAS system also
        the ones with a File Status indicating that it is bad.

        hostId:            Name of NGAS host on which the files reside. If
                           None is specified, the host is not taken into
                           account (string).

        fileIds:           List of file IDs for which to query information.
                           If not specified, all files of the referenced
                           host will be chosen (list/string|[]).

        diskId:            Used to refer to all files on a given disk
                           (string|None).

        ignore:            If set to 0 or 1, this value of ignore will be
                           queried for. If set to None, ignore is not
                           considered (None|0|1).

        Returns:           Cursor object (<NG/AMS DB Cursor Object API>).
        """
        info(4, "Entering getFileSummary2() ...")
        sqlQuery = "SELECT " + _sum2Cols + " " +\
                   "FROM ngas_disks nd noholdlock, " +\
                   "ngas_files nf noholdlock " +\
                   "WHERE nd.disk_id=nf.disk_id "
        if (ignore != None): sqlQuery += "AND nf.ignore=%d" % int(ignore)
        if (hostId): sqlQuery += " AND nd.host_id='" + hostId + "'"
        if (diskId): sqlQuery += " AND nf.disk_id='" + diskId + "'"
        if (fileIds != []):
            sqlQuery += " AND nf.file_id IN (" + str(fileIds)[1:-1] + ")"
        sqlQuery += " ORDER BY nf.ingestion_date"

        # Create a cursor and perform the query.
        curObj = self.dbCursor(sqlQuery)
        
        info(4, "Leaving getFileSummary2(). DB Time: " + self.getDbTimeStr())
        return curObj
        

    # IMPL: Make safe layer on top of this method like it is done for
    #       ngamsDb.getFileSummary1() (see ngamsb,umpFileSummary1()).
    def getFileSummarySpuriousFiles1(self,
                                     hostId = None,
                                     diskId = None,
                                     fileId = None,
                                     fileVersion = None):
        """
        Return summary information about spurious files, i.e. files registered
        in the DB as to be ignored and/or having a status indicating that
        they're not OK. The information is returned in a list containing
        again sub-lists with contents as defined by ngamsDb._sum1Cols
        (see general documentation of the ngamsDb Class.

        hostId:            Name of NGAS host on which the files reside
                           (string).

        diskId:            Disk ID of disk to take into account (string|None).

        fileId:            File ID of file(s) to take into account
                           (string|None).

        fileVersion:       Version of file(s) to take into account
                           (integer|None).

        Returns:           Cursor object (<NG/AMS DB Cursor Object API>).
        """
        info(5,"Entering getFileSummarySpuriousFiles1() ...")
        fileStatusList = [NGAMS_FILE_STATUS_OK, NGAMS_FILE_CHK_ACTIVE]
        sqlQuery = "SELECT " + _sum1Cols + " " +\
                   "FROM ngas_disks nd, ngas_files nf " +\
                   "WHERE nd.disk_id=nf.disk_id " +\
                   "AND (nf.ignore=1 OR " +\
                   "nf.file_status NOT IN (" + str(fileStatusList)[1:-1]+ "))"

        if (hostId): sqlQuery += " AND nd.host_id='" + hostId + "'"
        if (diskId): sqlQuery += " AND nd.disk_id='" + diskId + "'"
        if (fileId): sqlQuery += " AND nf.file_id='" + fileId + "'" 
        if (fileVersion):
            sqlQuery += " AND nf.file_version=" + str(fileVersion) 
        
        # Create a cursor and perform the query.
        curObj = self.dbCursor(sqlQuery)
        
        info(5, "Leaving getFileSummarySpuriousFiles1(). DB Time: " +\
             self.getDbTimeStr())
        return curObj


    def setFileChecksum(self,
                        fileId,
                        fileVersion,
                        diskId,
                        checksum,
                        checksumPlugIn):
        """
        Set the checksum value in the ngas_files table.

        fileId:          ID of file (string).
        
        fileVersion:     Version of file (integer).

        diskId:          ID of disk where file is stored (string).

        checksum:        Checksum of file (string).

        checksumPlugIn:  Name of plug-in used to generate the
                         checksum (string).

        Returns:         Reference to object itself (ngamsDb).
        """
        try:
            info(4, "Entering setFileChecksum() ...")
            sqlQuery = "UPDATE ngas_files SET checksum='" + checksum + "', " +\
                       "checksum_plugin='" + checksumPlugIn + "' " +\
                       "WHERE file_id='" + fileId + "' AND file_version=" +\
                       str(fileVersion) + " AND disk_id='" + diskId + "'"
            res = self.query(sqlQuery)
            self._storeWrTime()

            # Create a File Removal Status Document.
            if (self.__createSnapshot):
                dbFileInfo = self.getFileInfoFromFileIdHostId(getHostId(),
                                                              fileId,
                                                              fileVersion,
                                                              diskId)
                tmpFileObj = ngamsFileInfo.ngamsFileInfo().\
                             unpackSqlResult(dbFileInfo)
                self.createDbFileChangeStatusDoc(NGAMS_DB_CH_FILE_UPDATE,
                                                 [tmpFileObj])

            self.triggerEvents()
            info(4, "Leaving setFileChecksum(). DB Time: "+self.getDbTimeStr())
            return self
        except Exception, e:   
            raise e


    def _updateDiskFileStatus(self,
                              diskId,
                              fileSize):
        """
        Update the NGAS Disks Table according to a new file archived.

        diskId:       Disk ID (string).
        
        fileSize:     Size of file as stored on disk (integer).

        Returns:      Reference to object itself (
        """
        info(4,"Entering _updateDiskFileStatus() ...")
        try:
            self.takeGlobalDbSem()
            sqlQuery = "SELECT number_of_files, available_mb, " +\
                       "bytes_stored, mount_point " +\
                       "FROM ngas_disks WHERE disk_id='" + diskId + "'"
            res = self.query(sqlQuery, ignoreEmptyRes=0)
            self._storeRdTime()

            if (len(res[0]) != 1):
                format = "Cannot find entry for disk with ID: %s."
                errMsg = format % (diskId)
                raise Exception, errMsg

            # Generate new values.
            numberOfFiles    = res[0][0][0]
            bytesStored      = res[0][0][2]
            mountPoint       = res[0][0][3]
            newNumberOfFiles = (numberOfFiles  + 1)
            newAvailMb       = getDiskSpaceAvail(mountPoint)
            newBytesStored   = (bytesStored + fileSize)
            sqlQuery = "UPDATE ngas_disks SET" +\
                       " number_of_files=" + str(newNumberOfFiles) + "," +\
                       " available_mb=" + str(newAvailMb) + "," +\
                       " bytes_stored=" + str(newBytesStored) +\
                       " WHERE disk_id='" + diskId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.relGlobalDbSem()
            self.triggerEvents()
            info(4, "Leaving _updateDiskFileStatus(). DB Time: " +\
                 self.getDbTimeStr())
        except Exception, e:
            self.relGlobalDbSem()
            raise e

        
    def setFileStatus(self,
                      fileId,
                      fileVersion,
                      diskId,
                      status):
        """
        Set the checksum value in the ngas_files table.

        fileId:        ID of file (string).
        
        fileVersion:   Version of file (integer).

        diskId:        Disk ID for disk where file is stored (string).
        
        status:        File Status (8 bytes) (string).

        Returns:       Reference to object itself (ngamsDb).
        """
        try:
            info(5, "Entering setFileStatus() ...")
            sqlQuery = "UPDATE ngas_files SET file_status='" + status + "' " +\
                       "WHERE file_id='" + fileId + "' AND file_version=" +\
                       str(fileVersion) + " AND disk_id='" + diskId + "'"
            res = self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(5, "Leaving setFileStatus(). DB Time: " + self.getDbTimeStr())
            return self
        except Exception, e:   
            raise e


    def deleteDiskInfo(self,
                       diskId,
                       delFileInfo = 1):
        """
        Delete a record for a certain disk in the NGAS DB.

        CAUTION:  IF THE DB USER WITH WHICH THERE IS LOGGED IN HAS PERMISSION
                  TO EXECUTE DELETE STATEMENTS, THE INFORMATION ABOUT THE
                  DISK IN THE NGAS DB WILL BE DELETED! THIS INFORMATION
                  CANNOT BE RECOVERED!!

        diskId:        ID of disk for which to delete the entry (string).
        
        delFileInfo:   If set to 1, the file information for the files
                       stored on the disk is deleted as well (integer/0|1).
        
        Returns:       Reference to object itself (ngamsDb).
        """
        fileInfoDbmName = fileInfoDbm = None
        try:
            info(4, "Entering deleteDiskInfo() ...")
            if (delFileInfo and self.__createSnapshot):
                diskInfo = ngamsDiskInfo.ngamsDiskInfo().read(self, diskId)
            else:
                diskInfo = None
                
            if (delFileInfo):
                # Get the information about the files on the disk (before this
                # information is deleted).
                if (self.__createSnapshot):
                    ts = PccUtTime.TimeStamp().getTimeStamp()
                    fileName = ts + "_" + str(getUniqueNo()) + "_DISK_INFO"
                    fileInfoDbmName = os.path.normpath(self.__dbTmpDir + "/" +\
                                                       fileName)
                    fileInfoDbmName = self.dumpFileInfoList(diskId,
                                                            fileListDbmName=\
                                                            fileInfoDbmName)
                                                            
            # Delete the disk info.
            sqlQuery = "DELETE FROM ngas_disks WHERE disk_id='" + diskId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            
            # Delete file info if requested.
            if (delFileInfo):
                sqlQuery = "DELETE FROM ngas_files WHERE disk_id='"+diskId+"'"
                self.query(sqlQuery)
                self._storeWrTime()

                # Create a File Removal Status Document.
                if (self.__createSnapshot):
                    fileInfoDbm = ngamsDbm.ngamsDbm(fileInfoDbmName)
                    fileInfoDbm.initKeyPtr()
                    key, fileSqlInfo = fileInfoDbm.getNext()
                    while (key):
                        tmpFileObj = ngamsFileInfo.ngamsFileInfo().\
                                     unpackSqlResult(fileSqlInfo)
                        op = NGAMS_DB_CH_FILE_DELETE
                        self.createDbFileChangeStatusDoc(op, [tmpFileObj],
                                                         [diskInfo])
                        key, fileSqlInfo = fileInfoDbm.getNext()

            self.triggerEvents([diskInfo.getDiskId(),diskInfo.getMountPoint()])
            del fileInfoDbm
            rmFile(fileInfoDbmName + "*")
            info(4, "Leaving deleteDiskInfo(). DB Time: "+self.getDbTimeStr())
            return self
        except Exception, e:
            error("Error deleting disk info from DB: %s" % str(e))
            if (fileInfoDbm): del fileInfoDbm
            if (fileInfoDbmName): rmFile(fileInfoDbmName + "*")
            try:
                del cursorObj
            except:
                pass
            raise e

                   
    def deleteFileInfo(self,
                       diskId,
                       fileId,
                       fileVersion,
                       genSnapshot = 1):
        """
        Delete one record for a certain file in the NGAS DB.

        CAUTION:  IF THE DB USER WITH WHICH THERE IS LOGGED IN HAS PERMISSION
                  TO EXECUTE DELETE STATEMENTS, THE INFORMATION ABOUT THE
                  FILE(S) IN THE NGAS DB WILL BE DELETED! THIS INFORMATION
                  CANNOT BE RECOVERED!!

        diskId:         ID of disk hosting the file (string).

        fileId:         ID of file to be deleted. No wildcards accepted
                        (string).

        fileVersion:    Version of file to delete (integer)

        genSnapshot:    Generate Db Snapshot (integer/0|1).

        Returns:        Reference to object itself (ngamsDb).
        """
        fileInfoDbmName = fileInfoDbm = None
        try:
            info(4, "Entering deleteFileInfo() ...")
          
            # We have to update some fields of the disk hosting the file
            # when we delete a file (number_of_files, available_mb,
            # bytes_stored, also maybe later: checksum.
            dbDiskInfo = self.getDiskInfoFromDiskId(diskId)
            fileInfoDbmName = self.dumpFileInfoList(diskId,fileId,fileVersion,
                                                    None,fileInfoDbmName)
            fileInfoDbm = ngamsDbm.ngamsDbm(fileInfoDbmName)
            if (fileInfoDbm.getCount() > 0):
                dbFileInfo = fileInfoDbm.get("0")
            else:
                format = "Cannot remove file. File ID: %s, " +\
                         "File Version: %d, Disk ID: %s"
                errMsg = format % (fileId, fileVersion, diskId)
                raise Exception, errMsg
            sqlQuery = "DELETE FROM ngas_files WHERE disk_id='"+diskId+"' " +\
                       "AND file_id='" + fileId + "' AND file_version=" +\
                       str(fileVersion)
            self.query(sqlQuery)
            self._storeWrTime()

            # Create a File Removal Status Document.
            if (self.__createSnapshot and genSnapshot):
                tmpFileObj = ngamsFileInfo.ngamsFileInfo().\
                             unpackSqlResult(dbFileInfo)
                self.createDbFileChangeStatusDoc(NGAMS_DB_CH_FILE_DELETE,
                                                 [tmpFileObj])

            # Now update the ngas_disks entry for the disk hosting the file.
            if (dbDiskInfo):
                newNumberOfFiles = (dbDiskInfo[NGAS_DISKS_NO_OF_FILES] - 1)
                if (newNumberOfFiles < 0): newNumberOfFiles = 0
                newAvailMb = (dbDiskInfo[NGAS_DISKS_AVAIL_MB] +
                              int(float(dbFileInfo[NGAS_FILES_FILE_SIZE])/1e6))
                newBytesStored = (dbDiskInfo[NGAS_DISKS_BYTES_STORED] -
                                  dbFileInfo[NGAS_FILES_FILE_SIZE])
                if (newBytesStored < 0): newBytesStored = 0
                sqlQuery = "UPDATE ngas_disks SET" +\
                           " number_of_files=" + str(newNumberOfFiles) + "," +\
                           " available_mb=" + str(newAvailMb) + "," +\
                           " bytes_stored=" + str(newBytesStored) +\
                           " WHERE disk_id='" + diskId + "'"
                self.query(sqlQuery)
                self._storeWrTime()          

            self.triggerEvents()
            if (fileInfoDbmName): rmFile(fileInfoDbmName)
            info(4, "Leaving deleteFileInfo(). DB Time: "+self.getDbTimeStr())
            return self
        except Exception, e:
            if (fileInfoDbmName): rmFile(fileInfoDbmName)
            try:
                del cursorObj
            except:
                pass
            raise e
    

    def getLastDiskCheck(self,
                         hostId = ""):
        """
        Queries all the Last Check Flags for all disks or for all disks
        currently mounted in a specific host. A Dictionary is returned
        containining the Disk IDs as keys, and the time for the last check
        in seconds. If the value is NULL in the DB, it is set to 0.

        hostId:    If specified, only disks mounted in this system are
                   taken into account (string).

        Returns:   Dictionary with entry for each disk. Key is Disk
                   ID (dictionary).
        """
        sqlQuery = "SELECT disk_id, last_check from ngas_disks"
        if (hostId != ""): sqlQuery += " WHERE host_id = '" + hostId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        diskDic = {}
        for diskInfo in res[0]:
            if (not diskInfo[1]):
                timeSinceLastCheck = 0
            else:
                if (type(diskInfo[1]) == types.IntType):
                    timeSinceLastCheck = int(diskInfo[1])
                elif (type(diskInfo[1]) == types.StringType):
                    # Expects an ISO8601 timestamp.
                    timeSinceLastCheck = int(ngamsLib.\
                                             iso8601ToSecs(diskInfo[1])+0.5)
                else:
                    timeSinceLastCheck = diskInfo[1].ticks()
            diskDic[diskInfo[0]] = timeSinceLastCheck
        return diskDic


    def setLastCheckDisk(self,
                         diskId,
                         timeSecs):
        """
        Update the Last Check Flag for a disk.

        diskId:           ID of disk for which to update record (string).
        
        timeSecs:         Time in seconds to set for the disk (integer).

        Returns:          Reference to object itself (ngamsDb).
        """
        try:
            try:
                self.takeDbSem()
                dbTime = self.__dbDrv.convertTimeStamp(timeSecs)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e
            sqlQuery = "UPDATE ngas_disks SET last_check = '" + dbTime +\
                       "' WHERE disk_id='" + diskId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            return self
        except Exception, e:   
            raise e


    def reqWakeUpCall(self,
                      wakeUpHostId,
                      wakeUpTime):
        """
        Request a Wake-Up Call via the DB.
        
        wakeUpHostId:  Name of host where the NG/AMS Server requested for
                       the Wake-Up Call is running (string).
        
        wakeUpTime:    Absolute time for being woken up (seconds since
                       epoch) (integer).

        Returns:       Reference to object itself (ngamsDb). 
        """
        try:
            info(4,"Entering reqWakeUpCall() ...")
            try:
                self.takeDbSem()
                wakeUpTimeLoc = self.__dbDrv.convertTimeStamp(wakeUpTime)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e
            sqlQuery = "UPDATE ngas_hosts SET " +\
                       "srv_suspended=1, " +\
                       "srv_req_wake_up_srv='" + wakeUpHostId + "', " +\
                       "srv_req_wake_up_time='" + wakeUpTimeLoc + "' " +\
                       "WHERE host_id='" + getHostId() + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4,"Leaving reqWakeUpCall()")
            return self
        except Exception, e:   
            raise e


    def markHostSuspended(self,
                          hostId = None):
        """
        Mark a host as being suspended in the NGAS DB.

        hostId:    Name of host to mark as suspended. If not given the
                   local host name is used (string).

        Returns:   Reference to object itself (ngamsDb).
        """
        try:
            info(4,"Entering markHostSuspended() ...")
            if (hostId == None): hostId = getHostId()
            sqlQuery = "UPDATE ngas_hosts SET " + "srv_suspended=1 " +\
                       "WHERE host_id='" + hostId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4,"Leaving markHostSuspended()")
            return self
        except Exception, e:   
            raise e
       

    def resetWakeUpCall(self,
                        hostId = None,
                        resetSrvSusp = 0):
        """
        Cancel/reset the Wake-Up Call parameters.

        hostId:        If specified, another host ID than the one where
                       this NG/AMS Server is running can be indicated (string).

        Returns:       Reference to object itself (ngamsDb). 
        """
        try:
            info(4,"Entering resetWakeUpCall() ...")
            if (hostId == None): hostId = getHostId()
            sqlQuery = "UPDATE ngas_hosts SET srv_req_wake_up_srv=''"
            if (resetSrvSusp): sqlQuery += ", srv_suspended=0"
            sqlQuery += " WHERE host_id='" + hostId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4,"Leaving resetWakeUpCall()")
            return self
        except Exception, e:   
            raise e


    def getWakeUpRequests(self):
        """
        Generates a tuple with suspended NGAS Hosts that have requested
        to be woken up by this host.

        Returns:  Tuple containing sub-tuples with information about hosts
                  to be woken up:

                    (({host id}, {wake-up time (secs since epoch)}), ...)
                    
                                                                (list/tuple)
        """
        info(5,"Entering getWakeUpRequests() ...")
        sqlQuery = "SELECT host_id, srv_req_wake_up_time from ngas_hosts " +\
                   "WHERE srv_req_wake_up_srv='" + getHostId() + "' " +\
                   "AND srv_suspended=1"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(5,"Leaving getWakeUpRequests()")
        if (len(res[0]) >= 1):
            wakeUpReqs = []
            for req in res[0]:
                suspHost   = req[0]
                tmpWakeUpTime = timeRef2Iso8601(req[1])
                wakeUpTime = ngamsLib.iso8601ToSecs(tmpWakeUpTime)
                wakeUpReqs.append((suspHost, wakeUpTime))
            return wakeUpReqs
        else:
            return []


    def getPortNoFromHostId(self,
                            hostId):
        """
        Return the port number corresponding to the host ID.

        hostId:    Host ID (string).

        Return:    Port number (integer).
        """
        sqlQuery = "SELECT srv_port from ngas_hosts where host_id='" +\
                   hostId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (len(res[0]) == 1):
            return int(res[0][0][0])
        else:
            errMsg = "Error retrieving port number for host: " + hostId
            raise Exception, errMsg

       
    def addDiskHistEntry(self,
                         diskId,
                         synopsis,
                         descrMimeType = None,
                         descr = None,
                         origin = None,
                         date = None):
        """
        Add an entry in the NGAS Disks History Table (ngas_disks_hist)
        indicating a major action or event occurring in the context of a disk.

        dbConObj:       Instance of NG/AMS DB class (ngamsDb).
    
        diskId:         Disk ID for the disk concerned (string).

        synopsis:       A short description of the action
                        (string/max. 255 char.s).

        descrMimeType:  The mime-type of the contents of the description
                        field. Must be specified when a description is
                        given (string).

        descr:          An arbitrary long description of the action or event
                        in the life-time of the disk (string). 

        origin:         Origin of the history log entry. Can either be the
                        name of an application or the name of an operator.
                        If not specified (= None) it will be set to 
                        'NG/AMS - <host name>' (string).

        date:           Date for adding the log entry. If not specified (set
                        to None), the function takes the current date and
                        writes this in the new entry (string/ISO-8601).
    
        Returns:        Void.
        """
        try:
            info(4,"Entering addDiskHistEntry() ...")
            if (origin == None): origin = "NG/AMS@" + getHostId()
            tsObj = PccUtTime.TimeStamp()
            try:
                self.takeDbSem()
                if (date == None):
                    histDate = self.__dbDrv.\
                               convertTimeStamp(tsObj.getTimeStamp())
                else:
                    histDate = self.__dbDrv.convertTimeStamp(date)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e    
            if (descr != None):
                if (descrMimeType == None):
                    errMsg = "Mime-type must be specified for entry in the "+\
                             "NGAS Disks History when a Description is given!"
                    raise Exception, errMsg
                if (descrMimeType == NGAMS_XML_MT):
                    descr = re.sub("\n", "", descr)
                    descr = re.sub("> *<", "><", descr)
                    descr = re.sub(">\t*<", "><", descr)
                descr = "'" + descr + "'"
            else:
                descr = "None"
            if (descrMimeType != None):
                mt = "'" + descrMimeType + "'"
            else:
                mt = "None"
            sqlQuery = "INSERT INTO ngas_disks_hist " +\
                       "(disk_id, hist_date, hist_origin, hist_synopsis, " +\
                       "hist_descr_mime_type, hist_descr) VALUES (" +\
                       "'" + diskId + "', " +\
                       "'" + histDate + "', " +\
                       "'" + origin + "', " +\
                       "'" + synopsis + "', " +\
                       mt + ", " +\
                       descr + ")"
            res = self.query(sqlQuery)
            self._storeWrTime()
            info(2,"Added entry in NGAS Disks History Table - Disk ID: " +\
                 diskId + " - Date: " + tsObj.getTimeStamp() + " - Origin: " +\
                 origin + " - Synopsis: " + synopsis +\
                 " - Description Mime-type: " + str(mt) + " - Description: " +\
                 str(descr))        
            self.triggerEvents()
            info(4,"Leaving addDiskHistEntry()")
        except Exception, e:   
            raise e
             

    def getSubscriberInfo(self,
                          subscrId = "",
                          hostId = "",
                          portNo = -1):
        """
        Get the information for one or more Subcribers from the
        ngas_subscribers table and return the contents in a list. The format
        of this list is formatted as follows:

          [<Host ID>, <Port No>, <Priority>, <Subscriber ID>, <Subscriber URL>,
           <Subscription Start Date>, <Subscription Filter Plug-In>,
           <Subscription Filter Plug-In Parameters>,
           <Last File Ingestion Date>]

        subscrId:   ID of the Subcriber (string).

        hostId:     Limit the query to Subscribers in connection with one
                    host (Data Provider) (string).

        portNo:     Limit the query to Subscribers in connection with one
                    host (Data Provider) (integer).

        Returns:    If a Subscriber ID is specified: List with information
                    about the Subscriber (if found). Otherwise [] is returned
                    (list).

                    If no Subscriber ID is given: List with sub-lists with
                    information for all Subscribers. Otherwise [] is returned
                    (list/list).
        """
        info(4,"Entering getSubscriberInfo() ...")
        sqlQuery = "SELECT " + _ngasSubscribersCols +\
                   " FROM ngas_subscribers ns"
        if ((subscrId != "") or (hostId != "")): sqlQuery += " WHERE "
        if (subscrId != ""): sqlQuery += "subscr_id='" + subscrId + "'"
        if ((hostId != "") and (subscrId != "")):
            sqlQuery += " AND host_id='" + hostId + "'"
        if ((hostId != "") and (subscrId == "")):
            sqlQuery += " ns.host_id='" + hostId + "'"
        if ((portNo != -1) and ((hostId != ""))):
            sqlQuery += " AND ns.srv_port=" + str(portNo)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4,"Leaving getSubscriberInfo()")
        if (res == [[]]):
            return []
        else:
            if (subscrId != ""):
                return res[0][0]
            else:
                return res[0]


    def writeSubscriberEntry(self,
                             hostId,
                             portNo,
                             subscrId,
                             subscrUrl,
                             priority = 10,
                             startDate = "",
                             filterPlugIn = "",
                             filterPlugInPars = "",
                             lastFileIngDate = ""):
        """
        The method writes the information in connection with a Subscriber
        in the NGAS DB. If an entry already exists for that disk, it is updated
        with the information given as input parameters. Otherwise, a new
        entry is created.

        hostId:
        ...
        filterPlugInPars:    Parameters for the Subscriber (string).

        priority:            Priority of Subscriber (integer).

        startDate:           Date the subscription should start from
                             (string/ISO8601).

        lastFileIngDate:     Ingestion dtae of last file delivered
                             (string/ISO8601).

        Returns:             Returns 1 if a new entry was created in the DB
                             and 0 if an existing entry was updated
                             (integer/0|1).
        """
        try:
            info(4, "Entering writeSubscriberEntry() ...")
            try:
                self.takeDbSem()
                if ((startDate == "") or (startDate == None)):
                    startDate = "None"
                else:
                    startDate = "'" +\
                                self.__dbDrv.convertTimeStamp(startDate) + "'"
                if ((filterPlugIn == "") or (filterPlugIn == None)):
                    filterPlugIn = "''"
                else:
                    filterPlugIn = "'" + filterPlugIn + "'"
                if ((filterPlugInPars == "") or (filterPlugInPars == None)):
                    filterPlugInPars = "''"
                else:
                    filterPlugInPars = "'" + filterPlugInPars + "'"
                if ((lastFileIngDate == "") or (lastFileIngDate == None)):
                    lastFileIngDate = "'" +\
                                      self.__dbDrv.convertTimeStamp(0) + "'"
                else:
                    lastFileIngDate = "'" + self.__dbDrv.\
                                      convertTimeStamp(lastFileIngDate) + "'"
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e                  

            # Check if the entry already exists. If yes update it, otherwise
            # insert a new element.
            if (self.subscriberInDb(subscrId)):
                sqlQuery = "UPDATE ngas_subscribers SET " +\
                           "host_id='" + hostId + "', " +\
                           "srv_port=" + str(portNo) + ", " +\
                           "subscr_prio=" + str(priority) + ", " +\
                           "subscr_id='" + subscrId + "', " +\
                           "subscr_url='" + subscrUrl + "', " +\
                           "subscr_start_date=" + startDate + ", " +\
                           "subscr_filter_plugin=" + filterPlugIn + ", " +\
                           "subscr_filter_plugin_pars="+filterPlugInPars+", "+\
                           "last_file_ingestion_date=" + lastFileIngDate+" " +\
                           "WHERE subscr_id='" + subscrId + "' AND " +\
                           "host_id='" + hostId + "' AND " +\
                           "srv_port=" + str(portNo)
                addedEntry = 0
            else:
                sqlQuery = "INSERT INTO ngas_subscribers " +\
                           "(host_id, srv_port, subscr_prio, subscr_id," +\
                           " subscr_url, subscr_start_date," +\
                           " subscr_filter_plugin,"+\
                           " subscr_filter_plugin_pars," +\
                           " last_file_ingestion_date) " +\
                           " VALUES " +\
                           "('" + hostId + "', " + str(portNo) + ", " +\
                           str(priority) + ", '"+subscrId + "', '"+subscrUrl+\
                           "', " + startDate + ", " + filterPlugIn + ", " +\
                           filterPlugInPars + ", " + lastFileIngDate + ")"
                addedEntry = 1
            res = self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4, "Leaving writeSubscriberEntry(). DB Time: " +\
                 self.getDbTimeStr())
            return addedEntry
        except Exception, e:   
            raise e


    def deleteSubscriber(self,
                         subscrId):
        """
        Delete the information for one Subscriber from the NGAS DB.

        subscrId:   Subscriber ID (string).

        Returns:    Reference to object itself (ngamsDb).
        """
        try:
            info(4, "Entering deleteSubscriber() ...")
            sqlQuery = "DELETE FROM ngas_subscribers WHERE subscr_id='" +\
                       subscrId + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4, "Leaving deleteSubscriber(). DB Time: " +\
                 self.getDbTimeStr())
            return self
        except Exception, e:   
            raise e


    def getSubscriberStatus(self,
                            subscrIds,
                            hostId = "",
                            portNo = -1):
        """
        Method to query the information about the Ingestion Date of the
        last file delivered to the Subscriber. A list is returned, which
        contains the following:

          [(<Subscriber ID>, <Last File Ingestion Date (ISO8601)>), ...]

        subscrIds:      List of Subscriber ID to query (list/string).

        hostId:         Host name of Subscriber host (string).

        portNo:         Port number used by Subscriber host (integer).
                            
        Returns:        List with Subscriber status (list/tuple/string).
        """
        info(4, "Entering getSubscriberStatus() ...")
        if (not subscrIds):
            info(4, "Leaving getSubscriberStatus()")
            return []
        sqlQuery = "SELECT subscr_id, last_file_ingestion_date " +\
                   "FROM ngas_subscribers WHERE subscr_id IN (" +\
                   str(subscrIds)[1:-1] + ")"
        if (hostId != ""):
            sqlQuery += " AND host_id='" + hostId + "'"
            if (portNo != -1):
                sqlQuery += " AND srv_port=" + str(portNo)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSubscriberStatus()")
        if (res == [[]]):
            return []
        else:
            subscrStatus = []
            for subscrInfo in res[0]:
                if (subscrInfo[1]):
                    lastIngDate = PccUtTime.TimeStamp().\
                                  initFromSecsSinceEpoch(subscrInfo[1]).\
                                  getTimeStamp()
                else:
                    lastIngDate = None
                subscrStatus.append((subscrInfo[0], lastIngDate))
            return subscrStatus


    def subscrBackLogEntryInDb(self,
                               hostId,
                               portNo,
                               subscrId,
                               fileId,
                               fileVersion):
        """
        Check if there is an entry in the Subscription Back-Log for that
        file/Subscriber.

        hostId:          Host ID for NGAS host where Data Provider concerned
                         is running (string).

        portNo:          Port number used by Data Provider concerned (integer).

        subscrId:        Subscriber ID (string).
        
        fileId:          File ID (string).
        
        fileVersion:     File Version (string).

        Returns:         1 = file found, 0 = file no found (integer).
        """
        info(4, "Entering subscrBackLogEntryInDb() ...")
        sqlQuery = "SELECT file_id FROM ngas_subscr_back_log " +\
                   "WHERE host_id='" + hostId + "' " +\
                   "AND srv_port=" + str(portNo) + " " +\
                   "AND subscr_id='" + subscrId + "' " +\
                   "AND file_id='" + fileId + "' " +\
                   "AND file_version=" + str(fileVersion)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving subscrBackLogEntryInDb(). DB Time: " +\
             self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0] == fileId):
                return 1
            else:
                return 0
        else:
            return 0


    def addSubscrBackLogEntry(self,
                              hostId,
                              portNo,
                              subscrId,
                              subscrUrl,
                              fileId,
                              fileName,
                              fileVersion,
                              ingestionDate,
                              format):
        """
        Adds a Back-Log Entry in the DB. If there is already an entry
        for that file/Subscriber, a new entry is not created.

        hostId:          Host ID for NGAS host where Data Provider concerned
                         is running (string).

        portNo:          Port number used by Data Provider concerned (integer).

        subscrUrl:       Susbcriber URl to where the files are delivered
                         (string).
        
        subscrId:        Subscriber ID (string).
        
        fileId:          File ID (string).

        fileName:        Filename, i.e., name of file as stored in the
                         Subscription Back-Log Area (string).
        
        fileVersion:     File Version (integer).

        ingestionDate:   File Ingestion Date (string/ISO8601).

        format:          Mime-type of file (string).

        Returns:         Void.
        """
        try:
            info(4, "Entering addSubscrBackLogEntry() ...")
            try:
                self.takeDbSem()
                ingDate = self.__dbDrv.convertTimeStamp(ingestionDate)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e             
            if (not self.subscrBackLogEntryInDb(hostId, portNo, subscrId,
                                                fileId, fileVersion)):
                sqlQuery = "INSERT INTO ngas_subscr_back_log " +\
                           "(host_id, srv_port, subscr_id, subscr_url, " +\
                           "file_id, file_name, file_version, " +\
                           "ingestion_date, format) " +\
                           "VALUES " +\
                           "('" + hostId + "', " + str(portNo) + ", '" +\
                           subscrId + "', '" + subscrUrl + "', '" +\
                           fileId + "', '" + fileName + "', " +\
                           str(fileVersion) +\
                           ", '" + ingDate + "', '" + format + "')"
                self.query(sqlQuery)
                self._storeWrTime()
                self.triggerEvents()
            info(4, "Leaving addSubscrBackLogEntry(). DB Time: " +\
                 self.getDbTimeStr())
        except Exception, e:   
            raise e


    def delSubscrBackLogEntry(self,
                              hostId,
                              portNo,
                              subscrId,
                              fileId,
                              fileVersion):
        """
        Delete an entry in the Subscription Back-Log Table.

        hostId:          Host ID for NGAS host where Data Provider concerned
                         is running (string).

        portNo:          Port number used by Data Provider concerned (integer).
 
        subscrId:        Subscriber ID (string).
        
        fileId:          File ID (string).
        
        fileVersion:     File Version (string).
        
        fileName:        Filename, i.e., name of file as stored in the
                         Subscription Back-Log Area (string).

        Returns:         Void.
        """
        try:
            info(4, "Entering delSubscrBackLogEntry() ...")
            if (self.subscrBackLogEntryInDb(hostId, portNo, subscrId, fileId,
                                            fileVersion)):
                sqlQuery = "DELETE FROM ngas_subscr_back_log " +\
                           "WHERE host_id='" + hostId + "' " +\
                           "AND srv_port=" + str(portNo) + " " +\
                           "AND subscr_id='" + subscrId + "' " +\
                           "AND file_id='" + fileId + "' " +\
                           "AND file_version=" + str(fileVersion)
                res = self.query(sqlQuery)
                self._storeWrTime()
                self.triggerEvents()
            info(4, "Leaving delSubscrBackLogEntry(). DB Time: " +\
                 self.getDbTimeStr())
        except Exception, e:   
            raise e


    def updateSubscrStatus(self,
                           subscrId,
                           fileIngDate):
        """
        Update the Subscriber Status so that it reflects the File Ingestion
        Date of the last file ingested.

        subscrId:       Subscriber ID (string).
        
        fileIngDate:    File Ingestion Date (string/ISO8601).

        Returns:        Void.
        """
        try:
            info(4, "Entering updateSubscrStatus() ...")
            try:
                self.takeDbSem()
                ingDate = self.__dbDrv.convertTimeStamp(fileIngDate)
                self.relDbSem()
            except Exception, e:
                self.relDbSem()
                raise Exception, e
            sqlQuery = "UPDATE ngas_subscribers SET " +\
                       "last_file_ingestion_date='" +\
                       ingDate + "' WHERE subscr_id='" + subscrId + "' AND " +\
                       "last_file_ingestion_date < '" + ingDate + "'"
            self.query(sqlQuery)
            self._storeWrTime()
            self.triggerEvents()
            info(4, "Leaving updateSubscrStatus(). DB Time: " +\
                 self.getDbTimeStr())
        except Exception, e:   
            raise e


    def getSubscrBackLog(self,
                         hostId,
                         portNo):
        """
        Read all entries in the Subscriber Back-Log Table 'belonging'
        to a specific Data Provider, and return these in a list with sub-lists.

        hostId:      Host ID of Data Provider (string).
        
        portNo:      Port number used by Data Provider (integer).

        Returns:     List containing sub-list with the following information:

                       [[<Subscr. ID>, <Subscr. URL>, <File ID>, <Filename>,
                         <File Version>, <Ingestion Date>,
                         <Format <Mime-Type>], ...]

                     Note that the part of the list after the Subscriber URL
                     is the same as generated by ngamsDb.getFileSummary2()
                     (list/list).
        """
        info(4, "Entering getSubscrBackLog() ...")
        sqlQuery = "SELECT subscr_id, subscr_url, file_id, file_name, " +\
                   "file_version, ingestion_date, format " +\
                   "FROM ngas_subscr_back_log WHERE " +\
                   "host_id='" + hostId + "' AND srv_port=" + str(portNo)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSubscrBackLog(). DB Time: " + self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            # Convert the time into ISO8601.
            procList = []
            for fi in res[0]:
                ingDate = PccUtTime.TimeStamp().initFromSecsSinceEpoch(fi[5]).\
                          getTimeStamp()
                newItem = list(fi[0:5]) + [ingDate] + [fi[6]]
                procList.append(newItem)
            return procList


    def getSubscrsOfBackLogFile(self,
                                fileId,
                                fileVersion):
        """
        Get a list of Subscribers (their IDs) interested in this Back-Log
        Buffered file.

        fileId:       File ID (string).
        
        fileVersion:  File Version (string).

        Returns:      List with Subscriber IDs (if any) (list/string).
        """
        info(4, "Entering getSubscrsOfBackLogFile() ...")
        sqlQuery = "SELECT subscr_id FROM ngas_subscr_back_log " +\
                   "WHERE file_id='" + fileId + "' AND " +\
                   "file_version=" + str(fileVersion)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSubscrsOfBackLogFile(). DB Time: " +\
             self.getDbTimeStr())
        if (res == [[]]):
            return []
        else:
            return res[0]


    def getSumBytesStored(self,
                          diskId):
        """
        Get the total sum of the sizes of the data files stored on a disk
        and return this.

        diskId:    Disk ID of disk to get the sum for (string).

        Return:    Total sum of bytes stored on the disk (integer).
        """
        info(4, "Entering getSumBytesStored() ...")
        sqlQuery = "SELECT sum(file_size) from ngas_files WHERE " +\
                   "disk_id='" + diskId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getSumBytesStored(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0]):
                return res[0][0][0]
            else:
                return 0
        else:
            return 0


    def getNumberOfFiles(self,
                         diskId = "",
                         fileId = "",
                         fileVersion = -1,
                         ignore = None,
                         onlyOnlineFiles = 0):
        """
        Get the number of files stored on a disk.

        diskId:          Disk ID of disk to get the number of files for
                         (string).

        fileId:          File ID for file to be retrieved (string).

        fileVersion:     If a File Version is specified only information
                         for files with that version number and File ID
                         are taken into account. The version must be a
                         number in the range [1; oo[ (integer).

        ignore:          If set to 0 or 1, this value of ignore will be
                         queried for. If set to None, ignore is not
                         considered (None|0|1).

        onlyOnlineFiles: If specified, only files which are Online or on
                         suspended nodes are considered (integer/0|1).

        Return:          Number of files stored on the disk (integer).
        """
        info(4, "Entering getNumberOfFiles() ...")
        sqlQuery = "SELECT count(file_id) from ngas_files nf"

        # Build up the query, take only Online/Suspended files into
        # account if requested.
        if (diskId or fileId or (fileVersion != -1) or (ignore != None) or
            onlyOnlineFiles):
            sqlQuery += ", ngas_disks nd WHERE nf.disk_id=nd.disk_id"
        if (diskId): sqlQuery += " AND nf.disk_id='%s'" % diskId
        if (fileId): sqlQuery += " AND nf.file_id='%s'" % fileId
        if (fileVersion > 0): sqlQuery += " AND nf.file_version=%d" %\
                                          int(fileVersion)
        if (ignore != None): sqlQuery += " AND nf.ignore=%d" % int(ignore)
        if (onlyOnlineFiles):
            # We assume here that either Disk ID, File ID, File Version
            # or ignore=1 specified so that we can append and AND clause.
            sqlQuery += " AND nf.disk_id IN (SELECT nd.disk_id " +\
                        "FROM ngas_disks nd, ngas_hosts nh " +\
                        "WHERE (nd.host_id=nh.host_id) OR " +\
                        "((nd.last_host_id=nh.host_id) AND " +\
                        "(nh.srv_suspended=1)))"
            
        # Now, do the query.
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getNumberOfFiles(). DB Time: " + self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0]):
                return res[0][0][0]
            else:
                return 0
        else:
            return 0


    def createDbFileChangeStatusDoc(self,
                                    operation,
                                    fileInfoObjList,
                                    diskInfoObjList = []):
        """
        The function creates a File DB Change Status XML document in the
        '<Disk Mt Pt>/staging/NgamsRemoveStatus' directory from the information
        in the 'fileInfoObj' object.

        operation:        Has to be either ngams.NGAMS_DB_CH_FILE_INSERT or
                          ngams.NGAMS_DB_CH_FILE_DELETE (string). 

        fileInfoObj:      List of instances of NG/AMS File Info Object
                          containing the information about the file
                          (list/ngamsFileInfo).

        diskInfoObjList:  It is possible to give the information about the
                          disk(s) in question via a list of ngamsDiskInfo
                          objects (list/ngamsDiskInfo).
                         
        Returns:          Void.
        """
        # IMPL: Implementation concern: This class is suppose to be
        # at a lower level in the hierarchie than the ngamsFileInfo,
        # ngamsFileList, ngamsDiskInfo and ngamsStatus classes and as
        # such these should not be used from within this class. All usage
        # of these classes in the ngamsDb class should be analyzed. Probably
        # these classes should be made base classes for this class.

        # IMPL: Potential memory bottleneck.

        info(4,"Entering createDbFileChangeStatusDoc() ...")
        timeStamp = PccUtTime.TimeStamp().getTimeStamp()

        # Sort the File Info Objects according to disks.
        fileInfoObjDic = {}
        for fileInfo in fileInfoObjList:
            if (not fileInfoObjDic.has_key(fileInfo.getDiskId())):
                fileInfoObjDic[fileInfo.getDiskId()] = []
            fileInfoObjDic[fileInfo.getDiskId()].append(fileInfo)

        # Get the mount points for the various disks concerned.
        mtPtDic = {}
        if (diskInfoObjList == []):
            diskIdMtPtList = self.getDiskIdsMtPtsMountedDisks(getHostId())
            for diskId, mtPt in diskIdMtPtList:
                mtPtDic[diskId] = mtPt
        else:
            for diskInfoObj in diskInfoObjList:
                mtPtDic[diskInfoObj.getDiskId()] = diskInfoObj.getMountPoint()

        # Create on each disk the relevant DB Change Status Document.
        tmpFileList = None
        for diskId in fileInfoObjDic.keys():
            statFilePath = os.path.normpath(mtPtDic[diskId] + "/" +\
                                            NGAMS_DB_CH_CACHE)
            statFilename = os.path.normpath(statFilePath + "/" +\
                                            timeStamp + "_" +\
                                            str(getUniqueNo()) + ".pickle")
            tmpStatFilename = statFilename + ".tmp"

            if ((len(fileInfoObjList) == 1) and (diskInfoObjList == [])):
                tmpStatObj = fileInfoObjList[0].setTag(operation)
            else:
                dbId = self.__dbServer + "." + self.__dbName
                tmpFileObjList = fileInfoObjDic[diskId]
                tmpFileList = ngamsFileList.ngamsFileList(dbId, operation)
                for fileObj in tmpFileObjList:
                    tmpFileList.addFileInfoObj(fileObj)        
                tmpStatObj = ngamsStatus.ngamsStatus().\
                             setDate(timeStamp).\
                             setVersion(getNgamsVersion()).\
                             setHostId(getHostId()).\
                             setMessage(dbId).\
                             addFileList(tmpFileList)
            
            info(4,"Creating Temporary DB Snapshot: " + statFilename)
            pickleFo = None
            try:
                pickleFo = open(tmpStatFilename, "w")
                cPickle.dump(tmpStatObj, pickleFo, 1)
                pickleFo.close()
                commands.getstatusoutput("mv " + tmpStatFilename + " " +\
                                         statFilename)
            except Exception, e:
                if (pickleFo): pickleFo.close()
                raise e
            del tmpStatObj
            tmpStatObj = None
            del tmpFileList
            tmpFileList = None
        
        info(4,"Leaving createDbFileChangeStatusDoc()")


    def getMinLastDiskCheck(self,
                            hostId):
        """
        Get the timestamp for the disk that was checked longest time ago
        for all disks mounted in a specific NGAS Host.

        hostId:   Host ID of the host to consider (string).

        Returns:  Time since the 'oldest', last check (seconds since epoch)
                  (integer).
        """
        info(4,"Entering getMinLastDiskCheck() ...")
        sqlQuery = "SELECT min(last_check) FROM ngas_disks WHERE " +\
                   "host_id='" + hostId + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        info(4, "Leaving getMinLastDiskCheck(). DB Time: "+self.getDbTimeStr())
        if (len(res[0]) == 1):
            if (res[0][0][0]):
                if (type(res[0][0][0]) == types.IntType):
                    return res[0][0][0]
                elif (type(res[0][0][0]) == types.StringType):
                    # Expects an ISO8601 timestamp.
                    return int(ngamsLib.iso8601ToSecs(res[0][0][0]) + 0.5)
                else:
                    return int(res[0][0][0].ticks() + 0.5)
            else:
                return None
        else:
            return None


    def updateDataCheckStat(self,
                            hostId,
                            start,
                            remain,
                            estimTime,
                            rate,
                            checkMb,
                            checkedMb,
                            checkFiles,
                            checkedFiles):
        """
        Update the statistics for the Data Checking Thread.

        hostId:          ID of NGAS Host to update statistics for (string).

        start:           Start of checking in seconds since epoch (integer).
        
        remain:          Estimated remaining time in seconds (integer).

        estimTime:       Estimated total time in seconds to complete the check
                         cycle (integer)
        
        rate:            Rate of checking in MB/s (float).
        
        checkMb:         Amount of data to check in MB (float).
        
        checkedMb:       Amount checked in MB (float).
        
        checkFiles:      Number of files to check (integer).
        
        checkedFiles:    Number of files checked (integer).

        Returns:         Reference to object itself (ngamsDb).
        """
        info(4,"Entering updateDataCheckStat() ...")
        try:
            self.takeDbSem()
            startDbTime = self.__dbDrv.convertTimeStamp(start)
            endDbTime = self.__dbDrv.convertTimeStamp(start + estimTime)
            self.relDbSem()
        except Exception, e:
            self.relDbSem()
            raise Exception, e
        sqlQuery = "UPDATE ngas_hosts SET " +\
                   "srv_check_start='" + startDbTime + "', " +\
                   "srv_check_remain=" + str(remain) + ", " +\
                   "srv_check_end='" + endDbTime + "', " +\
                   "srv_check_rate=" + str(rate) + ", " +\
                   "srv_check_mb=" + str(checkMb) + ", " +\
                   "srv_checked_mb=" + str(checkedMb) + ", " +\
                   "srv_check_files=" + str(checkFiles) + ", " +\
                   "srv_check_count=" + str(checkedFiles) + " " +\
                   "WHERE host_id='" + hostId + "'"
        self.query(sqlQuery)
        self._storeWrTime()
        info(4,"Leaving updateDataCheckStat()")
        return self

  
    def hasCfgPar(self,
                  groupId,
                  parName):
        """
        Return 1 if the given configuration parameter (given by its
        Simplified XPath name) and Configuration Group ID is defined in the
        configuration table in the DB.

        groupId:      Group ID for the parameter (string).
        
        parName:      Name of parameter (string).

        Returns:      1 = parameter defined, 0 = parameter not defined
                      (integer/0|1).
        """
        queryFormat = "SELECT cfg_par FROM ngas_cfg_pars WHERE " +\
                   "cfg_group_id='%s' AND cfg_par='%s'"
        sqlQuery = queryFormat % (groupId, parName)
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeRdTime()
        if (res == [[]]):
            return 0
        else:
            return 1

    
    def writeCfgPar(self,
                    groupId,
                    parName,
                    value,
                    comment):
        """
        Write a configuration parameter to the NGAS DB. If the parameter
        is already defined, the value/comment are updated.

        groupId:   Configuration Group ID (string).
        
        parName:   Name of parameter (string).
        
        value:     Value of parameter (string).
        
        comment:   Comment for parameter (string).
        
        Returns:   Reference to object itself (ngamsDb).
        """
        if (self.hasCfgPar(groupId, parName)):
            queryFormat = "UPDATE ngas_cfg_pars SET cfg_val='%s'"
            if (comment): queryFormat += ", cfg_comment='%s'"
            queryFormat += " WHERE cfg_group_id='%s' AND cfg_par='%s'"
            if (comment):
                sqlQuery = queryFormat % (str(value), comment, groupId,
                                          parName)
            else:
                sqlQuery = queryFormat % (str(value), groupId, parName)
        else:
            if (not comment): comment = ""
            queryFormat = "INSERT INTO ngas_cfg_pars (cfg_group_id, " +\
                          "cfg_par, cfg_val, cfg_comment) " +\
                          "VALUES ('%s', '%s', '%s', '%s')"
            sqlQuery = queryFormat % (groupId, parName, str(value), comment)
        self.query(sqlQuery)
        self._storeWrTime()
        return self


    def getCfgPars(self,
                   name):
        """
        Return the list of configuration parameters from the DB associated
        to the given name.

        name:     Name of the configuration association (string).

        Returns:  List with sub-lists with the information. The format is:

                    [[<Group ID>, <Parameter>, <Value>, <Comment>], ...]

                    (list).
        """
        # We should query the parameters according to the order in which
        # the Cfg. Group IDs are listed.
        sqlQuery = "SELECT cfg_par_group_ids FROM ngas_cfg WHERE " +\
                   "cfg_name='" + name + "'"
        res = self.query(sqlQuery, ignoreEmptyRes=0)
        self._storeWrTime()
        if (res == [[]]):
            raise Exception, "Referenced configuration: " + name + " not " +\
                  "found in the DB!"
        groupIds = res[0][0][0].split(",")
        accuRes = []
        for groupId in groupIds:
            groupId = groupId.strip()
            if (not groupId): continue
            sqlQuery = "SELECT cfg_group_id, cfg_par, cfg_val, cfg_comment " +\
                       "FROM ngas_cfg_pars WHERE cfg_group_id='" + groupId +"'"
            res = self.query(sqlQuery, ignoreEmptyRes=0)
            self._storeWrTime()
            accuRes += res[0]
        return accuRes
        
    
if __name__ == '__main__':
    """
    Main routine to make a connection to the DB.
    """
    setLogCond(0, "", 0, "", 5)
    db = ngamsDb("TESTSRV", "ngas_dev", "ngas_dbo", "ngas_dbo_pw")
    db.close()
    res1 = []
    try:
        res1 = db.query("SELECT total_disk_write_time FROM ngas_disks")
    except Exception, e:
        print e
    print "res1="+str(res1)
    #print db.diskInDb("DiskId-1-1--")

    
# EOF
