#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: ngamsDbm.py,v 1.11 2004/08/04 16:45:30 ngasmgr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  25/03/2002  Created
#

"""
Contains definition of class for handling a DBM DB (BSDDB).
"""

import sys, os, bsddb, cPickle, random

from ngams import *


class ngamsDbm:
    """
    Class implementing interface to DBM DB.
    """

    def __init__(self,
                 dbmName,
                 cleanUpOnDestr = 0,
                 writePerm = 0,
                 autoSync = 100):
        """
        Constructor method.
        
        dbmName:           Name of the DBM DB file (string).
        
        cleanUpOnDestr:    If set to 1, the DBM file will be removed on
                           object destruction (integer/0|1).

        writePerm:         Open with write permission (integer/0|1).

        autoSync:          Synchonize the DB after the specified number
                           of changes have been introduced (integer).
        """
        # Add proper extension if not already added.
        tmpName, tmpExt = os.path.splitext(dbmName)
        if (tmpExt != "." + NGAMS_DBM_EXT): dbmName += "." + NGAMS_DBM_EXT

        self.__dbmObj         = None
        self.__dbmName        = dbmName
        self.__cleanUpOnDestr = cleanUpOnDestr
        self.__writePerm      = writePerm
        self.__keyPtr         = None
        self.__sem            = threading.Semaphore(1)
        self.__autoSync       = autoSync
        self.__changeCount    = 0
        if (not os.path.exists(dbmName)):
            info(4,"DBM file: %s does not exist - creating ..." % dbmName)
            perm = "c"
        elif (writePerm):
            info(4,"DBM file: %s being opened for writing ..." % dbmName)
            perm = "w"
        else:
            info(4,"DBM file: %s being opened for reading ..." % dbmName)
            perm = "r"
        info(4,"Opening/creating DBM: " + dbmName)
        self.__dbmObj = bsddb.hashopen(dbmName, perm)
        if (perm == "c"):
            self.__dbmObj[NGAMS_FILE_DB_COUNTER] = cPickle.dumps(0, 1)
            self.__dbmObj.sync()
        self.__dbmOpen = 1
        info(4,"Opened/created DBM: " + dbmName)


    def __del__(self):
        """
        Destructor method cleaning up.
        """
        if (self.__dbmObj): self.__dbmObj.sync()
        if (self.__cleanUpOnDestr): rmFile(self.__dbmName)


    def getDbmName(self):
        """
        Return the filename of the DBM.

        Returns:   DBM filename (string).
        """
        return self.__dbmName
    

    def cleanUp(self):
        """
        Close a possible open DB connection + remove the DB file from the disk.

        Returns:   Reference to object itself (ngasmDbm).
        """
        if (self.__dbmObj): self.__dbmObj.sync()
        rmFile(self.__dbmName)
        return self


    def _incrDbCount(self,
                     val):
        """
        Increment (or decrement) the DBM counter.

        val:      Value with which to increment/decrement the counter
                  (integer).

        Returns:  Reference to object itself (ngamsDbm).
        """
        newVal = (cPickle.loads(self.__dbmObj[NGAMS_FILE_DB_COUNTER]) + val)
        self.__dbmObj[NGAMS_FILE_DB_COUNTER] = cPickle.dumps(newVal, 1)
        return self


    def add(self,
            key,
            object,
            sync = 0):
        """
        Add an element in the DBM.

        key:       Key in DBM (string).

        object:    Object to store in connection with key (<Object>).

        sync:      Synchronize the DB to disk after adding the object
                   (integer/0|1).

        Returns:   Reference to object itself (ngamsDbm).
        """
        try:
            self.__sem.acquire()
            dbVal = cPickle.dumps(object, 1)
            if (not self.__dbmObj.has_key(key)): self._incrDbCount(1)
            self.__dbmObj[key] = dbVal
            self.__changeCount += 1
            if (sync):
                self.__dbmObj.sync()
            elif (self.__changeCount >= self.__autoSync):
                self.__dbmObj.sync()
                self.__changeCount = 0
            self.__sem.release()
            return self
        except Exception, e:
            self.__sem.release()
            raise e


    def addIncKey(self,
                  object,
                  sync = 0):
        """
        Add an element in the DBM, generate the key automatically
        (incrementally).

        object:    Object to store in connection with key (<Object>).

        sync:      Synchronize the DB to disk after adding the object
                   (integer/0|1).

        Returns:   Reference to object itself (ngamsDbm).
        """
        newKey = str(self.getCount() + 1)
        return self.add(newKey, object, sync)


    def rem(self,
            key):
        """
        Remove an element from the DBM. If the referred element is not
        contained in the DB, an exception is thrown.

        key:       Name of key (string).

        Returns:   Reference to object itself (ngamsDbm).
        """
        info(5,"Entering rem() ...")
        try:
            self.__sem.acquire()
            del self.__dbmObj[key]
            self._incrDbCount(-1)
            self.__sem.release()
            info(5,"Leaving rem()")
            return self
        except Exception, e:
            self.__sem.release()
            raise e


    def keys(self):
        """
        Return the keys in the DBM. Internal keys (of the form __<Key>__)
        are not returned.

        Returns:   List with the keys in the DBM (list).
        """
        keyList = self.__dbmObj.keys()
        for idx in range((len(keyList) - 1), -1, -1):
            if (keyList[idx].find("__") == 0): del keyList[idx]
        return keyList
        
        
    def hasKey(self,
               key):
        """
        Return 1 if the given key is in the DBM.

        key:      Key (string).

        Returns:  1 = key in DB otherwise 0 is returned (integer/0|1).
        """
        try:
            self.__sem.acquire()
            hasKey = self.__dbmObj.has_key(key)
            self.__sem.release()
            return hasKey
        except Exception, e:
            self.__sem.release()
            raise e       
        

    def sync(self):
        """
        Synchronize the DBm to disk.

        Returns:    Reference to object itself (ngamsDbm).
        """
        try:
            self.__sem.acquire()
            self.__dbmObj.sync()
            self.__sem.release()
            return self
        except Exception, e:
            self.__sem.release()
            raise e       
        

    def get(self,
            key,
            pop = 0):
        """
        Get a specific element from the DBM.

        key:       Name of the key referring to the element (string).

        Returns:   Element or None if not available (<Object>).
        """
        if (self.__dbmObj.has_key(key)):
            return cPickle.loads(self.__dbmObj[key])
        else:
            return None


    def initKeyPtr(self):
        """
        Initialize the internal key pointer. Subsequent calls to
        getNext() will then start from the beginning.

        Returns:   Reference to object itself (ngamsDbm).
        """
        self.__keyPtr = None
        return self
        

    def getNext(self,
                pop = 0):
        """
        Get the keys + objects sequentially from the DB. If pop=1,
        the entry will be removed from the DB before returning.

        Entries for keys of the format: '__<Key>__' are skipped.

        pop:        Remove the element from the DB after retrieving it
                    (integer/0|1).

        Returns:    Tuple with key + value (unpickled) (tuple).
        """
        info(5,"Entering getNext() ...")
        try:
            self.__sem.acquire()
            if (self.__keyPtr):
                while (self.__keyPtr):
                    try:
                        self.__keyPtr, dbVal = self.__dbmObj.next()
                    except:
                        self.__keyPtr, dbVal = (None, None)   
                    if (str(self.__keyPtr).find("__") != 0): break
            else:
                try:
                    self.__keyPtr, dbVal = self.__dbmObj.first()
                    while (str(self.__keyPtr).find("__") == 0):
                        self.__keyPtr, dbVal = self.__dbmObj.next()
                except:
                    self.__keyPtr, dbVal = (None, None)   
            if (self.__keyPtr and pop):
                del self.__dbmObj[self.__keyPtr]
                self._incrDbCount(-1)
                self.__changeCount += 1
                if (self.__changeCount >= self.__autoSync):
                    self.__dbmObj.sync()
                    self.__changeCount = 0
            self.__sem.release()
            info(5,"Leaving getNext()")
            if (not self.__keyPtr):
                return (None, None)
            else:
                return (self.__keyPtr, cPickle.loads(dbVal))
        except Exception, e:
            self.__sem.release()
            raise e


    def getRandom(self,
                  pop = 0):
        """
        Return a random element from the DB.

        pop:      Remove element from DBM (integer/0|1).

        Returns:  Tuple with key + value (unpickled) (tuple).
        """
        info(5,"Entering getRandom() ...")
        try:
            self.__sem.acquire()       
            keys = self.__dbmObj.keys()
            idx = int((random.random() * (len(keys) - 1)) + 0.5)
            key = keys[idx]
            del keys
            
            key, val = self.__dbmObj[key]
            val = cPickle.loads(val)
            if (pop):
                del self.__dbmObj[key]
                self._incrDbCount(-1)
                self.__changeCount += 1
                if (self.__changeCount >= self.__autoSync):
                    self.__dbmObj.sync()
                    self.__changeCount = 0
            self.__sem.release()
            info(5,"Leaving getRandom()")
            return (key, val)
        except Exception, e:
            self.__sem.release()
            raise e

        
    def getCount(self):
        """
        Return the number of elements stored in the DBM.

        Returns:    Number of elements stored in the DBM (integer).
        """
        return cPickle.loads(self.__dbmObj[NGAMS_FILE_DB_COUNTER])


if __name__ == '__main__':
    """
    Main function invoking the function to dump the contents of a DBM.
    """
    import Sybase
    
    if (len(sys.argv) != 2):
        print "\nCorrect usage:\n\n" +\
              "> python ngamsDbm <DBM File>\n\n"
        sys.exit(1)

    dbm = ngamsDbm(sys.argv[1], 0, 0)
    print 80 * "="
    print "Dumping contents of NG/AMS DBM:"
    print 80 * "-"
    print "Name:     %s" % dbm.getDbmName()
    print "Elements: %d\n" % dbm.getCount()
    print "Contents:"
    print 80 * "-"
    while (1):
        key, contents = dbm.getNext(0)
        if (not key): break
        print "%s:" % key
        print str(contents)
        print 80 * "-"
    print 80 * "=" + "\n"
    

# EOF
