#******************************************************************************
# ESO/DFS
#
# "@(#) $Id: ngamsJanitorThread.py,v 1.129 2006/01/13 15:56:52 jknudstr Exp $"
#
# Who       When        What
# --------  ----------  -------------------------------------------------------
# jknudstr  29/01/2002  Created
#

"""
This module contains the code for the Janitor Thread, which is used to perform
various background activities as cleaning up after processing, waking up
suspended NGAS hosts, suspending itself.
"""

# IMPL: Give overhaul to handling of the DB Snapshot: Use ngamsDbm instead
#       of bsddb + simplify the algorithm.

import os, sys, time, thread, threading, random, glob, cPickle, bsddb
import base64

import pcc, PccUtTime

from ngams import *
import ngamsLib, ngamsDbm, ngamsDb, ngamsEvent, ngamsHighLevelLib
import ngamsFileList, ngamsStatus
import ngamsFileInfo, ngamsDiskInfo, ngamsPlugInApi, ngamsArchiveUtils
import ngamsSrvUtils, ngamsNotification


def startJanitorThread(srvObj):
    """
    Start the Janitor Thread.

    srvObj:     Reference to server object (ngamsServer).
    
    Returns:    Void.
    """
    info(3,"Starting Janitor Thread ...")
    # The srvObj._janitorThreadRunSync threading.Event() object is
    # used to synchronize the stopping of the Janitor Thread from
    # the Python Main Thread. The Janitor Thread will run as long
    # as the event is set (true), if it is set to false, the thread
    # will stop. Before exiting, the thread will set the event back
    # to true to acknowledge that it received the instruction to
    # stop execution. The main thread will then set it to false.
    srvObj._janitorThreadRunSync.set()
    args = (srvObj, None)
    srvObj._janitorThread = threading.Thread(None, janitorThread,
                                             NGAMS_JANITOR_THR, args)
    srvObj._janitorThread.setDaemon(0)
    srvObj._janitorThread.start()
    srvObj.setJanitorThreadRunning(1)
    info(3,"Janitor Thread started")


def stopJanitorThread(srvObj):
    """
    Stop the Janitor Thread.

    srvObj:     Reference to server object (ngamsServer).
    
    Returns:    Void.
    """
    if (not srvObj.getJanitorThreadRunning()): return
    info(3,"Stopping Janitor Thread ...")
    srvObj._janitorThreadRunSync.clear()
    srvObj._janitorThreadRunSync.wait(10)
    srvObj._janitorThreadRunSync.clear()
    srvObj._janitorThread = None
    srvObj.resetJanitorThreadRunCount()
    info(3,"Janitor Thread stopped")


def checkStopJanitorThread(srvObj):
    """
    Used to check if the Janitor Thread should be stopped and in case
    yes, to stop it.

    srvObj:     Reference to server object (ngamsServer).
    
    Returns:    Void.
    """
    if (not srvObj._janitorThreadRunSync.isSet()):
        info(2,"Stopping Janitor Thread")
        srvObj._janitorThreadRunSync.set()
        srvObj.setJanitorThreadRunning(0)
        raise Exception, "_STOP_JANITOR_THREAD_"
    

def checkCleanDirs(startDir,
                   dirExp,
                   fileExp,
                   useLastAccess):
    """
    Check a tree of directories. Delete all empty directories older than
    the given Directory Expiration. Also files are deleted if the file is
    older than File Expiration given.

    startDir:       Starting directory. The function will move downwards from
                    this starting point (string).
    
    dirExp:         Expiration time in seconds for directories. Empty
                    directories older than this time are deleted (integer).
    
    fileExp:        Expiration time in seconds for file. Empty file older than
                    this time are deleted (integer).

    useLastAccess:  Rather than using the creation time as reference, the
                    last modification and access date should be used
                    (integer/0|1).

    Returns:        Void.
    """
    info(5,"Entering _checkCleanDir() ...")
    timeNow = time.time()
    # IMPL: Potential memory bottleneck. Use 'find > file' as for REGISTER
    #       Command.
    entryList = glob.glob(startDir + "/*")
    # Work down through the directories in a recursive manner. If some
    # directories are not deleted during this run because they have contents,
    # they might be deleted during one of the following runs.
    for entry in entryList:
        if (not useLastAccess):
            refTime = getFileCreationTime(entry)
        else:
            refTime1 = getFileModificationTime(entry)
            refTime2 = getFileAccessTime(entry)
            if (refTime1 > refTime2):
                refTime = refTime1
            else:
                refTime = refTime2
        if (os.path.isdir(entry)):
            checkCleanDirs(entry, dirExp, fileExp, useLastAccess)
            tmpGlobRes = glob.glob(entry + "/*")
            if (tmpGlobRes == []):
                if ((timeNow - refTime) > dirExp):
                    info(4,"Deleting temporary directory: " + entry)
                    rmFile(entry)
        else:
            if (fileExp):
                if ((timeNow - refTime) > fileExp):
                    info(4,"Deleting temporary file: " + entry)
                    rmFile(entry)
    info(5,"Leaving _checkCleanDir()")


def _addInDbm(snapShotDbObj,
              key,
              val,
              sync = 0):
    """
    Add an entry in the DB Snapshot. This entry is pickled in binary format.

    snapShotDbObj:    Snapshot DB file (bsddb).
    
    key:              Key in DB (string).
    
    val:              Value to be put in the DB (<object>).

    sync:             Sync the DB to the DB file (integer/0|1).
    
    Returns:          Void.
    """
    snapShotDbObj[key] = cPickle.dumps(val, 1)
    if (sync): snapShotDbObj.sync()


def _readDb(snapShotDbObj,
            key):
    """
    Read and unpickle a value referenced by its key from the file DB.

    snapShotDbObj:   Open DB object (bsddb).

    key:             Key to extract value from (string).

    Returns:         Void.
    """
    return cPickle.loads(snapShotDbObj[key])
                     

def _genFileKey(fileInfo):
    """
    Generate a dictionary key from information in the File Info object,
    which is either a list with information from ngas_files, or an
    ngamsFileInfo object.

    fileInfo:       File Info as read from the ngas_files table or
                    an instance of ngamsFileInfo (list|ngamsFileInfo).

    Returns:        File key (string).
    """
    if ((type(fileInfo) == types.ListType) or
        (type(fileInfo) == types.TupleType)):
        fileId  = fileInfo[ngamsDb.NGAS_FILES_FILE_ID]
        fileVer = fileInfo[ngamsDb.NGAS_FILES_FILE_VER]
    else:
        fileId  = fileInfo.getFileId()
        fileVer = fileInfo.getFileVersion()
    return ngamsLib.genFileKey(None, fileId, fileVer)


##############################################################################
# DON'T CHANGE THESE IDs!!!
##############################################################################
NGAMS_SN_SH_ID2NM_TAG   = "___ID2NM___"
NGAMS_SN_SH_NM2ID_TAG   = "___NM2ID___"
NGAMS_SN_SH_MAP_COUNT   = "___MAP_COUNT___"
##############################################################################

def _encName(dbSnapshot,
             name):
    """
    Encode a name and add the name and its corresponding mapping ID (integer)
    in the file DB. The mapping is such that the name itself is referred to by

      NGAMS_SN_SH_ID2NM_TAG + ID -> <Name>

    The get from the name to the corresponding ID the following mapping
    should be used:
    
      NGAMS_SN_SH_NM2ID_TAG + <Name> -> <ID>

    dbSnapshot:      Open DB object (bsddb).
    
    name:            Name to be encoded (string).

    Returns:         The ID allocated to that name (integer).
    """
    nm2IdTag = NGAMS_SN_SH_NM2ID_TAG + name
    if (dbSnapshot.has_key(nm2IdTag)):
        nameId = _readDb(dbSnapshot, nm2IdTag)
    else:
        if (dbSnapshot.has_key(NGAMS_SN_SH_MAP_COUNT)):
            count = (_readDb(dbSnapshot, NGAMS_SN_SH_MAP_COUNT) + 1)
            #_addInDbm(dbSnapshot, NGAMS_SN_SH_MAP_COUNT, count)
        else:
            #_addInDbm(dbSnapshot, NGAMS_SN_SH_MAP_COUNT, 0)
            count = 0
        #nameId = _readDb(dbSnapshot, NGAMS_SN_SH_MAP_COUNT)
        nameId = count
        id2NmTag = NGAMS_SN_SH_ID2NM_TAG + str(nameId)

        # Have to ensure that all three keys are entered in the DBM (this might
        # not be the right way, maybe there is something that can be done at
        # bsddb level.
        try:
            _addInDbm(dbSnapshot, NGAMS_SN_SH_MAP_COUNT, count)
        except Exception, e:
            _addInDbm(dbSnapshot, NGAMS_SN_SH_MAP_COUNT, count)
            _addInDbm(dbSnapshot, nm2IdTag, nameId)
            _addInDbm(dbSnapshot, id2NmTag, name, 1)
            raise e
        try:
            _addInDbm(dbSnapshot, nm2IdTag, nameId)
        except Exception, e:
            _addInDbm(dbSnapshot, NGAMS_SN_SH_MAP_COUNT, count)
            _addInDbm(dbSnapshot, nm2IdTag, nameId)
            _addInDbm(dbSnapshot, id2NmTag, name, 1)
            raise e
        _addInDbm(dbSnapshot, id2NmTag, name, 1)
        
    return nameId


def _unPickle(pickleObject):
    """
    Unpickle a pickled object.

    pickleObject:   Pickled object (<Pickle Data>).

    Returns:        Reference to unpickled object (<Object>).
    """
    return cPickle.loads(pickleObject)


def _encFileInfo(dbSnapshot,
                 fileInfo):
    """
    Encode the information about a file contained in a list as read
    from the DB and generate a dictionary with these values. The column
    names are encoded and mappings between the code (ID) and name are stored
    in the file DB.

    The elements in the list can be refferred to by the 'constants'

      ngamsDb.NGAS_FILES_DISK_ID ... ngamsDb.NGAS_FILES_CREATION_DATE

    dbSnapshot:    Open DB object (bsddb).

    fileInfo:      List with information about file from the DB (list).

    Returns:       Dictionary with encoded column names (dictionary).
    """
    tmpDic = {}
    for n in range(ngamsDb.NGAS_FILES_CREATION_DATE + 1):
        colName = ngamsDb._ngasFilesNameMap[n]
        colId = _encName(dbSnapshot, colName)
        tmpDic[colId] = fileInfo[n]
    return tmpDic


def _encFileInfo2Obj(dbSnapshot,
                     encFileInfoDic):
    """
    Convert an encoded file info from the snapshot into an NG/AMS File
    Info Object.

    dbSnapshot:      Open DB object (bsddb).

    encFileInfoDic:  Dictionary containing the encoded file information
                     (dictionary).

    Returns:         NG/AMS File Info Object (ngamsFileInfo).
    """
    sqlFileInfo = []
    for n in range (ngamsDb.NGAS_FILES_CREATION_DATE + 1):
        sqlFileInfo.append(None)
    idxKeys = encFileInfoDic.keys()
    for idx in idxKeys:
        colName = _readDb(dbSnapshot, NGAMS_SN_SH_ID2NM_TAG + str(idx))
        sqlFileInfoIdx = ngamsDb._ngasFilesNameMap[colName]
        sqlFileInfo[sqlFileInfoIdx] = encFileInfoDic[idx]
    tmpFileInfoObj = ngamsFileInfo.ngamsFileInfo().unpackSqlResult(sqlFileInfo)
    return tmpFileInfoObj
    
        
def _updateSnapshot(ngamsCfgObj):
    """
    Return 1 if the DB Snapshot should be updated, otherwise 0 is
    returned.

    ngamsCfgObj:   NG/AMS Configuration Object (ngamsConfig).

    Returns:       1 = update DB Snapshot, 0 = do not update DB Snapshot
                   (integer/0|1). 
    """
    if (ngamsCfgObj.getAllowArchiveReq() or ngamsCfgObj.getAllowRemoveReq()):
        return 1
    else:
        return 0
        

def _openDbSnapshot(ngamsCfgObj,
                    mtPt):
    """
    Open a bsddb file DB. If the file exists and this is not
    a read-only NGAS system the file is opened for reading and writing.
    If this is a read-only NGAS system it is only opened for reading.

    If the file DB does not exist, a new DB is created.

    If the file DB does not exist and this is a read-only NGAS system,
    None is returned.

    The name of the DB file is:

      <Disk Mount Point>/NGAMS_DB_DIR/NGAMS_DB_NGAS_FILES

    ngamsCfgObj:    NG/AMS Configuration Object (ngamsConfig).
      
    mtPt:           Mount point (string).

    Returns:        File DB object (bsddb|None).
    """
    snapShotFile = os.path.normpath(mtPt + "/" + NGAMS_DB_DIR + "/" +\
                                    NGAMS_DB_NGAS_FILES)
    checkCreatePath(os.path.normpath(mtPt + "/" + NGAMS_DB_CH_CACHE))
    if (os.path.exists(snapShotFile)):
        if (_updateSnapshot(ngamsCfgObj)):
            # Open the existing DB Snapshot for reading and writing.
            snapshotDbm = bsddb.hashopen(snapShotFile, "w")
        else:
            # Open only for reading.
            snapshotDbm = bsddb.hashopen(snapShotFile, "r")
    else:
        if (_updateSnapshot(ngamsCfgObj)):
            # Create a new DB Snapshot.
            snapshotDbm = bsddb.hashopen(snapShotFile, "c")
        else:
            # There is no DB Snapshot and it is not possible to
            # create one - the check cannot be carried out.
            snapshotDbm = None

    # Remove possible, old /<mt pt>/.db/NgasFiles.xml snapshots.
    # IMPL: Remove when it can be assumed that all old XML snapshots have
    #       been removed.
    rmFile(os.path.normpath(mtPt + "/" + NGAMS_DB_DIR + "/NgasFiles.xml"))

    return snapshotDbm


def _delFileEntry(dbConObj,
                  fileInfoObj):
    """
    Delete a file entry in the NGAS DB. If the file does not exist,
    nothing is done. If the file exists, it will be deleted.

    dbConObj:        NG/AMS DB object (ngamsDB).

    fileInfoObj:     File Info Object (ngamsFileInfo).
    
    Returns:         Void.
    """
    info(4,"Entering _delFileEntry() ...")
    if (dbConObj.fileInDb(fileInfoObj.getDiskId(),
                          fileInfoObj.getFileId(),
                          fileInfoObj.getFileVersion())):
        try:
            dbConObj.deleteFileInfo(fileInfoObj.getDiskId(),
                                    fileInfoObj.getFileId(),
                                    fileInfoObj.getFileVersion(), 0)
        except:
            pass
    info(4,"Leaving _delFileEntry()")


def checkUpdateDbSnapShots(srvObj):
    """
    Check if a DB Snapshot exists for the DB connected. If not, this is
    created according to the contents of the NGAS DB (if possible). During
    this creation it is checked if the file are physically stored on the
    disk.

    srvObj:        Reference to NG/AMS server class object (ngamsServer). 

    Returns:       Void.
    """
    tmpSnapshotDbm = None
    lostFileRefsDbm = None
    
    if (not srvObj.getCfg().getDbSnapshot()):
        info(3,"NOTE: DB Snapshot Feature is switched off")
        return

    info(4,"Entering checkUpdateDbSnapShots() ...")

    info(4,"Generate list of disks to check ...")
    tmpDiskIdMtPtList = srvObj.getDb().getDiskIdsMtPtsMountedDisks(getHostId())
    diskIdMtPtList = []
    for diskId, mtPt in tmpDiskIdMtPtList:
        diskIdMtPtList.append([mtPt, diskId])
    diskIdMtPtList.sort()
    info(4,"Generated list of disks to check: " + str(diskIdMtPtList))

    # Generate temporary snapshot filename.
    ngasId = ngamsHighLevelLib.genNgasId(srvObj.getCfg())
    tmpDir = ngamsHighLevelLib.getTmpDir(srvObj.getCfg())

    # Temporary DBM with file info from the DB.
    tmpSnapshotDbmName = os.path.normpath(tmpDir + "/" + ngasId + "_" +\
                                          NGAMS_DB_NGAS_FILES)

    # Temporary DBM to contain information about 'lost files', i.e. files,
    # which are registered in the DB and found in the DB Snapshot, but
    # which are not found on the disk.
    info(4,"Create DBM to hold information about lost files ...")
    lostFileRefsDbmName = os.path.normpath(tmpDir + "/" + ngasId +\
                                           "_LOST_FILES")
    rmFile(lostFileRefsDbmName + "*")
    lostFileRefsDbm = ngamsDbm.ngamsDbm(lostFileRefsDbmName, writePerm=1)
    info(4,"Created DBM to hold information about lost files")

    # Carry out the check.
    for mtPt, diskId in diskIdMtPtList:
        info(2,"Check/create/update DB Snapshot for disk with " +\
             "mount point: " + mtPt)

        snapshotDbm = _openDbSnapshot(srvObj.getCfg(), mtPt)
        if (snapshotDbm == None): continue
        
        # The scheme for synchronizing the Snapshot and the DB is:
        #
        # - Loop over file entries in the Snapshot:
        #  - If in DB:
        #    - If file on disk     -> OK, do nothing.
        #    - If file not on disk -> Accumulate + issue collective warning.
        #
        #  - If entry not in DB:
        #    - If file on disk     -> Add entry in DB.
        #    - If file not on disk -> Remove entry from Snapshot.
        #
        # - Loop over entries for that disk in the DB:
        #  - If entry in Snapshot  -> OK, do nothing.
        #  - If entry not in Snapshot:
        #    - If file on disk     -> Add entry in Snapshot.
        #    - If file not on disk -> Remove entry from DB.

        # Create a temporary DB Snapshot with the files from the DB.
        #
        # IMPL: This algorithm could be improved such that the intermediate
        #       DBM (tmpSnapshotDbm) is not created. I.e., tmpFileListDbm
        #       is used diretly futher down.
        tmpSnapshotDbm = None
        tmpFileListDbm = None
        tmpFileListDbmName = None
        try:
            rmFile(tmpSnapshotDbmName + "*")
            tmpSnapshotDbm = bsddb.hashopen(tmpSnapshotDbmName, "c")
            tmpFileListDbmName = srvObj.getDb().dumpFileInfoList(diskId,
                                                                 ignore=None)
            tmpFileListDbm = ngamsDbm.ngamsDbm(tmpFileListDbmName)
            while (1):
                key, fileInfo = tmpFileListDbm.getNext()
                if (not key): break
                fileKey = _genFileKey(fileInfo)
                encFileInfoDic = _encFileInfo(tmpSnapshotDbm, fileInfo)
                _addInDbm(tmpSnapshotDbm, fileKey, encFileInfoDic)
                checkStopJanitorThread(srvObj)
                time.sleep(0.005)
            tmpSnapshotDbm.sync()
            del tmpFileListDbm
            rmFile(tmpFileListDbmName)
        except Exception, e:
            if (tmpSnapshotDbm): del tmpSnapshotDbm
            rmFile(tmpSnapshotDbmName)
            if (tmpFileListDbmName): rmFile(tmpFileListDbmName)
            if (tmpFileListDbm): del tmpFileListDbm
            raise e

        #####################################################################
        # Loop over the possible entries in the DB Snapshot and compare
        # these against the DB.
        #####################################################################
        info(4,"Loop over file entries in the DB Snapshot - %s" % diskId)
        count = 0
        try:
            key, pickleValue = snapshotDbm.first()
        except:
            key = None
        while (key):
            value = _unPickle(pickleValue)

            # Check if an administrative element, if yes add it if necessary.
            if (key.find("___") != -1):
                if (not tmpSnapshotDbm.has_key(key)):
                    tmpSnapshotDbm[key] = pickleValue
            else:
                tmpFileObj = _encFileInfo2Obj(snapshotDbm, value)
                complFilename = os.path.normpath(mtPt + "/" +\
                                                 tmpFileObj.getFilename())
                
                # Is the file in the DB?
                if (tmpSnapshotDbm.has_key(key)):
                    # Is the file on the disk?
                    if (not os.path.exists(complFilename)):
                        fileVer = tmpFileObj.getFileVersion()
                        tmpFileObj.setTag(complFilename)
                        fileKey = ngamsLib.genFileKey(tmpFileObj.getDiskId(),
                                                      tmpFileObj.getFileId(),
                                                      fileVer)
                        lostFileRefsDbm.add(fileKey, tmpFileObj)
                        lostFileRefsDbm.sync()
                elif (not tmpSnapshotDbm.has_key(key)):
                    tmpFileObj = _encFileInfo2Obj(snapshotDbm, value)

                    # Is the file on the disk?
                    if (os.path.exists(complFilename)):
                        # Add this entry in the NGAS DB.
                        tmpFileObj.write(srvObj.getDb(), 0, 1)
                        tmpSnapshotDbm[key] = pickleValue
                    else:
                        # Remove this entry from the DB Snapshot.
                        if (_updateSnapshot(srvObj.getCfg())):
                            del snapshotDbm[key]
                    del tmpFileObj
 
            # Be friendly, make a break every now and then + sync the DB file.
            count += 1
            if ((count % 100) == 0):
                if (_updateSnapshot(srvObj.getCfg())): snapshotDbm.sync()
                checkStopJanitorThread(srvObj)
                tmpSnapshotDbm.sync()
                time.sleep(0.010)
            else:
                time.sleep(0.002)
            try:
                key, pickleValue = snapshotDbm.next()
            except:
                key = None
        info(4,"Looped over file entries in the DB Snapshot - %s" % diskId)
        # End-Loop: Check DB against DB Snapshot. ###########################
        if (_updateSnapshot(srvObj.getCfg())): snapshotDbm.sync()
        tmpSnapshotDbm.sync()

        info(2,"Checked/created/updated DB Snapshot for disk with " +\
             "mount point: " + mtPt)

        #####################################################################
        # Loop over the entries in the DB and compare these against the
        # DB Snapshot.
        #####################################################################
        info(4,"Loop over the entries in the DB - %s..." % diskId)
        count = 0
        try:
            key, pickleValue = tmpSnapshotDbm.first()
        except:
            key = None
        while (key):
            value = _unPickle(pickleValue)
 
            # Check if it is an administrative element, if yes add it if needed
            if (key.find("___") != -1):
                if (not snapshotDbm.has_key(key)):
                    snapshotDbm[key] = pickleValue
            else:
                # Is the file in the DB Snapshot? 
                if (not snapshotDbm.has_key(key)):
                    tmpFileObj = _encFileInfo2Obj(tmpSnapshotDbm, value)

                    # Is the file on the disk?
                    complFilename = os.path.normpath(mtPt + "/" +\
                                                     tmpFileObj.getFilename())
                    if (os.path.exists(complFilename)):
                        # Add this entry in the DB Snapshot.
                        if (_updateSnapshot(srvObj.getCfg())):
                            snapshotDbm[key] = pickleValue
                    else:
                        # Remove this entry from the DB (if it is there).
                        _delFileEntry(srvObj.getDb(), tmpFileObj)
                    del tmpFileObj
                else:
                    # We always update the DB Snapshot to ensure it is
                    # in-sync with the DB entry.
                    if (_updateSnapshot(srvObj.getCfg())):
                        snapshotDbm[key] = pickleValue
           
            # Be friendly and make a break every now and then +
            # sync the DB file.
            count += 1
            if ((count % 100) == 0):
                if (_updateSnapshot(srvObj.getCfg())): snapshotDbm.sync()
                checkStopJanitorThread(srvObj)
                time.sleep(0.010)
            else:
                time.sleep(0.002)
            try:
                key, pickleValue = tmpSnapshotDbm.next()
            except:
                key = None
        info(4,"Checked DB Snapshot against DB - %s" % diskId)
        # End-Loop: Check DB Snapshot against DB. ###########################
        if (_updateSnapshot(srvObj.getCfg())): snapshotDbm.sync()
        snapshotDbm.close()

        # Make a small break between each disk/mount point.
        time.sleep(0.1)

    # Check if lost files found.
    info(4,"Check if there are Lost Files ...")
    noOfLostFiles = lostFileRefsDbm.getCount()
    if (noOfLostFiles):
        statRep = os.path.normpath(tmpDir + "/" + ngasId +\
                                   "_LOST_FILES_NOTIF_EMAIL.txt")
        fo = open(statRep, "w")
        timeStamp = PccUtTime.TimeStamp().getTimeStamp()
        tmpFormat = "JANITOR THREAD - LOST FILES DETECTED:\n\n" +\
                    "==Summary:\n\n" +\
                    "Date:                       %s\n" +\
                    "NGAS Host ID:               %s\n" +\
                    "Lost Files:                 %d\n\n" +\
                    "==File List:\n\n"
        fo.write(tmpFormat % (timeStamp, getHostId(), noOfLostFiles))

        tmpFormat = "%-32s %-32s %-12s %-80s\n"
        fo.write(tmpFormat % ("Disk ID", "File ID", "File Version",
                              "Expected Path"))
        fo.write(tmpFormat % (32 * "-", 32 * "-", 12 * "-", 80 * "-"))

        # Loop over the files an generate the report.
        lostFileRefsDbm.initKeyPtr()
        while (1):
            key, fileInfoObj = lostFileRefsDbm.getNext()
            if (not key): break
            diskId      = fileInfoObj.getDiskId()
            fileId      = fileInfoObj.getFileId()
            fileVersion = fileInfoObj.getFileVersion()
            filename    = fileInfoObj.getTag()
            fo.write(tmpFormat % (diskId, fileId, fileVersion, filename))
        fo.write("\n\n==END\n")
        fo.close()
        ngamsNotification.notify(srvObj.getCfg(), NGAMS_NOTIF_DATA_CHECK,
                                 "LOST FILE(S) DETECTED", statRep,
                                 [], 1, NGAMS_TEXT_MT,
                                 NGAMS_JANITOR_THR + "_LOST_FILES", 1)
        rmFile(statRep)
    info(4,"Checked if there are Lost Files. Number of lost files: %d" %\
         noOfLostFiles)

    # Clean up.
    del tmpSnapshotDbm
    rmFile(tmpSnapshotDbmName + "*")
    del lostFileRefsDbm
    rmFile(lostFileRefsDbmName + "*")

    info(4,"Leaving checkUpdateDbSnapShots()")


def checkDbChangeCache(srvObj,
                       diskId,
                       diskMtPt):
    """
    The function merges the information in the DB Change Snapshot Documents
    in the DB cache area on the disk concerned, into the Main DB Snapshot
    Document in a safe way which prevents that any information is lost.

    srvObj:        Reference to NG/AMS server class object (ngamsServer). 

    diskId:        ID for disk (string).
    
    diskMtPt:      Mount point of the disk, e.g. '/NGAS/disk1' (string).
    
    Returns:       Void.
    """
    if (not srvObj.getCfg().getDbSnapshot()): return
    if (not _updateSnapshot(srvObj.getCfg())): return

    info(5,"Entering checkDbChangeCache() ...")
    snapshotDbm = None
    try:
        snapshotDbm = _openDbSnapshot(srvObj.getCfg(), diskMtPt)
        if (snapshotDbm == None): return

        # Remove possible, old /<mt pt>/.db/cache/*.xml snapshots.
        # IMPL: Remove when it can be assumed that all old XML snapshots have
        #       been removed.
        rmFile(os.path.normpath(diskMtPt + "/" + NGAMS_DB_CH_CACHE + "/*.xml"))

        # Update the Status document with the possibly new entries.
        # IMPL: Potential memory bottleneck. Use 'find > file' as for
        #       REGISTER Command.
        dbCacheFilePat = os.path.normpath(diskMtPt + "/" + NGAMS_DB_CH_CACHE +\
                                          "/*.pickle")
        tmpCacheFiles = glob.glob(dbCacheFilePat)
        tmpCacheFiles.sort()
        cacheStatObj = None
        count = 0
        fileCount = 0
        noOfCacheFiles = len(tmpCacheFiles)
        timer = PccUtTime.Timer()
        for cacheFile in tmpCacheFiles:
            checkStopJanitorThread(srvObj)
            cacheStatObj = ngamsLib.loadObjPickleFile(cacheFile)
            if (isinstance(cacheStatObj, ngamsFileInfo.ngamsFileInfo)):
                tmpFileInfoObjList = [cacheStatObj]
                operation = cacheStatObj.getTag()
            else:
                cacheFileListObj = cacheStatObj.getFileListList()[0]
                tmpFileInfoObjList = cacheFileListObj.getFileInfoObjList()
                operation = cacheFileListObj.getComment()

            # Loop over the files in the temporary snapshot.
            for tmpFileInfoObj in tmpFileInfoObjList:
                fileKey = _genFileKey(tmpFileInfoObj)
                fileInfoList = tmpFileInfoObj.genSqlResult()
                encFileInfoDic = _encFileInfo(snapshotDbm, fileInfoList)
                if ((operation == NGAMS_DB_CH_FILE_INSERT) or
                    (operation == NGAMS_DB_CH_FILE_UPDATE)):
                    _addInDbm(snapshotDbm, fileKey, encFileInfoDic)
                    tmpFileInfoObj.write(srvObj.getDb(), 0)
                elif (operation == NGAMS_DB_CH_FILE_DELETE):
                    if (snapshotDbm.has_key(fileKey)): del snapshotDbm[fileKey]
                    _delFileEntry(srvObj.getDb(), tmpFileInfoObj)
                else:
                    # Should not happen.
                    pass
            del cacheStatObj
            
            # Sleep if not last iteration (or if only one file).
            fileCount += 1
            if (fileCount < noOfCacheFiles): time.sleep(0.010)
            
            # Synchronize the DB.
            count += 1
            if (count == 100):
                snapshotDbm.sync()
                checkStopJanitorThread(srvObj)
                count = 0

        # Clean up, delete the temporary File Remove Status Document.
        snapshotDbm.sync()
        snapshotDbm.close()
        for cacheFile in tmpCacheFiles:
            rmFile(cacheFile)
        totTime = timer.stop()

        tmpMsg = "Handled DB Snapshot Cache Files. Mount point: %s. " +\
                 "Number of Cache Files handled: %d."
        tmpMsg = tmpMsg % (diskMtPt, fileCount)
        if (fileCount):
            tmpMsg += "Total time: %.3fs. Time per file: %.3fs." %\
                      (totTime, (totTime / fileCount))
        info(4,tmpMsg)
        info(5,"Leaving checkDbChangeCache()")
    except Exception, e:
        try:
            if (snapshotDbm): snapshotDbm.sync()
        except:
            pass
        try:
            if (snapshotDbm): snapshotDbm.close()
        except:
            pass
        try:
            if (snapshotDbm): del snapshotDbm
        except:
            pass
        raise e


def updateDbSnapShots(srvObj,
                      diskInfo = None):
    """
    Check/update the DB Snapshot Documents for all disks.

    srvObj:            Reference to NG/AMS server class object (ngamsServer). 

    diskInfo:          If a Snapshot should only be updated for a specific
                       disk, this can be specifically indicated by giving
                       the Disk ID and Mount Point of the disk (list).
    
    Returns:           Void.
    """
    info(5,"Entering updateDbSnapShots() ...")
    if (diskInfo):
        diskId = diskInfo[0]
        mtPt = diskInfo[1]
        if (diskId and mtPt):
            mtPt = diskInfo[1]
        else:
            mtPt = srvObj.getDb().getMtPtFromDiskId(diskId)
        if (not mtPt):
            notice("No mount point returned for Disk ID: %s" % diskId)
            return
        try:
            checkDbChangeCache(srvObj, diskId, mtPt)
        except Exception, e:
            msg = "Error checking DB Change Cache for " +\
                  "Disk ID:mountpoint: %s:%s. Error: %s"
            msg = msg % (diskId, str(mtPt), str(e))
            error(msg)
            raise Exception, msg
    else:
        tmpDiskIdMtPtList = srvObj.getDb().\
                            getDiskIdsMtPtsMountedDisks(getHostId())
        diskIdMtPtList = []
        for diskId, mtPt in tmpDiskIdMtPtList:
            diskIdMtPtList.append([mtPt, diskId])
        diskIdMtPtList.sort()
        for mtPt, diskId in diskIdMtPtList:
            info(5,"Check/Update DB Snapshot Document for disk with " +\
                 "mount point: " + mtPt)
            try:
                checkDbChangeCache(srvObj, diskId, mtPt)
            except Exception, e:
                msg = "Error checking DB Change Cache for " +\
                      "Disk ID:mountpoint: %s:%s. Error: %s"
                msg = msg % (diskId, str(mtPt), str(e))
                error(msg)
                raise Exception, msg
    info(5,"Leaving updateDbSnapShots()")


def janitorThread(srvObj,
                  dummy):
    """
    The Janitor Thread runs periodically when the NG/AMS Server is
    Online to 'clean up' the NG/AMS environment. Task performed are
    checking if any data is available in the Back-Log Buffer, and
    archiving of these in case yes, checking if there are any Processing
    Directories to be deleted.

    srvObj:      Reference to server object (ngamsServer).

    dummy:       Needed by the thread handling ... 
    
    Returns:     Void.
    """
    # Make the event object to synchronize DB Snapshot updates available
    # for the ngamsDb class.
    dbChangeSync = ngamsEvent.ngamsEvent()
    srvObj.getDb().addDbChangeEvt(dbChangeSync)

    # => Update NGAS DB + DB Snapshot Document for the DB connected.
    try:
        checkUpdateDbSnapShots(srvObj)
    except Exception, e:
        errMsg = "Problem updating DB Snapshot files: " + str(e)
        warning(errMsg)

    suspendTime = ngamsLib.isoTime2Secs(srvObj.getCfg().\
                                        getJanitorSuspensionTime())
    while (1):
        # Incapsulate this whole block to avoid that the thread dies in
        # case a problem occurs, like e.g. a problem with the DB connection.
        try:
            checkStopJanitorThread(srvObj)
            info(5,"Janitor Thread running ...")

            ##################################################################
            # => Check if there are any Temporary DB Snapshot Files to handle.
            ##################################################################
            try:
                updateDbSnapShots(srvObj)
            except Exception, e:
                error("Error encountered updating DB Snapshots: " + str(e))

            # => Check Back-Log Buffer (if appropriate).
            if (srvObj.getCfg().getAllowArchiveReq() and \
                srvObj.getCfg().getBackLogBuffering()):
                info(5,"Checking Back-Log Buffer ...")
                try:
                    ngamsArchiveUtils.checkBackLogBuffer(srvObj)
                except Exception, e:
                    errMsg = genLog("NGAMS_ER_ARCH_BACK_LOG_BUF", [str(e)])
                    error(errMsg)
            ##################################################################
                    
            ##################################################################
            # => Check if we need to clean up Processing Directory (if
            #    appropriate). If a Processing Directory is more than
            #    30 minutes old, it is deleted.
            ##################################################################
            info(4,"Checking/cleaning up Processing Directory ...")
            procDir = os.path.normpath(srvObj.getCfg().\
                                       getProcessingDirectory() +\
                                       "/" + NGAMS_PROC_DIR)
            checkCleanDirs(procDir, 1800, 1800, 0)
            info(4,"Processing Directory checked/cleaned up")
            ##################################################################
            
            ##################################################################
            # => Check if there are old Requests in the Request DBM, which
            #    should be removed.
            ##################################################################
            reqDbm = None
            try:
                reqDbm = ngamsDbm.ngamsDbm(srvObj.getReqDbName(),
                                               cleanUpOnDestr = 0,
                                               writePerm = 1)
                reqDbm.initKeyPtr()
                while (1):
                    key, reqPropsObj = reqDbm.getNext()
                    if (not key): break
                    checkStopJanitorThread(srvObj)

                    # Remove a Request Properties Object from the queue if
                    #
                    # 1. The request handling is completed for more than
                    #    24 hours (86400s).
                    # 2. The request status has not been updated for more
                    #    than 24 hours (86400s).
                    # 3. If the object has not been touched for more than
                    #    24 hours (86400s).
                    timeNow = time.time()
                    if (reqPropsObj.getCompletionTime() != None):
                        complTime = reqPropsObj.getCompletionTime()
                        if ((timeNow - complTime) >= 86400):
                            info(4,"Removing request with ID from " +\
                                 "Request DBM: %s" % str(key))
                            srvObj.delRequest(key)
                            continue
                    if (reqPropsObj.getLastRequestStatUpdate() != None):
                        lastReq = reqPropsObj.getLastRequestStatUpdate()
                        if ((timeNow - lastReq) >= 86400):
                            srvObj.delRequest(key)
                            continue
                    if ((timeNow - reqPropsObj.getRequestTime()) >= 86400):
                        srvObj.delRequest(key)
                    time.sleep(0.020)
                if (reqDbM):  del reqDbm
            except Exception, e:
                # Assume that an exception means that there is no more
                # data in the DB.
                if (reqDbm): del reqDbm
            ##################################################################
                
            ##################################################################
            # => Check if we need to clean up Subscription Back-Log Buffer.
            ##################################################################
            info(4,"Checking/cleaning up Subscription Back-Log Buffer ...")
            backLogDir = os.path.\
                         normpath(srvObj.getCfg().getBackLogBufferDirectory()+\
                                  "/" + NGAMS_SUBSCR_BACK_LOG_DIR)
            expTime = ngamsLib.\
                      isoTime2Secs(srvObj.getCfg().getBackLogExpTime())
            checkCleanDirs(backLogDir, expTime, expTime, 0)
            info(4,"Subscription Back-Log Buffer checked/cleaned up")

            # => Check if there are left-over files in the NG/AMS Temp. Dir.
            info(4,"Checking/cleaning up NG/AMS Temp Directory ...")
            tmpDir = ngamsHighLevelLib.getTmpDir(srvObj.getCfg())
            expTime = (12 * 3600)
            checkCleanDirs(tmpDir, expTime, expTime, 1)
            info(4,"NG/AMS Temp Directory checked/cleaned up")

            # => Check for retained Email Notification Messages to send out.
            ngamsNotification.checkNotifRetBuf(srvObj.getCfg())

            # => Check if its time to carry out a rotation of the log file.
            logFile = srvObj.getCfg().getLocalLogFile()
            logPath = os.path.dirname(logFile)
            if (os.path.exists(srvObj.getCfg().getLocalLogFile())):
                info(4,"Checking if a Local Log File rotate is due ...")
                logFo = None
                try:
                    takeLogSem()

                    # For some reason we cannot use the creation date ...
                    logFo = open(logFile, "r")
                    while (1):
                        line = logFo.readline().strip()
                        if ((line == "") or (line.find("[INFO]") != -1)): break
                        time.sleep(0.005)
                    logFo.close()
                    if (line != ""):
                        creTime = line.split(" ")[0].split(".")[0]
                        logFileCreTime = ngamsLib.iso8601ToSecs(creTime)
                        logRotInt = ngamsLib.isoTime2Secs(srvObj.getCfg().\
                                                          getLogRotateInt())
                        deltaTime = (time.time() - logFileCreTime)
                        if (deltaTime >= logRotInt):
                            # It's time to rotate the current Local Log File.
                            rotLogFile = "LOG-ROTATE-" +\
                                         PccUtTime.TimeStamp().getTimeStamp()+\
                                         ".nglog"
                            rotLogFile = os.path.\
                                         normpath(logPath + "/" + rotLogFile)
                            PccLog.info(1, "Rotating log file: %s -> %s" %\
                                        (logFile, rotLogFile), getLocation())
                            logFlush()
                            commands.getstatusoutput("mv " + logFile + " " +\
                                                     rotLogFile)
                            open(logFile, "w").close()
                            msg = "NG/AMS Local Log File Rotated (%s)"
                            PccLog.info(1,msg % getHostId(), getLocation())
                    relLogSem()
                except Exception, e:
                    relLogSem()
                    if (logFo): logFo.close()
                    raise e
                info(4,"Checked for Local Log File rotatation") 
            ##################################################################
                
            ##################################################################
            # => Check if there are rotated Local Log Files to remove.
            ##################################################################
            info(4,"Check if there are rotated Local Log Files to remove ...")
            rotLogFilePat = os.path.normpath(logPath + "/LOG-ROTATE-*.nglog")
            rotLogFileList = glob.glob(rotLogFilePat)
            delLogFiles = (len(rotLogFileList) -\
                           srvObj.getCfg().getLogRotateCache())
            if (delLogFiles > 0):
                rotLogFileList.sort()
                for n in range(delLogFiles):
                    info(1,"Removing Rotated Local Log File: " +\
                         rotLogFileList[n])
                    rmFile(rotLogFileList[n])
            info(4,"Checked for expired, rotated Local Log Files")
            ##################################################################

            ##################################################################
            # => Check if there is enough disk space for the various
            #    directories defined.
            ##################################################################
            try:
                srvObj.checkDiskSpaceSat()
            except Exception, e:
                alert(str(e))
                alert("Bringing the system to Offline State ...")
                # We use a small trick here: We send an Offline Command to
                # the process itself.
                #
                # If authorization is on, fetch a key of a defined user.
                if (srvObj.getCfg().getAuthorize()):
                    authHdrVal = srvObj.getCfg().\
                                 getAuthHttpHdrVal(NGAMS_HTTP_INT_AUTH_USER)
                else:
                    authHdrVal = ""
                ngamsLib.httpGet(getHostName(), srvObj.getCfg().getPortNo(),
                                 NGAMS_OFFLINE_CMD, 0,
                                 [["force", "1"], ["wait", "0"]],
                                 "", 65536, 30, 0, authHdrVal)
            ##################################################################

            ##################################################################
            # => Check if this NG/AMS Server is requested to wake up
            #    another/other NGAS Host(s).
            ##################################################################
            timeNow = time.time()
            for wakeUpReq in srvObj.getDb().getWakeUpRequests():
                # Check if the individual host is 'ripe' for being woken up.
                suspHost = wakeUpReq[0]
                if (timeNow > wakeUpReq[1]):
                    info(2,"Found suspended NG/AMS Server: "+ suspHost + " " +\
                         "that should be woken up by this NG/AMS Server: " +\
                         getHostId() + " ...")
                    ngamsSrvUtils.wakeUpHost(srvObj, suspHost)
            ##################################################################

            ##################################################################
            # => Check if the conditions for suspending this NGAS Host are met.
            ##################################################################
            srvDataChecking = srvObj.getDb().getSrvDataChecking(getHostId())
            if ((not srvDataChecking) and 
                (srvObj.getCfg().getIdleSuspension()) and
                (not srvObj.getHandlingCmd())):
                timeNow = time.time()
                # Conditions are that the time since the last request was
                # handled exceeds the time for suspension defined.
                if ((timeNow - srvObj.getLastReqEndTime()) >=
                    srvObj.getCfg().getIdleSuspensionTime()):
                    # Conditions are met for suspending this NGAS host.
                    info(2,"NG/AMS Server: %s suspending itself ..." %\
                         getHostId())

                    # If Data Checking is on, we request a wake-up call.
                    if (srvObj.getCfg().getDataCheckActive()):
                        wakeUpSrv = srvObj.getCfg().getWakeUpServerHost()
                        nextDataCheck = srvObj.getNextDataCheckTime()
                        srvObj.reqWakeUpCall(wakeUpSrv, nextDataCheck)

                    # Now, suspend this host.
                    srvObj.getDb().markHostSuspended()
                    suspPi = srvObj.getCfg().getSuspensionPlugIn()
                    info(3,"Invoking Suspension Plug-In: " + suspPi + " to " +\
                         "suspend NG/AMS Server: " + getHostId() + " ...")
                    logFlush()
                    try:
                        exec "import " + suspPi
                        eval(suspPi + "." + suspPi + "(srvObj)")
                    except Exception, e:
                        errMsg = "Error suspending NG/AMS Server: " +\
                                 getHostId() + " using Suspension Plug-In: "+\
                                 suspPi + ". Error: " + str(e)
                        error(errMsg)
                        ngamsNotification.notify(srvObj.getCfg(),
                                                 NGAMS_NOTIF_ERROR,
                                                 "ERROR INVOKING SUSPENSION "+\
                                                 "PLUG-IN", errMsg)
            ##################################################################

            # PATCH: JANITOR-PLUG-IN
                      
            # Update the Janitor Thread run count.
            srvObj.incJanitorThreadRunCount()

            # Suspend the thread for the time indicated.
            info(5,"Janitor Thread executed - suspending for " +\
                 str(suspendTime) + "s ...")
            startTime = time.time()
            while ((time.time() - startTime) < suspendTime):
                checkStopJanitorThread(srvObj)
                # Check if we should update the DB Snapshot.
                if (dbChangeSync.isSet()):
                    time.sleep(0.5)
                    try:
                        tmpList = dbChangeSync.getEventInfoList()
                        dbChangeSync.clear()
                        diskInfo = None
                        if (tmpList):
                            for diskInfo in tmpList:
                                updateDbSnapShots(srvObj, diskInfo)
                    except Exception, e:
                        if (diskInfo):
                            msg = "Error encountered handling DB Snapshot " +\
                                  "for disk: %s/%s. Exception: %s"
                            msg = msg % (diskInfo[0], diskInfo[1], str(e))
                        else:
                            msg = "Error encountered handling DB Snapshot. " +\
                                  "Exception: %s" % str(e)
                        error(msg)
                        time.sleep(5)
                time.sleep(0.100)
                    
        except Exception, e:
            if (str(e).find("_STOP_JANITOR_THREAD_") != -1): thread.exit()
            errMsg = "Error occurred during execution of the Janitor " +\
                     "Thread. Exception: " + str(e)
            alert(errMsg)
            # We make a small wait here to avoid that the process tries
            # too often to carry out the tasks that failed.
            time.sleep(2.0)


# EOF
